# -*- test-case-name: atop.test.test_credup -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python.components import implements, Interface
from twisted.cred.portal import IRealm
from twisted.cred.credentials import IUsernamePassword, IUsernameHashedPassword
from twisted.cred.checkers import ICredentialsChecker, ANONYMOUS
from twisted.cred.error import UnauthorizedLogin
from twisted.python import log

from atop.powerup import Storeup
from atop.store import SubStore, Pool, NameIndex
from atop.fwd import ItemForwardingComponent

def _localpartDomainToDBEmail(localpart, domain):
    dparts = domain.split('.')
    dparts.reverse()
    revdomain = '.'.join(dparts)
    return (revdomain+'@'+localpart)


class Avatar(SubStore):

    securePassword = None
    insecurePassword = None

    def setSecurePassword(self, passwd):
        """Save away a string for use with protocols that do not expose your
        password on the network.
        """
        self.securePassword = passwd

    def setInsecurePassword(self, passwd):
        """Save away a string for use with protocols that expose your password
        on the internet.
        """
        self.insecurePassword = passwd

    def getPassword(self, passwordSecure):
        """Retrieve a password.
        """
        if passwordSecure:
            return self.securePassword
        else:
            return self.insecurePassword

    def __str__(self):
        return '<Avatar %r at %d in %s>' % (self.name, self.storeID, self.store)


class AnonymousUser:
    ## Maybe this is pointless
    ## Maybe the adapter being registered on this class
    ## Should just be registered on the Store itself
    def __init__(self, store):
        self.store = store


class Domain(Pool):
    def __init__(self, store, name):
        Pool.__init__(self, store, name)
        self.addIndex(NameIndex())

class Preauthenticated:
    __implements__ = IUsernameHashedPassword

    def __init__(self, username):
        self.username = username

    def checkPassword(self, password):
        return True

def preauth(store, username):
    """Utility function to retrieve an avatar from a database by name.
    """
    return store.getItemByID(
        ICredentialsChecker(store).
        requestAvatarId(Preauthenticated(username)))

class FwdChecker(ItemForwardingComponent):
    __implements__ = ICredentialsChecker
    def __init__(self, fwdTo):
        ItemForwardingComponent.__init__(self, fwdTo)
        self.credentialInterfaces = fwdTo.getItem().credentialInterfaces

class FwdRealm(ItemForwardingComponent):
    __implements__ = IRealm

class StoreCredentialsChecker(Storeup):

    __implements__ = ICredentialsChecker

    credentialInterfaces = (IUsernamePassword, IUsernameHashedPassword)

    def setUpPools(self, store):
        self.touch()
        domains = Pool(store, 'domains')
        domains.addIndex(NameIndex())
        self.domainPoolRef = domains.referenceTo()
        store.setComponent(ICredentialsChecker, self)

    def getDomains(self):
        """Retrieve an iterator of all domains in this store.
        """
        return self.domainPoolRef.getItem().queryIndex()

    def getAvatars(self):
        """Retrieve an iterator of all avatars regardless of domain.

        This will yield certain avatars twice.
        """
        for domain in self.getDomains():
            for avatar in domain.queryIndex():
                yield avatar

    def getDomain(self, name=None):
        return self.domainPoolRef.getItem().getIndexItem('name', name)

    def addDomain(self, domainName):
        x = Domain(self.store, domainName)
        self.domainPoolRef.getItem().addItem(x)
        return x

    def addAvatar(self, avatar, domainName=None):
        def _():
            domain = self.getDomain(domainName)
            av2 = domain.getIndexItem("name", avatar.name, None)
            if av2 is not None:
                raise UnauthorizedLogin("You can't choose that username in that domain.")
            domain.addItem(avatar)
        return self.store.transact(_)
    addAvatarToDomain = addAvatar

    def getAvatar(self, localpart, domainName=None):
        def _():
            domain = self.getDomain(domainName)
            return domain.getIndexItem("name", localpart)
        return self.store.transact(_)
    getAvatarFromDomain = getAvatar

    def requestAvatarId(self, credentials):
        passwordSecure = implements(credentials, IUsernameHashedPassword)
        try:
            usplit = credentials.username.split('@', 1)
            avatar = self.getAvatarFromDomain(*usplit)
            password = avatar.getPassword(passwordSecure)
            if credentials.checkPassword(password):
                return avatar.storeID
        except KeyError:
            pass
        raise UnauthorizedLogin()


class ILoginRelationRegistry(Interface):
    def addRefusalRelation(self, relation):
        """Add a relation to be examined for refusal of logins.

        @type relation: Callable taking an Avatar and an interface,
        and raising an Exception if login for that Avatar on that
        interface is to be denied.
        """

class StoreRealm(Storeup):

    dependencies = [StoreCredentialsChecker]

    __implements__ = IRealm,

    def setUpPools(self, store):
        self.relations = []
        store.setComponent(IRealm, self)
        # store.setComponent(ILoginRelationRegistry, self)

    def addRefusalRelation(self, relation):
        self.touch()
        self.relations.append(relation)

    def checkApproval(self, avatar, iface):
        for r in self.relations:
            try:
                r(avatar, iface)
            except Exception, e:
                raise UnauthorizedLogin(str(e))

    ##
    ## Actual IRealm implementation
    ##

    def requestAvatar(self, avatarId, mind, *interfaces):
        """
        """
        def _():
            if avatarId is ANONYMOUS:
                avatar = AnonymousUser(self.store)
            else:
                avatar = self.store.getItemByID(avatarId)
            for iface in interfaces:
                avatarAspect = iface(avatar, default=None)
                if avatarAspect is None:
                    continue

                ##
                ## Possibly the exception checkApproval raises should
                ## be caught, saved, and the loop continued.  If, after
                ## all interfaces have been examined, no access has been
                ## granted, the exception may be re-raised.  In this
                ## way, less desirable services (priority being determined
                ## by the order of the interfaces tuple, of course) may
                ## still be granted even when the configured relations
                ## deny access to those which are more desirable.  Which of
                ## the first, last, or some other arbitrary refusal should
                ## be propegated is unclear at this time.
                ##
                ## However, none of our code currently makes requests for
                ## multiple services, making the point moot (for now), so
                ## I will leave the implementation simple.  - jp
                ##

                self.checkApproval(avatar, iface)

                logout = getattr(avatarAspect, 'logout', noLogout)
                log.msg(login=avatarId, interface=iface, mind=mind)
                return (iface, avatarAspect, logout)
            log.msg(login=None, interface=iface, mind=mind)
            raise UnauthorizedLogin("You're not allowed to access that.")
        return self.store.transact(_)


def noLogout():
    pass
