# -*- test-case-name: atop.test_sched -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time

from twisted.application.service import Service
from twisted.python.components import Adapter, registerAdapter, Interface
from twisted.python import log

from atop.powerup import Powerup, Storeup, IPowerStation
from atop.store import INDEX_AND_ITEM, SubStore, Pool, DateIndex
from atop.fwd import ItemForwardingComponent


class IScheduler(Interface):
    """Marker interface for Scheduler.
    """
    def callItemLater(self, item):
        pass
    def clock(self):
        pass

class IScheduledEvent(Interface):
    """Interface which scheduled events must implement.

    Subclass Schedulable and don't worry about it, okay?  Also, subclass item.

    I know, this isn't really a component.  Live with it.
    """

class Schedulable:
    """I am a mix-in for items which wish to schedule timed tasks.
    """

    __implements__ = IScheduledEvent

    status = ''

    def schedule(self, newTime):
        """Schedule my run() method to be executed at the given local time.
        (Values returned from time.time() are local times, so use that and
        adjust.)
        """
        self.setIndexKey("scheduled", newTime)
        IScheduler(self.store).callItemLater(self)

    def delay(self, secondsLater):
        """Delay the scheduled call.
        @param secondsLater: how many seconds from its current firing time to delay
        """
        self.schedule(self.getIndexKey(None, "scheduled") + secondsLater)

    def reset(self, secondsFromNow):
        """Reset the scheduled call's timer.
        @param secondsFromNow: how many seconds from now it should fire,
            equivalent to C{self.cancel()} and then doing another
            C{reactor.callLater(secondsLater, ...)}
        """
        self.schedule(IScheduler(self.store).clock() + secondsFromNow)

    def unschedule(self):
        self.setIndexKey("scheduled", None)
        IScheduler(self.store).poolRef.getItem().removeItem(self)

    def run(self):
        """Implement
        """
        raise NotImplementedError("")

class SchedulerBase(Schedulable):
    __implements__ = IScheduler

    clock = staticmethod(time.time)

    maxEventsPerTick = 50

    def setUpPools(self, store):
        p = Pool(store, "Scheduled tasks")
        p.invisible = True
        p.addIndex(DateIndex("scheduled"))
        p.addIndex(DateIndex("scheduled_errors"))
        self.poolRef = p.referenceTo()
        getattr(store, 'rootRef', getattr(store, 'rootPoolRef', None)).getItem().addItem(p)

    def callItemLater(self, item):
        """Call an item later, according to its date index.

        You probably want to use Schedulable.schedule(newTime) rather than
        calling this directly.
        """
        self.poolRef.getItem().addItem(item)

    def getEvents(self):
        return self.poolRef.getItem().queryIndex("scheduled", result=INDEX_AND_ITEM)

    def callItemLater(self, item):
        p = self.poolRef.getItem()
        if item not in p:
            # These are almost always rescheduled items, not new
            # items, so the check saves us a function call, not
            # to mention cleans up our log output.
            p.addItem(item)
        for localtime, item in self.getEvents():
            self.schedule(localtime)
            return

    def tick(self):
        i = self.maxEventsPerTick
        now = self.clock()
        for localtime, item in self.getEvents():
            i -= 1
            if i and localtime < now:
                try:
                    self.store.subtransact(IScheduledEvent(item).run)
                except:
                    # TODO: later this should also actually catch the Failure
                    # here and pickle it somewhere.
                    log.err()
                    item.setIndexKey('scheduled', None)
                    item.setIndexKey('scheduled_errors', localtime)
                else:
                    when = item.getIndexKey(None, "scheduled")
                    if when is None:
                        self.poolRef.getItem().removeItem(item)
                    elif when <= localtime:
                        log.msg("%s did not unschedule itself!" % item)
                        item.setIndexKey('scheduled', None)
                        item.setIndexKey('scheduled_errors', localtime)
            else:
                behind = localtime - now
                if behind < 0:
                    log.msg("Warning: Scheduler falling behind (%0.2f seconds)" % -behind)
                return localtime


class SubTicker(Adapter):
    __implements__ = IScheduledEvent
    def __init__(self, original):
        Adapter.__init__(self, original)
        self.store = original

    def run(self):
        self.schedule(self.schedulerRef.getItem().tick())

    def schedule(self, newTime):
        return Schedulable.schedule.im_func(self.original, newTime)

    def unschedule(self):
        return Schedulable.unschedule.im_func(self.original)

registerAdapter(SubTicker, SubStore, IScheduledEvent)

class ISchedule(Interface):
    pass

class SystemScheduler(SchedulerBase, Storeup):
    __implements__ = SchedulerBase.__implements__, Powerup.__implements__
    persistenceForgets = ['nextCall']

    def setUpPools(self, store):
        SchedulerBase.setUpPools(self, store)
        store.setComponent(IScheduler, self)

    nextCall = None

    def callLater(self, secs, f, *a, **b):
        from twisted.internet import reactor
        return reactor.callLater(secs, f, *a, **b)

    def schedule(self, newTime):
        try:
            self.nextCall.cancel()
        except:
            pass
        self.nextCall = None
        if newTime is not None:
            self.nextCall = self.callLater(max(newTime - self.clock(),0), self.callLaterEr)

    def callLaterEr(self):
        nexttime = self.store.transact(self.tick)
        self.schedule(nexttime)

class SubScheduler(SchedulerBase, Powerup):
    __implements__ = SchedulerBase.__implements__, Powerup.__implements__, ISchedule
    powerupName = 'Scheduled Tasks'

    storeDependencies = [SystemScheduler]

    def setUpPools(self, substore):
        SchedulerBase.setUpPools(self, substore)
        substore.touch()
        IScheduledEvent(substore).schedulerRef = self.referenceTo()
        substore.setComponent(IScheduler, self)

    def schedule(self, newTime):
        IScheduledEvent(self.store).schedule(newTime)

class SchedulerService(Service):
    def __init__(self, store):
        # TODO: for store resurrection, this needs to be able to suck it out of
        # the powerups pool.
        sched = IPowerStation(store).installPowerup(SystemScheduler)
        self.ref = sched.referenceTo()

    def startService(self):
        Service.startService(self)
        self.ref.getItem().callLaterEr()

    def stopService(self):
        Service.stopService(self)
        self.ref.getItem().schedule(None)
