# -*- test-case-name: atop.test.test_store -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time
import os
import sys
import itertools
import operator

try:
    import cPickle as pickle
except ImportError:
    import pickle

from os.path import join as opj

from bsddb.db import *

from twisted.python import log
from twisted.python import plugin
from twisted.python import context
from twisted.python import failure
from twisted.python import components
from twisted.python.failure import Failure
from twisted.application import service
from twisted.internet import defer

from twisted.python import reflect
from twisted.python.components import Interface, Componentized

from weakref import ref

from twisted.persisted.styles import Versioned, Ephemeral, doUpgrade

from atop.tpython import Forgetter, mkgmtime
from atop.fwd import ItemForwardingComponent, makeForwarder

# This is a value that is greater than anything that might show up inside
# one part of a key string, but less than a special magic value that lets
# us know when a particular prefix has ended.  It cannot show up in any parts
# of any key names inserted into the database.
SEP = chr(0x00)
END = chr(0xff)

# We use this because index keys contain their storeIDs to allow multiple
# values per key.  When we do queries we need to make sure that they ignore the
# storeID.

MAGIC = chr(0x01)

# DB_TXN_WRITE_NOSYNC = 0x080000  #Write, don't sync, on txn commit.

def md(dirname):
    if os.path.isdir(dirname):
        return False
    os.makedirs(dirname)
    return True

MAX_LOCKS = 1000
MAX_LOCKERS = 1000
MAX_OBJECTS = 1000

def envdb(dirname, joining=False):
    """Open a BSDDB environment and a single database.
    """
    dirname = os.path.abspath(dirname)
    md(dirname)
    dirname = dirname
    envdir = opj(dirname,"env")
    datafile = opj(dirname, "db")
    md(envdir)
    env = DBEnv()

    env.set_lk_max_locks(MAX_LOCKS)
    env.set_lk_max_lockers(MAX_LOCKERS)
    env.set_lk_max_objects(MAX_OBJECTS)

    if joining:
        print 'I AM JOINING OMGOMGOMGOMG'
        baseFlags = (DB_JOINENV)
        env.open(envdir, baseFlags)
    else:
        baseFlags = (DB_INIT_MPOOL |  DB_INIT_LOG | DB_CREATE| DB_INIT_TXN
                     | DB_INIT_LOCK)
        for flag in DB_RECOVER, DB_RECOVER_FATAL:
            try:
                env.open(envdir, baseFlags | flag)
            except DBRunRecoveryError:
                if flag == DB_RECOVER_FATAL:
                    log.msg("FATAL RECOVERY FAILED.  CALL SLEEPYCAT IMMEDIATELY.")
                    raise
                else:
                    log.msg("Regular recovery failed.")
            else:
                break

    env.set_flags(DB_TXN_NOSYNC, 1)
    env.set_flags(DB_AUTO_COMMIT, 1)

    db = DB(env)
    db.set_flags(0)
    txn = env.txn_begin()
    try:
        db.open(datafile, "db", DB_BTREE, DB_CREATE, txn=txn)
    except:
        txn.abort()
        db.close()
        env.close()
        raise
    txn.commit()
    return env, db


def updateIDKey(db, keyName):
    txn = context.get('txn')
    if txn is None:
        raise NotImplementedError("Updating ID outside of transaction")
    val = db.get(keyName, flags=DB_RMW, txn=txn)
    if val is None:
        val = 0
    ret = int(val)
    db.put(keyName, str(ret + 1), txn=txn)
    return ret

class LoginNotFound(KeyError):
    pass

class DuplicateLogin(KeyError):
    pass

class NoSuchIndex(KeyError):
    pass

class Inconsistency(Exception):
    """
    A slightly weaker warning than an assert.  Something bad happened, but it
    was recoverable, and it's the kind of thing we need to consistently check
    for at runtime (it can't be disabled because it's not really specified by
    any invariant, and could be exploited by attackers).
    """

class AtomicFile(file):
    """I am a file which is moved from temporary to permanent storage when it
    is closed.

    After I'm closed, I will have a 'finalpath' property saying where I went.
    """
    def __init__(self, tempname, destdir):
        self._destdir = destdir
        file.__init__(self, tempname, 'wb')

    def close(self, destpath):
        file.close(self)
        fullpath = os.path.join(self._destdir, destpath)
        md(os.path.dirname(fullpath))
        self.finalpath = fullpath
        os.rename(self.name, fullpath)
        return fullpath

    def abort(self):
        os.unlink(self.name)

tempCounter = itertools.count()

def noLogout():
    pass

def _fcremover(w, k, f):
    def remove(o):
        try:
            f()
        except:
            log.msg("Exception in finalizer cannot be propegated")
            log.err()
        self = w()
        if self is not None:
            del self.data[k]
    return remove

class FinalizingCache:
    """Possibly useful for infrastructure?  This would be a nice addition (or
    perhaps even replacement) for twisted.python.finalize.
    """
    def __init__(self):
        self.data = {}
        self.has = self.data.has_key

    def cache(self, key, value):
        fin = value.__finalizer__()
        self.data[key] = ref(value, _fcremover(ref(self), key, fin))
        return value

    def uncache(self, key):
        del self.data[key]

    def get(self, key):
        o = self.data[key]()
        if o is None:
            raise Inconsistency("FinalizingCache has %r but its value is no more." % (key,))
        log.msg(db_cache="hit", key=key)
        return o

class Store(Versioned):
    try:
        from twisted.python.components import IComponentized
        __implements__ = IComponentized
    except ImportError:
        print 'TODO: add IComponentized to Twisted so that we can do this without relying on getattr magic.'

    store = None
    storeID = None
    persistenceForgets = ['cache', 'env', 'db']

    service = None

    def __init__(self, storeID, datadir, filedir):
        """
        @param storeID: A unique identifier allocated for this store by the master store.  If two
        servers in the same group use the same storeID, the world will end.  The master store should
        be instantiated with a storeID of 0.

        @param datadir: The directory in which to instantiate database files.

        @param filedir: The directory into which to place message files.
        """
        self.storeID = storeID
        self.filedir = filedir
        self.datadir = datadir
        self.tempdir = os.path.join(self.filedir, 'tmp')
        md(self.tempdir)
        dataWasntThere = md(self.datadir)
        self.setup()
        if dataWasntThere:
            self.transact(self.live)
        else:
            self.transact(self.resurrect)

    def resurrect(self):
        """Re-instantiate a database.
        """
        self.rootRef = ItemReference(self, 0)
        pupool = self.rootRef.getItem().getIndexItem('name', 'powerups')
        self.powerupsRef = pupool.referenceTo()
        from atop.powerup import PowerStation
        self.powerstation = PowerStation(self)

        # really, this ought to be handled by a different implementation of
        # IPowerStation considering the extra specialness that Store has to do,
        # but this at least fixes incorrect behavior with a relatively small
        # change... -gl

        for pup in pupool.queryIndex():
            self.powerstation.powerups[reflect.qual(pup.__class__)] = pup.referenceTo()

    def live(self):
        """This is the first time this store has been initialized.  Perform
        in-database initialization to create various objects which we will
        need.
        """

        # This is being done in this apparently round-about way (with root
        # vs. powerup pool) because we _need_ to be able to resurrect these
        # items deterministically if we lose track of the original Store
        # object.  If we lose the .tap, we need to be able to re-create it

        root = Pool(self, "root")
        root.addIndex(NameIndex())
        powerups = Pool(self, "powerups")
        powerups.addIndex(PowerupIndex())
        root.addItem(powerups)
        from atop.powerup import PowerStation
        self.powerstation = PowerStation(self)
        self.rootRef = root.referenceTo()
        self.powerupsRef = powerups.referenceTo()

    def getComponent(self, interface, registry=None, default=None):
        # I am my own registry, really.
        from atop.powerup import IPowerStation
        if interface == IPowerStation:
            return self.powerstation
        ppool = self.powerupsRef.getItem()
        pitem = ppool.getIndexItem("powerup", reflect.qual(interface), default=default)
        if isinstance(pitem, ItemForwardingComponent):
            return pitem.fwdTo.getItem()
        return pitem

    def setComponent(self, interface, newItemComponent):
        if newItemComponent.store is not self:
            raise Inconsistency("You may only store top-level items as top-level powerups.")
        newItemComponent.touch()
        newItemComponent.setIndexKey("powerup", reflect.qual(interface))
        self.powerupsRef.getItem().addItem(newItemComponent)

    def newSubFileForItem(self, substore):
        tmpname = os.path.join(self.tempdir, str(tempCounter.next())+".tmp")
        destpath = os.path.join(self.filedir, str(substore.storeID))
        return AtomicFile(tmpname, destpath)

    def newSubDirForItem(self, item, name):
        thisPath = os.path.join(self.filedir, str(item.storeID), name)
        md(thisPath)
        return thisPath

    def __setstate__(self, state):
        self.__dict__ = state
        self.setup()

    def setup(self):
        self.cache = FinalizingCache()
        self.env, self.db = envdb(self.datadir)

    def close(self):
        self.db.close()
        self.env.close()

    def nextID(self):
        """Allocate and return the next available store ID.
        """
        return updateIDKey(self.db, "next"+SEP+"id")

    def nextSubID(self, item):
        """Allocate and return the next available store ID for the given item.

        These IDs are independent of any other IDs, including the ones for this
        Store instance.
        """
        return updateIDKey(self.db, self._formatIDPath(
            item.getItemPath()) + SEP+'next'+SEP+'id')

    def setSubKey(self, item, key, value):
        ipfx = self._formatIDPath(item.getItemPath())
        self.db.put(ipfx + key, value, txn=context.get('txn'))

    def delSubKey(self, item, key, value):
        ipfx = self._formatIDPath(item.getItemPath())
        self.db.delete(ipfx + key, txn=context.get('txn'))

    def _loginkey(self, path, method):
        return 'login'+SEP+'%s/%s' % (method, path)

    def _loadItemByPath(self, ctxstore, *path):
        """
        Retrieve an item given a sequence of integers.
        """
        stringKey = self._formatIDPath(path)
        stringValue = self.db.get(stringKey, txn=context.get('txn'))
        if stringValue is None:
            return None
        myObject = context.call({"store": ctxstore}, pickle.loads, stringValue)
        log.msg(path=path, db_get=None, item=myObject, pickle=stringValue)
        doUpgrade()
        return myObject

    def cacheItem(self, item, path):
        assert isinstance(path, tuple)
        if self.cache.has(path):
            assert self.cache.get(path) is item, "inconsistency!?!  the db manages this now.  Cached item %r at %r is not %r" % (self.cache.get(path), path, item)
        else:
            self.cache.cache(path, item)

    def _saveItemToPath(self, item, *path):
        assert path[0] == self.storeID
        self.cacheItem(item, path)
        stringKey = self._formatIDPath(path)
        stringValue = context.call({'item': item}, pickle.dumps, item, pickle.HIGHEST_PROTOCOL)
        log.msg(path=path, db_put=None, item=item, pickle=stringValue)
        self.db.put(stringKey, stringValue, txn=context.get('txn'))

    def _getItemByPath(self, ctxstore, *path):
        assert path[0] == self.storeID
        if self.cache.has(path):
            return self.cache.get(path)
        else:
            obj = self._loadItemByPath(ctxstore, *path)
            if obj is None:
                raise KeyError(path)
            self.cacheItem(obj, path)
            obj.activate()
            return obj

    def getItemByID(self, itemID):
        """Return the item from this Store with the given ID.

        @type itemID: C{int}
        """
        return self._getItemByPath(self, self.storeID, itemID)

    def getSubItem(self, subStore, subItemID):
        p = subStore.getItemPath()
        p.append(subItemID)
        return self._getItemByPath(subStore, *p)

    def lockItem(self, subItem):
        self.lockSubItem(self, subItem)

    def lockSubItem(self, subStore, subItem):
        p = subItem.getItemPath()
        stringKey = self._formatIDPath(p)
        self.db.get(stringKey, flags=DB_RMW, dlen=0, doff=0, txn=context.get('txn'))
        log.msg(db_lock=stringKey, path=p, item=subItem)

    def _inMemoryRevert(self, collection):
        """Revert a collection of objects to their state in the database.
        """
        for item in collection:
            path = tuple(item.getItemPath())
            dupItem = self._loadItemByPath(item.store, *path)
            if dupItem is not None:
                try:
                    item.restoreStateFromStore(dupItem)
                except:
                    log.err()
            elif self.cache.has(path):
                self.cache.uncache(path)
            else:
                log.msg("Reversion peculiarity: %s" (item))

    def new_txn(self):
        return self.env.txn_begin(context.get('txn'))

    def subtransact(self, f, *a, **k):
        """Open a database transaction object and commit it.
        """
        txn = self.new_txn()
        storageContext = {}
        cursorFactory = CursorFactory(self, txn)
        try:
            retvalue = context.call({'txn': txn,
                                     'StorageContext': storageContext,
                                     'CursorFactory': cursorFactory},
                                    f, *a, **k)
            context.call({'txn': txn},
                         self._inMemoryCommit,
                         storageContext.itervalues())
        except:
            cursorFactory.abort()
            txn.abort()
            log.msg(db_txn="abort")
            # is there a way to make this transaction read-only?  -glyph
            revertReadTxn = self.new_txn()
            try:
                context.call({'txn': revertReadTxn},
                             self._inMemoryRevert, storageContext.itervalues())
            finally:
                revertReadTxn.abort()
            raise
        else:
            cursorFactory.commit()
            txn.commit()            # this can fail with a RUNRECOVERY - that would be bad.
            log.msg(db_txn="commit")
            return retvalue

    def _inMemoryCommit(self, collection):
        for item in collection:
            # save the item by ID, either overwriting or adding
            path = item.getItemPath()
            self._saveItemToPath(item, *path)

    def transact(self, f, *a, **k):
        """
        Run a function or method within the context of the storage system.

        This will keep track of all items created or touched during the
        duration of this function and save or revert them based on whether the
        function completes successfully.
        """
        if context.get('txn'):
            return f(*a,**k)
        return self.subtransact(f, *a, **k)

    def transback(self, result, f, *a, **k):
        """Utility method for using 'transact' as a callback to a Deferred.

        All this does is reverse the order of the first two arguments to
        'transact'.  Its common usage is:

            d = doSomethingLater()
            d.addCallback(store.transback, F, 'hello')

        As you might imagine, this causes F to be called as

            F(deferred-result, 'hello')
        """
        return self.transact(f, result, *a, **k)

    def _formatIDPath(self, path):
        L = ['%0.10d'] * len(path)
        L.insert(0, 'id')
        return ((SEP.join(L)) % tuple(path))

class CursorFactory:
    def __init__(self, store, txn):
        self.store = store
        self.txn = txn
        self.cursors = []
        self.done = False

    def cursor(self):
        if self.done:
            raise Inconsistency("Re-use of a storage factory.  You suck!")
        self.cursors.append(self.store.db.cursor(self.txn))
        return self.cursors[-1]

    def closeCursors(self):
        self.done = True
        for cursor in self.cursors:
            cursor.close()

    def commit(self):
        self.closeCursors()

    def abort(self):
        self.closeCursors()

class ItemReference:
    def __init__(self, store, storeID, iface=None):
        self.store = store
        self.storeID = storeID
        self.iface = iface

    def getItem(self):
        if self.iface is not None:
            return self.iface(self.store.getItemByID(self.storeID))
        return self.store.getItemByID(self.storeID)

    def __str__(self):
        return '<ItemReference to %d, %d>' % (self.store.storeID, self.storeID)

    def __hash__(self):
        return hash((self.store.storeID, self.storeID))

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'store'):
            st = context.get('store')
            if st is None:
                raise Inconsistency("You're loading an item reference outside the context of a store.")
            self.store = st

    def __ne__(self, other):
        #PYTHON IS SO AWESOME
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, ItemReference):
            return False
        return (other.store is self.store) and (other.storeID == self.storeID)

    def __getstate__(self):
        # get the store we're saving to.
        topitem = context.get('item')
        if topitem is None:
            return self.__dict__
##             raise Inconsistency("Why are you saving an item reference outside the scope of an item?  "
##                                 "Consider using RemoteItemReference (when it becomes available.  "
##                                 "Talk to glyph.")
        elif topitem.store is self.store:
            # If the store we're saving to is the same store as the one for the
            # item we're storing a reference to, remember it from context on
            # load.
            dct = self.__dict__.copy()
            del dct['store']
            return dct
        elif topitem is self.store:
            # Most likely, what's happening here is that we're saving a
            # reference to a sub-item of the avatar within the avatar itself.
            # This will only happen in substores, because 'topitem is
            # self.store' means it's a sub-item of the top item which we are
            # currently saving.

            # In this case it's perfectly safe to return a synchronous
            # reference to the store, because it's going to be remembered by
            # the pickler.
            return self.__dict__
        else:
            if self.getItem() is topitem.store:
                raise Inconsistency("Trying to store an ItemReference on %s to its own store! Don't *do* that." % (topitem,))
            raise Inconsistency("Ref to %s stored from %s in %s should be %s." % (self.getItem(),
                                                                                  topitem,
                                                                                  topitem.store,
                                                                                  self.getItem().store))


class Item(Forgetter, Versioned):
    persistenceForgets = ['store']
    contextRemembers = [('store', 'store')]
    name = None

    def __init__(self, store, name=None):
        if name is not None:
            self.name = name
        self.storeID = store.nextID()
        self.store = store
        self.store.cacheItem(self, tuple(self.getItemPath()))
        from atop.sched import IScheduler
        # self.dateCreated = time.gmtime(IScheduler(store).clock())
        self.dateCreated = time.gmtime(time.time())
        self.touch()

    def __statecheck__(self):
        if self is not context.get('item'):
            raise Inconsistency("You were saving %s but I found %s in it." % (context.get('item'), self))

    def __getstate__(self):
        self.__statecheck__()
        return Forgetter.__getstate__(self)

    def fwd(self, *interfaces):
        if not interfaces:
            interfaces = self.__implements__
        return makeForwarder(interfaces)(self.referenceTo())

    def referenceTo(self, iface=None):
        return ItemReference(self.store, self.storeID, iface)
    getRef = referenceTo

    def restoreStateFromStore(self, otherItem):
        """Called when a transaction is aborted.

        The default behavior is to reset the objects stored state to
        whatever it was before the transaction began and to discard all
        non-stored state.
        """
        self.__dict__ = otherItem.__dict__
        self.__class__ = otherItem.__class__
        # ^ If you need this, you are being faaaar too clever for quotient.

    def getItemPath(self):
        """Return a list describing this item's location in the main store.
        """
        ids = [self.storeID]
        ss = self
        while ss.store is not None:
            ss = ss.store
            ids.append(ss.storeID)
        ids.reverse()
        return ids

    def setDataKey(self, key, value):
        return self.store.setSubKey(self, key, value)

    def delDataKey(self, key, value):
        return self.store.delSubKey(self, key, value)

    def __finalizer__(self):
        return lambda : None

    def activate(self):
        """This method is called when this item has been loaded from the DB.
        """

    def touch(self):
        """
        Mark this object for persistence soon.

        This must be run in the context of StorQ.transact
        """
        sctx = context.get('StorageContext')
        if sctx is None:
            raise Inconsistency("You're touching an object outside a transaction.")
        ikey = tuple(self.getItemPath())
        sctx[ikey] = self

    def iterpools(self):
        if not self.pools:
            return
        for sid in self.pools.keys():
            yield self.store.getItemByID(sid)

    def change(self, notificationData=None):
        """Touch me.  Send out a notification as to what is about to change.

        @param notificationData: an object that represents the change.  XXX
        TODO: donovan, what kind of change object do you want here?
        """
        if self.pools:
            for value in self.iterpools():
                value._announceItem(self, "itemChanged", notificationData)

    def lock(self):
        """Prepare this object for database update.
        """
        self.touch()
        self.store.lockItem(self)

    def setIndexKey(self, indexName, newKey, notify=False):
        self.touch()
        if self.pools:
            for pool in self.iterpools():
                oldKey = self.getIndexKey(pool, indexName)
                changedKey = self._changeIndexKey(pool, indexName, newKey)
                pool.changeKeyForItem(self, indexName, oldKey, changedKey, notify=notify)
        self._setIndexKeyData(indexName, newKey)

    def appendToIndex(self, indexName, indexKey):
        """Be careful with this method.  The store will not automatically
        remove state added using this method from a pool.  removeFromIndex,
        similarly.
        """
        for pool in self.iterpools():
            pool.changeKeyForItem(self, indexName, None, indexKey)

    def removeFromIndex(self, indexName, indexKey):
        for pool in self.iterpools():
            pool.changeKeyForItem(self, indexName, indexKey, None)


    def getIndexKey(self, pool, indexName):
        """Get a key value for a particular index.  The key value must match
        what the index expects.
        """
        m = getattr(self, "index_" + indexName, None)
        if m:
            return m(pool)
        if self._indexes:
            return self._indexes.get(indexName)

    _indexes = None

    def _setIndexKeyData(self, indexName, newValue):
        m = getattr(self, "setindex_" + indexName, None)
        if m:
            return m(newValue)
        if not self._indexes:
            self._indexes = {}
        self._indexes[indexName] = newValue

    def _changeIndexKey(self, pool, indexName, newValue):
        m = getattr(self, "changeindex_" + indexName, None)
        if m:
            return m(pool, newValue)
        return newValue

    pools = None

    def _initpools(self):
        self.lock()
        self.pools = {}                 # {int poolID:int UID}

    def _addedToPool(self, pool):
        if pool.store != self.store:
            raise Inconsistency(
                "You can't add foriegn items directly to pools.")
        if not self.pools:
            self._initpools()
        if not self.pools.has_key(pool.storeID):
            self.touch()
            uid = pool.nextUID()
            self.pools[pool.storeID] = uid
        # else: already in pool?  what to do?

    def _removedFromPool(self, pool):
        if pool.store != self.store:
            raise Inconsistency(
                "You can't add foriegn items directly to pools.")
        self.touch()
        del self.pools[pool.storeID]

    def index_uid(self, pool):
        return self.pools.get(pool.storeID)

    def setindex_uid(self, newuid):
        raise NotImplementedError("You can't change an item's UID.")

    def setindex_name(self, newname):
        self.name = newname
        return newname

    def index_name(self, pool):
        return self.name

    def index_received(self, pool):
        ### TODO: Change the name of this index, to make a uniform 'date created' index
        return self.dateCreated

    def newCacheDir(self, dirname):
        cur = self.store
        while cur.store is not None:
            cur = cur.store
        thisPath = os.path.join(cur.filedir, str(self.store.storeID), 'cache', str(self.storeID), dirname)
        md(thisPath)
        return thisPath


class SubStore(Forgetter, Item, Componentized):
    def __init__(self, *a, **b):
        Componentized.__init__(self)
        Item.__init__(self, *a, **b)
        self.idcounter = 0
        p = Pool(self, name='My Pools')
        p.addIndex(NameIndex())
        p.suggestIndex('name', False)
        self.rootPoolRef = p.referenceTo()

    def getRootPool(self):
        """Return the root pool for this store.
        """
        return self.rootPoolRef.getItem()

    def __getstate__(self, dict=None):
        Item.__statecheck__(self)
        if dict is None:
            dict = self.__dict__.copy()
        x = dict['_adapterCache'] = dict['_adapterCache'].copy()
        # kill all ephemeral adapters.  Didn't Componentized do this at one
        # point!?!?
        for k,v in x.items():
            if isinstance(v, Ephemeral):
                del x[k]
        return Forgetter.__getstate__(self, dict)

    # make sure that all appropriate interfaces touch me before making a
    # potentially destructive change!!  TODO: add special-casing for ephemerals

    def setComponent(self, interfaceClass, component):
        self.touch()
        if isinstance(component, Item):
            # if it's going to be automatic, it might as well be symmetrical
            component = ItemForwardingComponent(component.referenceTo())
        return Componentized.setComponent(self, interfaceClass, component)

    def getComponent(self, *a, **k):
        component = Componentized.getComponent(self, *a, **k)
        if isinstance(component, ItemForwardingComponent):
            return component.fwdTo.getItem()
        return component

    def addComponent(self, *a, **k):
        self.touch()
        Componentized.addComponent(self, *a, **k)

    def unsetComponent(self, *a, **k):
        self.touch()
        Componentized.unsetComponent(self, *a, **k)

    def removeComponent(self, *a, **k):
        self.touch()
        Componentized.removeComponent(self, *a, **k)

    def newFile(self):
        return self.store.newSubFileForItem(self)

    def newDirectory(self, name):
        """Create a directory in the user's "address
        space" or return the existing one
        """
        return self.store.newSubDirForItem(self, name)

    def setSubKey(self, item, key, value):
        return self.store.setSubKey(item, key, value)

    def delSubKey(self, item, key, value):
        return self.store.delSubKey(item, key, value)

    def nextID(self):
        return self.store.nextSubID(self)

    def nextSubID(self, item):
        return self.store.nextSubID(item)

    def cacheItem(self, item, path):
        return self.store.cacheItem(item, path)

    def getItemByID(self, itemID):
        return self.store.getSubItem(self, itemID)

    def transact(self, f, *a, **k):
        return self.store.transact(f, *a, **k)

    def subtransact(self, f, *a, **k):
        return self.store.subtransact(f, *a, **k)

    def lockItem(self, item):
        return self.store.lockSubItem(self, item)

class StorageService(service.Service):
    def __init__(self, store):
        self.setName("atop.Store")
        self.store = store
        self.store.service = self
        self._startupNotifications = []

    def startService(self):
        def _():
            for i in self._startupNotifications:
                i.getItem().startup()
        from twisted.internet import reactor
        reactor.callWhenRunning(_)

    def stopService(self):
        self.store.close()

    def notifyOnStartup(self, item):
        self._startupNotifications.append(item.referenceTo())

# subsubstore.store = substore; substore.store = realstore;
# subsubstore.getItemByPath => context.get('store') == subsubstore

class IPoolIndex(Interface):
    def getIndexName(self):
        """Return a qualified name for this index, used to look up index_*
        methods on Items."""

    def keyToSortableString(self, key):
        """Convert a given 'key' object (such as a date) to a fixed-length
        sortable string: e.g. the date represented by the float
        1066147720.700978 might be converted to '2003/10/14/12/08/40'
        """

    def sortableStringToKey(self, sortableString):
        """Convert a value produced by keyToSortableString back into its
        original format.
        """

CREATED = 'created'

# the time that this object was first created by the system; when it was
# initially instantiated

RECEIVED = 'received'

# The time that this object was first reliably 'received' by a trusted
# forwarder.  Corresponds to a heuristic based on the last-found rfc822
# 'received' header.

MODIFIED = 'modified'

# Last time this object was changed.


class DateIndex:
    __implements__ = IPoolIndex
    def __init__(self, dateType=RECEIVED):
        self.iname = dateType

    def getIndexName(self):
        return self.iname

    def sortableStringToKey(self, sstr):
        return mkgmtime(tuple(map(int, sstr.split(".")))+(0,0,0))

    def keyToSortableString(self, key):
        if isinstance(key, (list, tuple, time.struct_time)):
            gmt = key
        else:
            gmt = time.gmtime(key)
        return "%04d.%02d.%02d.%02d.%02d.%02d" % gmt[:6]

class IntIndex:
    __implements__ = IPoolIndex

    name = 'int'
    def __init__(self, name='int'):
        self.name = name

    def getIndexName(self):
        return self.name

    def keyToSortableString(self, key):
        return '%0.10d' % key

    def sortableStringToKey(self, sstr):
        return int(sstr)

class UIDIndex(IntIndex):
    __implements__ = IPoolIndex

    name = 'uid'
    def __init__(self):
        IntIndex.__init__(self, self.name)

class NameIndex:
    __implements__ = IPoolIndex
    name = 'name'

    def __init__(self, name='name'):
        self.name = name

    def getIndexName(self):
        return self.name

    def keyToSortableString(self, key):
        return key

    def sortableStringToKey(self, sstr):
        return sstr

class PowerupIndex(NameIndex):
    __implements__ = IPoolIndex
    def getIndexName(self):
        return 'powerup'

class ItemReferenceIndex(Forgetter):
    '''Allow access to a "foreign key"
    '''

    persistenceForgets = ['store']
    contextRemembers = [('store','store')]

    def __init__(self, attrname, store):
        self.name = attrname
        self.store = store

    def getIndexName(self):
        return self.name

    def keyToSortableString(self, key):
        # key ought to be an ItemReference.
        return '%0.10d' % key.storeID

    def sortableStringToKey(self, sstr):
        return ItemReference(self.store, int(sstr))

ITEM           = object()
INDEX          = object()
INDEX_AND_ITEM = object()
STORE_ID       = object()

class IPipe(Interface):
    def itemAdded(self, item):
        pass

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, notificationData):
        pass

class Pool(Item):
    persistenceForgets = ['sequence', 'ephemeralPipes']
    sequence = None

    # These attributes are for use by UI.  I won't change the default sort
    # order because the semantics of queryIndex make sense as they are.

    suggestedIndex = 'uid'
    suggestedBackwards = True
    suggestedView = 'Table'

    def __init__(self, store, name=None):
        self.indexes = {}
        self.pipes = []
        self.activate()
        Item.__init__(self,store,name)
        self.addIndex(UIDIndex())

    def suggestIndex(self, index, backwards=False):
        self.suggestedIndex = index
        self.suggestedBackwards = backwards

    def suggestView(self, view):
        self.suggestedView = view

    def addPipe(self, pipe, ephemeral=False):
        """Add a 'pipe', an object that is notified of items arriving in this
        pool, leaving it, and changing while inside it.

        The 'ephemeral' flag is intended for use by UI elements and other
        objects which do not reside in the database.  You will need to keep a
        strong reference to the pool around somewhere if you want ephemeral
        notifications.  (Normally this will not be a big deal because you've
        already got the pool open for reading, writing, or both, somewhere else
        anyway.)

        DO NOT USE THE EPHEMERAL FLAG FOR OBJECTS INSIDE THE DATABASE.  Failure
        of a non-ephemeral observer will _not_ abort the transaction, and will
        therefore put your database into an inconsistent state if you use it
        incorrectly.
        """
        if isinstance(pipe, Item):
            pipe = pipe.referenceTo()
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.append(pipe)

    def removePipe(self, pipe, ephemeral=False):
        """See addPipe.

        Note that pipes added with the ephemeral flag _must_ be removed with
        the ephemeral flag.
        """
        if isinstance(pipe, Item):
            pipe = pipe.referenceTo()
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.remove(pipe)

    def activate(self):
        self.ephemeralPipes = []

    def restoreStateFromStore(self, other):
        ep = self.ephemeralPipes
        Item.restoreStateFromStore(self, other)
        self.ephemeralPipes = ep

    def _getSuperStore(self):
        ss = self.store
        while ss.store is not None:
            ss = ss.store
        return ss

    def queryIndex(self, iname='uid', startKey=None, endKey=None, result=ITEM,
                   backwards=False, limit=None, startItemID=None, endItemID=None):
        """Retrieve information about the contents of this pool.

        A generator which yields the requested information about the requested Items is
        returned.  This generator I{must} be iterated in a storage context.

        @type iname: C{str}
        @param iname: The index to query.

        @type endKey: A type appropriate for the specified index.
        @param endKey: If specified, the last key at which to return a result.

        @param result: If specified as C{ITEM}, return the L{Item} itself
        for each content of the pool returned.  If specified as C{INDEX},
        return the key by which each item is indexed in this pool.  If
        specified as C{STORE_ID}, return the store ID for each content of
        the pool returned.

        @raise NoSuchIndex: If the specified index does not exist for this pool.
        """
        indexPfx = self._getSuperStore()._formatIDPath(self.getItemPath()) + SEP + 'index' + SEP + iname + SEP
        try:
            indexObj = self.indexes[iname]
        except KeyError:
            raise NoSuchIndex("Pool %r has no index %r" % (self.name, iname))

        if startKey is not None:
            startStr = indexPfx + indexObj.keyToSortableString(startKey)
            if startItemID:
                startStr += SEP + ('%0.10d' % startItemID)
            elif backwards:
                startStr += MAGIC
        elif backwards:
            startStr = indexPfx + END
        else:
            startStr = indexPfx

        if endKey is not None:
            endStr = indexPfx + indexObj.keyToSortableString(endKey)
            if endItemID:
                endStr += SEP + ('%0.10d' % endItemID)
            elif not backwards:
                endStr += MAGIC
        elif backwards:
            endStr = indexPfx
        else:
            endStr = indexPfx + END

        c = context.get('CursorFactory').cursor()
        dummy = c.set_range(startStr)
        if backwards:
            oper = operator.lt
            gogogo = c.prev
            correctForDB = c.next
            if dummy[0] == startStr:
                correctForDB(dlen=0,doff=0)
        else:
            oper = operator.gt
            gogogo = c.next
            correctForDB = c.prev
            correctForDB(dlen=0, doff=0)

        if oper(startStr, endStr):
            raise Inconsistency("%squery starting at %s ending at %s will have no items: specify them in order!." %
                                (backwards and 'backwards ' or '', startStr.split(SEP), endStr.split(SEP)))
        count = 1
        for skey, svalue in iter(gogogo, None):
            if limit is not None:
                if count > limit:
                    return
                count += 1
            if oper(skey, endStr): # I need an exit.
                return
            if result == STORE_ID:
                storeID = int(svalue)
                yield storeID
            elif result == ITEM:
                storeID = int(svalue)
                yield self.store.getItemByID(storeID)
            elif result == INDEX:
                indexString = skey[len(indexPfx):].split(SEP)[0]
                indexKeyObj = indexObj.sortableStringToKey(indexString)
                yield indexKeyObj
            elif result == INDEX_AND_ITEM:
                indexString = skey[len(indexPfx):].split(SEP)[0]
                indexKeyObj = indexObj.sortableStringToKey(indexString)
                storeID = int(svalue)
                item = self.store.getItemByID(storeID)
                yield (indexKeyObj, item)
            else:
                raise Inconsistency(
                    "Invalid result-mode specified for index query: %s" % (result,))

    def getSequence(self):
        if self.sequence is None:
            self.sequence = list(self.queryIndex('uid',result=INDEX))
        return self.sequence

    _noIndexItem = object()

    def getIndexItem(self, indexName, indexKey=None, default=_noIndexItem):

        """Retrieve an item from an index.  If no indexKey is specified, I will
        return the first item in the index.

        Upon failure, (if there are no items in the index, or the specified key
        is not found), I will raise KeyError, or return the 'default' argument,
        if it is supplied.
        """
        def getter():
            for k, v in self.queryIndex(indexName, startKey=indexKey, result=INDEX_AND_ITEM):
                # only iterates once, haha.
                if indexKey is None:
                    return v
                if k != indexKey:
                    if default is Pool._noIndexItem:
                        raise KeyError(indexKey)
                    return default
                return v
            if default is Pool._noIndexItem:
                raise KeyError(indexKey)
            return default
        return self.store.transact(getter)

    def nextUID(self):
        return self.store.nextSubID(self)

    def _manipItem(self, item, adding):
        """
        Either add or remove an Item from this Pool. This method is used by
        addItem and removeItem.
        """
        if adding:
            announceMethodName = "itemAdded"
            item._addedToPool(self)
        else:
            announceMethodName = "itemRemoved"
        for ipref, index in self.indexes.items():
            iname = index.getIndexName()
            ikey = item.getIndexKey(self, iname)
            if adding:
                oldKey = None
                newKey = ikey
            else:
                oldKey = ikey
                newKey = None
            self.changeKeyForItem(item, ipref, oldKey, newKey)
        if not adding:
            item._removedFromPool(self)
        self._announceItem(item, announceMethodName)

    def __contains__(self, item):
        if item.pools:
            return item.pools.has_key(self.storeID)
        return False

    def changeKeyForItem(self, item, ipref, oldKey, newKey, notify=False):
        if not self.indexes.has_key(ipref):
            return
        index = self.indexes[ipref]
        sisid = '%0.10d' % item.storeID
        if oldKey is not None:
            # delete the old key.
            istr = index.keyToSortableString(oldKey)
            self.delDataKey(SEP+'index'+SEP+ipref+SEP+istr+SEP+sisid, sisid)
        if newKey is not None:
            # add the new key.
            istr = index.keyToSortableString(newKey)
            self.setDataKey(SEP+'index'+SEP+ipref+SEP+istr+SEP+sisid, sisid)

    def _announceItem(self, item, announceMethodName, *args):
        for pipe in self.pipes:
            getattr(pipe, announceMethodName)(item, *args)
        for epipe in self.ephemeralPipes:
            # UI elements are non-transactional anyway, so we will not honor
            # their failures.  DO NOT ADD EPHEMERAL PIPES FOR DATABASE OBJECTS.
            try:
                getattr(epipe, announceMethodName)(item, *args)
            except:
                log.err()

    def addItem(self, item):
        if item in self:
            import warnings
            warnings.warn("Adding %s to %s, but it is already there." % (item, self),
                          Warning, stacklevel=2)
            return
        self._manipItem(item, True)

    def removeItem(self, item):
        self._manipItem(item, False)

    def addIndex(self, index, name=None):
        self.lock()
        if name is None:
            name = index.getIndexName()
        self.indexes[name] = index

    def removeIndex(self, prefix):
        self.lock()
        del self.indexes['prefix']

class HardPoint(Item):
    """
    I almost called this a PoolPipe since
    it is part Pool and part Pipe. It is a pipe
    that can be attached to one pool (via setPoint()).
    Other pipes can then be added to this faux pool.
    This allows pipes to be insulated from the source
    being moved.
    """
    __implements__ = IPipe
    persistenceForgets = ['ephemeralPipes']

    def __init__(self, store, name=None):
        Item.__init__(self,store,name)
        self.pipes = []
        self.activate()
        self.sourcePoolRef = None
        self.sourcePipe = None

    def setPoint(self, pool):
        if self.sourcePoolRef:
            p = self.sourcePoolRef.getItem()
            p.removePipe(self.fwd(IPipe))
        self.sourcePoolRef = pool.referenceTo()
        pool.addPipe(self.fwd(IPipe))
        self.touch()

    def addPipe(self, pipe, ephemeral=False):
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.append(pipe)

    def removePipe(self, pipe, ephemeral=False):
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.remove(pipe)

    def activate(self):
        self.ephemeralPipes = []

    def restoreStateFromStore(self, other):
        ep = self.ephemeralPipes
        Item.restoreStateFromStore(self, other)
        self.ephemeralPipes = ep

    def _announceItem(self, item, announceMethodName, *args):
        for pipe in self.pipes:
            getattr(pipe, announceMethodName)(item, *args)
        for epipe in self.ephemeralPipes:
            # UI elements are non-transactional anyway, so we will not honor
            # their failures.  DO NOT ADD EPHEMERAL PIPES FOR DATABASE OBJECTS.
            try:
                getattr(epipe, announceMethodName)(item, *args)
            except:
                log.err()

    def itemAdded(self, item):
        self._announceItem(item, 'itemAdded')

    def itemRemoved(self, item):
        self._announceItem(item, 'itemRemoved')

    def itemChanged(self, item, notificationData):
        self._announceItem(item, 'itemChanged')
