# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import gc
import os
import time
import shutil
from os.path import join as opj

from atop.store import Store, Item, SubStore, FinalizingCache
from atop.store import Pool, DateIndex, UIDIndex
from atop.store import ITEM, INDEX, STORE_ID
from atop.store import Inconsistency, NameIndex
from atop.test import unittest
from atop import fwd

from twisted.cred.portal import Portal
from twisted.cred.credentials import UsernamePassword
from twisted.python.components import Interface
from twisted.python.util import sibpath
from twisted.python import log
from twisted.trial import unittest as trialtest

class OtherNameIndex(NameIndex):
    def getIndexName(self):
        return "name_other"

class ExampleException(Exception):
    pass

class ExampleAvatar(SubStore):

    def __init__(self, store, frobnitzes=-99):
        self.frobnitzes = frobnitzes
        SubStore.__init__(self, store)

    def addFrobnitzThenBreak(self):
        self.touch()
        self.frobnitzes += 1
        raise ExampleException('whoops sorry')

    def noReallyAddFrobnitz(self):
        self.frobnitzes += 1
        return 'YES TO HAPPINESS'

class ExampleSubitem(Item):
    def __init__(self, store):
        Item.__init__(self, store)

class StoreMixin:
    pathToDB = None
    def setUp(self):
        if self.pathToDB is None:
            self.pathToDB = self.mktemp()
        self.store = Store(0, opj(self.pathToDB, "db"), opj(self.pathToDB, "files"))
        # THREE shall be the number of the count
        self.avatar = self.store.transact(ExampleAvatar, self.store, 3)

    def tearDown(self):
        self.store.close()
        try:
            del self.pathToDB
        except AttributeError:
            pass

class CachingTestCase(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = ExampleAvatar(self.store, 3)

    def testCreationCachesObjects(self):
        i = self.store.transact(Item, self.avatar)
        self.failUnless(self.store.cache.has(tuple(i.getItemPath())))

    def dbtestCachedBeforeTransactionCommit(self):
        i = Item(self.avatar)
        self.failUnless(self.store.cache.has(tuple(i.getItemPath())))

    def dbtestTransactUncallable(self):
        self.assertRaises(TypeError, self.store.transact, None, 1, 2, "3")

    def dbtestSubTransactionUncallable(self):
        def subItem():
            return Item(self.avatar)
        def subException():
            raise TypeError()
        def subTrans():
            i1 = self.avatar.transact(subItem)
            self.avatar.transact(subException)
        self.assertRaises(TypeError, self.avatar.transact, subTrans)

    def testAbortedTransactionClearsCache(self):
        def _():
            self.item = Item(self.avatar)
            self.itemPath = tuple(self.item.getItemPath())
            raise ExampleException("Failure")
        self.failUnlessRaises(ExampleException, self.store.transact, _)
        self.failIf(self.store.cache.has(self.itemPath))

    def testSynchronousReferences(self):
        def _():
            self.item = Item(self.avatar)
            self.otheritem = Item(self.avatar)
            self.weirdItem = Item(self.store)
            self.ref = self.item.referenceTo()
            self.otherref = self.otheritem.referenceTo()
            self.weirdref = self.weirdItem.referenceTo()
        self.store.transact(_)
        def _2():
            self.item.touch()
            self.item.ref = self.otheritem.referenceTo()
        self.store.transact(_2)
        del self.item
        del self.otheritem
        import gc
        gc.collect()
        def _3():
            self.assertIdentical(self.ref.getItem().ref.getItem(),
                                 self.otherref.getItem())
        self.store.transact(_3)
        def _4():
            i = self.ref.getItem()
            i.touch()
            i.brokeref = self.weirdItem.referenceTo()
        self.assertRaises(Inconsistency, self.store.transact, _4)
        def _5():
            i = self.ref.getItem()
            self.avatar.touch()
            self.avatar.goodref = i.referenceTo()
            self.avatar.topref = self.weirdItem.referenceTo()
        self.store.transact(_5)
        del self.weirdItem
        def _6():
            i = self.ref.getItem()
            w = self.weirdref.getItem()
            self.assertIdentical(self.avatar.goodref.getItem(), i)
            self.assertIdentical(self.avatar.topref.getItem(), w)
        self.store.transact(_6)

class UIExampleAccess(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = ExampleAvatar(self.store, 3)

    def testSomethingInteresting(self):
        self.assertRaises(ExampleException, self.avatar.transact, self.avatar.addFrobnitzThenBreak)
        self.assertEquals(self.avatar.frobnitzes, 3) # and the number of the counting shall be THREE
        self.avatar.transact(self.avatar.noReallyAddFrobnitz)
        self.assertEquals(self.avatar.frobnitzes, 4)

    def testSubitemStorage(self):
        subitem = self.avatar.transact(ExampleSubitem, self.avatar)
        self.assertIdentical(self.avatar.getItemByID(subitem.storeID),
                             subitem)
        try:
            nosub = self.store.getItemByID(subitem.storeID)
        except KeyError:
            return
        else:
            # trapping the exception to avoid potential re-numbering problems
            # and potential number-conflicts at multiple namespace levels, but
            # if we got something back from the storage there, it had better
            # not be the same item we put into sub-storage!
            self.assertNotEquals(nosub, subitem)


class PoolTestCase(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = ExampleAvatar(self.store, 3)
        self.p = Pool(self.store)
        self.items = []

    nAdds = 300
    nRuns = 5
    nItems = nAdds * nRuns

    def queryIndexAddItems(self):
        # Put some stuff into the pool
        for i in range(self.nAdds):
            i = Item(self.store)
            self.p.addItem(i)
            self.items.append(i)

    def queryIndexAssertState(self):
        query = list(self.p.queryIndex('uid', result=ITEM))
        self.assertEquals(len(query), self.nItems)
        for (orig, fromPool) in zip(self.items, query):
            self.assertIdentical(orig, fromPool)

        query = list(self.p.queryIndex('uid', result=INDEX))
        self.assertEquals(len(query), self.nItems)

        for (ans, fromPool) in zip(range(self.nItems), query):
            self.assertEquals(ans, fromPool)

        query = list(self.p.queryIndex('uid', result=STORE_ID))
        self.assertEquals(len(query), self.nItems)
        for (orig, fromPool) in zip(self.items, query):
            self.assertEquals(orig.storeID, fromPool)

        query = list(self.p.queryIndex('uid', limit=10))
        self.assertEquals(len(query), 10)
        itmz = self.items[:10]
        for (orig, fromPool) in zip(itmz, query):
            self.assertEquals(orig, fromPool)
        query = list(self.p.queryIndex('uid', backwards=True))
        itmz = self.items[:]
        itmz.reverse()

        self.assertEquals(len(query), len(itmz))
        for (orig, fromPool) in zip(itmz, query):
            self.assertEquals(orig, fromPool)

        sk = self.nAdds // 5
        ek = self.nAdds // 2
        query = list(self.p.queryIndex('uid', startKey=sk, endKey=ek, result=INDEX))
        self.assertEquals(query[0],  sk)
        self.assertEquals(query[-1], ek)
        quitter = self.p.queryIndex('uid', startKey=sk, endKey=ek, result=INDEX, backwards=True)
        self.assertRaises(Inconsistency, list, quitter)
        query = list(self.p.queryIndex('uid', startKey=ek, endKey=sk, result=INDEX, backwards=True))
        self.assertEquals(query[0],   ek)
        self.assertEquals(query[-1],  sk)

    multitxnQueryIndex = [queryIndexAddItems] * nRuns + [queryIndexAssertState]

    def testStartKey(self):
        nadds = 300
        startKey = 100
        items = []
        def _1():
            for i in range(nadds):
                i = self.store.transact(Item, self.store)
                self.p.addItem(i)
                items.append(i)

        def _2():
            query = list(self.p.queryIndex('uid', startKey=startKey, result=INDEX))
            self.assertEquals(len(query), nadds - startKey)
            self.assertEquals(query, range(100, 300))

        self.store.transact(_1)
        self.store.transact(_2)

    def testEndKey(self):
        nadds = 300
        endKey = 200
        items = []
        def _1():
            for i in range(nadds):
                i = self.store.transact(Item, self.store)
                self.p.addItem(i)
                items.append(i)

        def _2():
            query = list(self.p.queryIndex('uid', endKey=endKey, result=INDEX))
            self.assertEquals(len(query), endKey + 1)
            self.assertEquals(query, range(endKey + 1))

        self.store.transact(_1)
        self.store.transact(_2)

    def testBothKeys(self):
        nadds = 300
        startKey = 75
        endKey = 225
        items = []
        def _1():
            for i in range(nadds):
                i = self.store.transact(Item, self.store)
                self.p.addItem(i)
                items.append(i)

        def _2():
            items
            query = list(self.p.queryIndex('uid', startKey=startKey, endKey=endKey, result=INDEX))
            self.assertEquals(len(query), endKey - startKey + 1)
            self.assertEquals(query, range(startKey, endKey + 1))

        self.store.transact(_1)
        self.store.transact(_2)

    def dbtestIndexMutation(self):
        i = Item(self.store, name='Fred')
        self.p.addIndex(NameIndex())
        p2 = Pool(self.store)
        p2.addItem(i)
        self.p.addItem(i)
        self.assertEquals(list(self.p.queryIndex('name', result=INDEX)), ['Fred'])
        i.setIndexKey("name", 'Bob')
        self.assertEquals(list(self.p.queryIndex('name', result=INDEX)), ['Bob'])
        i.setIndexKey("name", None)
        self.assertEquals(list(self.p.queryIndex('name', result=INDEX)), [])
        i2 = Item(self.store, name='JJ')
        i2.setIndexKey('name', 'KK')
        self.p.addItem(i2)
        self.assertEquals(list(self.p.queryIndex('name', result=INDEX)), ['KK'])

    def testMultipleIndexes(self):
        p = self.p
        def _():
            i = Item(self.store, name='Fred')
            i.setIndexKey("received", time.time())
            p.addIndex(NameIndex())
            p.addIndex(OtherNameIndex())
            p.addIndex(DateIndex())
            p.addItem(i)
            return i

        def _2(i):
            self.assertEquals(list(p.queryIndex('name')), [i])
            self.assertEquals(list(p.queryIndex('received')), [i])
            self.assertEquals(list(p.queryIndex('name_other')), [])
        i = self.store.transact(_)
        self.store.transact(_2, i)

    def dbtestItemsMovement(self):
        i = Item(self.store, "mover")
        self.p.addItem(i)
        self.assertEquals(list(self.p.queryIndex()), [i])
        self.failUnless(i in self.p)
        self.p.removeItem(i)
        self.assertEquals(list(self.p.queryIndex()), [])
        self.failUnless(i not in self.p)
        self.p.addItem(i)
        self.assertEquals(list(self.p.queryIndex()), [i])

import gc
class ResurrectionTest(unittest.TestCase):
    def runResurrectionWork(self, storeID):
        p = self.mktemp()
        s = Store(storeID, opj(p, 'db'), opj(p, 'files'))
        a = s.transact(SubStore, s)
        sid = a.storeID
        i = s.transact(Item, a, name="Foo Foo")
        iid = i.storeID

        s.close()
        del s
        gc.collect()

        s = Store(storeID, opj(p, 'db'), opj(p, 'files'))
        a = s.getItemByID(sid)
        self.failUnless(isinstance(a, SubStore))
        i = a.getItemByID(iid)
        self.failUnless(isinstance(i, Item))
        self.assertEquals(i.name, "Foo Foo")

    def testMasterResurrection(self):
        self.runResurrectionWork(0)

    def testNonMasterResurrection(self):
        self.runResurrectionWork(11)

class ItemForwardingComponentTestCase(unittest.TestCase):
    def testEquality(self):
        C = fwd.ItemForwardingComponent
        self.assertEquals(C("fwd"), C("fwd"))
        self.failIfEqual(C("fwd"), "fwd")
        self.failIfEqual(C("fwd"), C("bar"))

    def dbtestInstantiationRestriction(self):
        self.assertRaises(RuntimeError,
                          fwd.ItemForwardingComponent,
                          Item(self.store))

import weakref

class Empty:
    def __finalizer__(self):
        return lambda: None

class Finalizable:
    L = []

    def __init__(self, L=None):
        if L is None:
            L = []
        self.L = L

    def __finalizer__(self):
        return lambda L=self.L, i=id(self): L.append(i)

def assert_(L, c, p):
    r = []
    if c.has(p):
        r.append(c.get(p))
    L.append(r)

class ViciouslyDefectiveFinalizable:
    def __init__(self, cache, path, L):
        self.cache = cache
        self.path = path
        self.L = L

    def __finalizer__(self):
        return lambda L=self.L, c=self.cache, p=self.path: assert_(L, c, p)

class ExceptionRaisingFinalizable:
    def __finalizer__(self):
        return lambda: 1 / 0

class FinalizingCacheTestCase(trialtest.TestCase):
    def testNoStrongReferences(self):
        o = Empty()
        r = weakref.ref(o)
        c = FinalizingCache()
        c.cache("path", o)
        del o
        self.assertEquals(r(), None)
        self.failIf(c.has("path"))
        self.assertRaises(KeyError, c.get, "path")

    def testFinalization(self):
        L = []
        o = Finalizable(L)
        i = id(o)
        c = FinalizingCache()
        c.cache("path", o)
        del o
        self.assertEquals(L, [i])
        self.failIf(c.has("path"))
        self.assertRaises(KeyError, c.get, "path")

    def testGarbageCollectorInteraction(self):
        L = []
        o1 = Finalizable(L)
        i1 = id(o1)
        o2 = Finalizable(L)
        i2 = id(o2)

        o1.ref = o2
        o2.ref = o1

        c = FinalizingCache()
        c.cache("path1", o1)
        c.cache("path2", o2)

        del o1, o2
        import gc; gc.collect()

        L.sort()
        L2 = [i1, i2]
        L2.sort()
        self.assertEquals(L, L2)
        self.failIf(c.has("path1"))
        self.failIf(c.has("path2"))
        self.assertRaises(KeyError, c.get, "path1")
        self.assertRaises(KeyError, c.get, "path2")

    def testGarbageFinalization(self):
        L = []
        c = FinalizingCache()
        o1 = ViciouslyDefectiveFinalizable(c, "path2", L)
        i1 = id(o1)
        o2 = ViciouslyDefectiveFinalizable(c, "path1", L)
        i2 = id(o2)

        o1.ref = o2
        o2.ref = o1

        c.cache("path1", o1)
        c.cache("path2", o2)

        import gc
        # gc.collect()
        # print len(gc.get_objects())
        del o1, o2
        gc.collect()
        # print len(gc.get_objects())

        self.failIf([o for o in gc.garbage if getattr(o, '__class__') is ViciouslyDefectiveFinalizable])

        for e in L:
            self.failUnless(e)

    def testFinalizingException(self):
        c = FinalizingCache()
        o = ExceptionRaisingFinalizable()
        c.cache("path", o)
        del o
        self.failIf(c.has("path"))
        self.assertRaises(KeyError, c.get, "path")
        log.flushErrors(ZeroDivisionError)
