# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import types
from os.path import join as opj

from twisted.trial import unittest
from twisted.python import reflect
from twisted.python import log
from twisted.python import components

from atop import sched
from atop import store
from atop import credup
from atop import powerup

def remove(sequence, omit):
    r = []
    for e in sequence:
        try:
            r.append(remove(e, omit))
        except:
            if e not in omit:
                r.append(e)
    return tuple(r)

_unittestSpecials = ['todo', 'skip']

class MetaTestClass(type):
    def makeTransactor(klass, f, name):
        def tf(self):
            try:
                return self.store.transact(f, self)
            except:
                from twisted.python import log
                log.err()
        tf = types.FunctionType(tf.func_code, f.func_globals, name, f.func_defaults, tf.func_closure)
        for attr in _unittestSpecials:
            hasattr(f, attr) and setattr(tf, attr, getattr(f, attr))
        return tf
    makeTransactor = classmethod(makeTransactor)

    def makeMultiTransactor(klass, fns, name):
        def tf(self):
            for f in fns:
                self.store.transact(f, self)
        f = fns[0]
        return types.FunctionType(tf.func_code, f.func_globals, name, f.func_defaults, tf.func_closure)
    makeMultiTransactor = classmethod(makeMultiTransactor)

    def __new__(klass, name, bases, attrs):
        for (key, value) in attrs.items():
            if key.startswith('dbtest') and isinstance(value, types.FunctionType):
                del attrs[key]
                attrs[key[2:]] = klass.makeTransactor(value, key[2:])
            elif key.startswith('multitxn') and isinstance(value, (list, tuple)):
                del attrs[key]
                attrs['test' + key[8:]] = klass.makeMultiTransactor(value, 'test' + key[8:])
        if name != 'TestCase':
            return types.ClassType.__new__(types.ClassType, name, remove(bases, [object, TestCase]) + (_TestCase,), attrs)
        return type.__new__(klass, name, bases, attrs)

class TestCase(object):
    """Base class for test cases which rely heavily on the database.

    Methods beginning with the prefix "dbtest" will be run as normal
    tests, but in a database transaction.

    Tuples and lists beginning with "multitxn" will have the functions they
    contain run as a unit test, each function being run in a separate transaction.

    @cvar dbPath: The path relative to L{storePath} beneath which the database
    environment is created.

    @cvar mailPath: The path relative to L{storePath} beneath which message files
    are stored.

    @cvar storePowerups: A tuple of the powerups to be installed when the store is
    created.

    @ivar storePath: The path beneath the database was created.

    @ivar store: A database instance unique to the running test method.
    """

    __metaclass__ = MetaTestClass

class _TestCase(unittest.TestCase):
    dbPath = "db"
    mailPath = "mail"

    schedService = None

    storePowerups = ()

    AvatarFactory = credup.Avatar

    closeDbOnTearDown = True

    def setUp(self):
        self.storePath = self.mktemp()
        self.store = store.Store(0, opj(self.storePath, self.dbPath), opj(self.storePath, self.mailPath))
        store.StorageService(self.store)
        self.store.transact(self.installPowerups, self.store, self.collectPowerups('storePowerups'))
        self.store.transact(self.dbSetUp)

        try:
            sched.IScheduler(self.store)
        except components.CannotAdapt:
            pass
        else:
            self.schedService = sched.SchedulerService(self.store)
            self.schedService.startService()

    def tearDown(self):
        if self.schedService:
            self.schedService.stopService()
        self.store.transact(self.dbTearDown)
        if self.closeDbOnTearDown:
            self.store.close()

    def dbSetUp(self):
        pass

    def dbTearDown(self):
        pass

    def installPowerups(self, store, powerups):
        ps = powerup.IPowerStation(store)
        for p in powerups:
            ps.installPowerup(p)

    def collectPowerups(self, name):
        d = dict.fromkeys(getattr(self, name, ()))
        for k in reflect.allYourBase(self.__class__):
            d.update(dict.fromkeys(getattr(k, name, ())))
        return d.keys()

    def createAvatar(self, powerups, *a, **kw):
        def _():
            avatar = self.AvatarFactory(self.store, *a, **kw)
            ps = powerup.IPowerStation(avatar)
            for p in powerups:
                ps.installPowerup(p)
            return avatar
        return self.store.transact(_)

deferredResult = unittest.deferredResult
deferredError = unittest.deferredError
