# -*- test-case-name: atop.test.test_tpython -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.
# objects that should be in tpython

import time
import sys

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.protocol import Protocol
from twisted.python import reflect, context, log

from twisted.persisted.styles import Versioned

def mkgmtime(gmtt):
    'convert GMT time-tuple to local time'
    if time.daylight and gmtt[-1]:
        zone = time.altzone
    else:
        zone = time.timezone
    return time.mktime(gmtt) - zone

def gmtForYearMonthDay(year, month, day):
    return time.gmtime(                          # 3 to a gmt tuple...
        time.mktime(                             # 2 to a localtime float...
            (year, month, day, 0, 0, 0, 0, 0, 0) # 1 go from a localtime tuple...
        )
    )

class Forgetter:
    """Utility class for forgetting particular attributes when persisted.

    @cvar persistenceForgets: A list of strings, naming attributes which should
    be deleted when I am persisted.

    @cvar contextRemembers: A list of 2-tuples of strings, (contextName,
    attributeName), that name attributes which should be retrieved from
    twisted.python.context upon loading me.
    """

    def __getstate__(self, state=None):
        if state is None:
            state = self.__dict__.copy()
        if isinstance(self, Versioned):
            d = Versioned.__getstate__(self, state)
        else:
            d = state
        l = []
        reflect.accumulateClassList(self.__class__, 'persistenceForgets', l)
        for ig in l:
            if d.has_key(ig):
                del d[ig]
        return d

    def __setstate__(self, state):
        if isinstance(self, Versioned):
            Versioned.__setstate__(self, state)
        else:
            self.__dict__ = state
        l = []
        reflect.accumulateClassList(self.__class__, 'contextRemembers', l)
        for contextName, attributeName in l:
            setattr(self, attributeName, context.get(contextName))


class DebugShimTransport(Protocol):
    def __init__(self, p, f=sys.stdout):
        self.p = p
        self.f = f

    # Protocol
    def makeConnection(self, t):
        self.t = t
        self.p.makeConnection(self)
        self.f.write('*\n')

    def dataReceived(self, data):
        self.f.write( 'C: %r\n' % data )
        self.p.dataReceived(data)

    def connectionLost(self, reason):
        self.f.write( 'C: *\n' )
        self.p.connectionLost(reason)

    # Transport
    
    def write(self, data):
        self.f.write( 'S: %r\n' % data )
        self.t.write(data)

    def writeSequence(self, iovec):
        self.f.write( 'S: %s\n' % repr(iovec))
        self.t.writeSequence(iovec)

    def getPeer(self):
        return self.t.getPeer()

    disconnecting = False

    def loseConnection(self):
        self.f.write( 'S: *\n' )
        self.t.loseConnection()
        self.disconnecting = True

    def registerProducer(self, producer, streaming):
        self.t.registerProducer(producer, streaming)

    def unregisterProducer(self):
        self.t.unregisterProducer()

def iterateInReactor(i, delay=0.0001, holder=[None]):
    d = defer.Deferred()
    def callLaterer():
        try:
            maybe = i.next()
            if isinstance(maybe, defer.Deferred):
                def rslts(x):
                    #print 'rslts', x
                    holder[0] = reactor.callLater(delay, callLaterer)
                    return x
                maybe.addBoth(rslts)
                return
        except StopIteration:
            d.callback(i)
        except:
            d.errback()
        else:
            holder[0] = reactor.callLater(delay, callLaterer)
    callLaterer()
    return d


PERIODS = (
    ('week', 60 * 60 * 24 * 7),
    ('day', 60 * 60 * 24),
    ('hour', 60 * 60),
    ('minute', 60),
    ('second', 1),
)

def humanInterval(seconds):
    s = []
    for (p, d) in PERIODS:
        n = int(seconds / d)
        seconds %= d
        if n:
            s.append('%d %s%s' % (n, p, n > 1 and 's' or ''))
    if s:
        return ', '.join(s)
    return '0 seconds'


class _counter:
    persec = 0
    def accum(self):
        self.persec += 1
        reactor.callLater(0, self.accum)
    def display(self):
        log.msg("FPS: %d" % self.persec)
        self.persec = 0
        reactor.callLater(1, self.display)
    
def framecounter():
    c = _counter()
    c.display()
    c.accum()
