# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import warnings

from twisted.python import components
from twisted.python.reflect import qual

from nevow.iwoven import IData
from nevow.stan import specialMatches, specials, Tag
from nevow.tags import invisible

class _Marker(object):
    pass


def megaGetInterfaces(adapter):
    adrs = [qual(x) for x in components.getInterfaces(adapter)]
    ## temporarily turn this off till we need it
    if False: #hasattr(adapter, '_adapterCache'):
        adrs.extend(adapter._adapterCache.keys())
    return adrs


dataqual = qual(IData)


class NodeNotFound(KeyError):
    def __str__(self):
        return "The %s named %r wasn't found in the template." % tuple(self.args[:2])

class TooManyNodes(Exception):
    def __str__(self):
        return "More than one %r with the name %r was found." % tuple(self.args[:2])


def stripContexts(tag):
    for i in range(len(tag.children)):
        obj = tag.children[i]
        if isinstance(obj, WovenContext):
            tag.children[i] = obj.tag
            obj = obj.tag
        if isinstance(obj, Tag):
            stripContexts(obj)


class WovenContext(object):
    cloned = False
    key = None
    _remembrances = {}
    tag = None
    def __init__(self, parent=None, tag=None, precompile=False, remembrances=None, key=None):
        self.tag = tag
        self.parent = parent
        if key is not None:
            if self.parent is not None and self.parent.key is not None:
                self.key = '.'.join((self.parent.key, key))
            else:
                self.key = key
            #print "KEY", self.key
        else:
            ## Bubble the value down to the bottom so it's always immediately accessible
            if self.parent is not None:
                self.key = self.parent.key
        if remembrances is None:
            self._remembrances = {}
        else:
            self._remembrances = remembrances
        self.precompile = precompile

    def __repr__(self):
        rstr = ''
        if self._remembrances:
            rstr = ', remembrances=%r' % self._remembrances
        return "Context(tag=%r%s)" % (self.tag, rstr)

    def remember(self, adapter, interface=None):
        """Remember an object that implements some interfaces.
        Later, calls to .locate which are passed an interface implemented
        by this object will return this object.
        
        If the 'interface' argument is supplied, this object will only
        be remembered for this interface, and not any of
        the other interfaces it implements.
        """
        if interface is None:
            interfaceList = megaGetInterfaces(adapter)
            if not interfaceList:
                interfaceList = [dataqual]
        else:
            interfaceList = [qual(interface)]
        for interface in interfaceList:
            self._remembrances[interface] = adapter
        return self

    def with(self, tag):
        """Remember a few things in a new context stack entry.
        
        tag: the tag the new context stack entry will be associated with.
        data: The data to pass to any rendering functions.
        renderer: The rendering function to use to render this node.
        observer: The observer function to notify when an
            event occurs to this node.
        """
#        data=None, renderer=None, observer=None, remembrances=None

        new = WovenContext(self, tag, self.precompile, key=tag.key)
        if tag.data is not None:
            new.remember(tag.data, IData)
        if tag.remember is not None:
            new.remember(tag.remember)
#        if renderer is not None:
#            # push a renderer onto the stack
#            pass
# woah this seems broken
#            new.remember(renderer, IRendererFactory)
#        if observer is not None:
#            # push an observer onto the stack
#            pass
        return new

    def locate(self, interface, depth=1):
        """Locate an object which implements a given interface.
        Objects will be searched through the context stack top
        down.
        """
        key = qual(interface)
        if depth < 0:
            full = []
            while True:
                try:
                    full.append(self.locate(interface, len(full)+1))
                except KeyError:
                    break
            #print "full", full, depth
            if full:
                return full[depth]
            return None
        if self._remembrances.has_key(key):
            depth -= 1
            if not depth:
                return self._remembrances[key]
        if self.parent is None:
            raise KeyError, "Interface %s was not remembered." % key
        return self.parent.locate(interface, depth)

    def chain(self, context):
        """Go to the top of this context's context chain, and make
        the given context the parent, thus continuing the chain
        into the given context's chain.
        """
        top = self
        while top.parent is not None:
            if top.parent.tag is None:
                ## If top.parent.tag is None, that means this context (top)
                ## has been cloned. We want to insert the current context
                ## (context) as the parent of this context (top) to chain properly.
                break
            top = top.parent
            if top is context: # this context is already in the chain; don't create a cycle
                return
        top.parent = context

    def patterns(self, pattern, raiseIfMissing=True):
        """Generate clones of pattern tags forever, looping around to the beginning
        when we run out of unique matches.
        """
        tag = self.tag.clone()
        stripContexts(tag)
        patterner = self._locatePatterns(tag, pattern, raiseIfMissing)
        return PatternTag(patterner)

    def slotted(self, slot):
        """Locate and return an existing slot in the current context tag, clearing any
        dummy content in the process. You can then fill the slot with real content.
        """
        return self._locateOne(slot, self._locateSlots, 'slot')

    def keyed(self, key):
        """Locate an existing keyed node in the current context tag. This returns the node
        unchanged, without clearing or cloning anything.
        """
        return self._locateOne(key, self._locateKeys, 'key')

    def _generatePatterns(self, pattern):
        warnings.warn("use patterns instead", stacklevel=2)
        return self.patterns(pattern)

    def _locatePatterns(self, tag, pattern, raiseIfMissing):
        keeplooking = True
        while keeplooking:
            keeplooking = False
            for x in specialMatches(tag, 'pattern', pattern):
                keeplooking = True
                cloned = x.clone()
                cloned.pattern = None
                yield cloned
        if raiseIfMissing:
            raise RuntimeError, "Pattern %s was not found." % pattern
        while True:
            yield invisible

    def _locateOne(self, name, locator, descr):
        found = False
        for node in locator(name):
            if found:
                raise TooManyNodes(descr, name)
            found = node
        if not found:
            raise NodeNotFound(descr, name)
        return found

    def _locateSlots(self, slot):
        for match in specialMatches(self.tag, 'slot', slot):
            yield match.clear()

    def _locateKeys(self, key):
        for keySpecial in specials(self.tag, 'key'):
            if keySpecial.key.endswith(key):
                yield keySpecial        

    def clone(self, includeTag=True):
        if self.parent is None:
            parent = None
        else:
            parent=self.parent.clone(includeTag=False)
        if includeTag:
            tag = self.tag.clone()
        else:
            tag = None
        return WovenContext(
            parent = parent,
            tag = tag,
            remembrances=self._remembrances.copy()
        )


class PatternTag(object):
    def __init__(self, patterner):
        self.patterner = patterner

    def __call__(self, **kw):
        return self.patterner.next()(**kw)

    def __getitem__(self, item):
        return self.patterner.next()[item]
