# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


from twisted.internet.defer import Deferred
from twisted.python import reflect
from twisted.python import components
from twisted.python.urlpath import URLPath
from twisted.web import resource
from twisted.web import static
from twisted.web import server

from nevow import formless
from nevow import iwoven
from nevow.renderer import flatten
from nevow.tags import *
from nevow.iwoven import IRendererFactory


class PrefixerDict(dict):
    def __init__(self, prefix, errors):
        self.prefix = prefix
        self.errors = errors
        dict.__init__(self)

    def __setitem__(self, key, value):
        self.errors['.'.join((self.prefix, key))] = value

    def __getitem__(self, key):
        return self.errors['.'.join((self.prefix, key))]

    def update(self, other):
        for key, value in other.items():
            self[key] = value



class IFormDefaults(components.Interface):
    """Default values for the current form
    """


class FormDefaults(components.Adapter):
    def __init__(self, original):
        """original should be a request"""
        self.original = original
        self.defaults = {}

    def setDefault(self, key, value, context=None):
        self.defaults[key] = value

    def getDefault(self, key, context=None):
        # 1) Check on the request
        current = self.defaults.get(key, None)
        if current is None:
            # 2) Check on the session
            if context is not None:
                sessionDefaults = IFormDefaults(context.locate(iwoven.IRequest).getSession())
                if sessionDefaults is not self:
                    current = sessionDefaults.getDefault(key)
                    if current is not None:
                        return current
                # 3) Ask the Binding instance for the default values
                try:
                    configurable = context.locate(formless.IConfigurable)
                except KeyError:
                    return ''
                return configurable.getDefault(context.locate(iwoven.IData))
        return current

    def getAllDefaults(self, key):
        return PrefixerDict(key, self.defaults)

    def clearAll(self):
        self.defaults = {}


class IFormErrors(components.Interface):
    """An object which keeps track of which forms have which errors
    """


class FormErrors(components.Adapter):
    """An object which keeps track of which forms have which errors
    """
    __implements__ = IFormErrors,
    def __init__(self, original):
        """original should be a session"""
        self.original = original
        self.errors = {}

    def setError(self, errorKey, error):
        self.errors[errorKey] = error

    def getError(self, errorKey):
        return self.errors.get(errorKey) 

    def getAllErrors(self, formName):
        return PrefixerDict(formName, self.errors)

    def updateErrors(self, formName, errors):
        PrefixerDict(formName, self.errors).update(errors)

    def clearErrors(self, formName):
        for key in self.errors.keys():
            if key.startswith(formName):
                del self.errors[key]

    def clearAll(self):
        self.errors = {}


class IBindingRenderer(components.Interface):
    """Something which can render a formless Binding.
    
    A Binding represents an atomic form which can be
    submitted to cause a named thing to occur;
    a MethodBinding will cause a method to be called;
    a PropertyBinding will cause a property to change.
    """


class IActionRenderer(components.Interface):
    """An alternate rendering of a formless Binding which
    is usually represented as a smaller, no-input-fields
    toolbar-style button. Should call a MethodBinding.
    
    An action is distinct from a MethodBinding in that
    an action does not explicitly solicit input from the user,
    but instead gathers all the information it needs to run
    a method from implicit context, such as the render
    context, or the current state of the selection.
    """


class ITypedRenderer(components.Interface):
    """Something which can render a formless Typed.
    Renders input fields in html which will gather information
    from the user which will be passed to the Typed.coerce
    method when the entire form is submitted.
    """


class IRedirectAfterPost(components.Interface):
    """A marker interface which can be set to something which can be cast
    to a string to indicate that the browser should be redirected to the
    resultant url after posting the form successfully. This component can
    be set by any form post validators, or by the configurable method which
    is being automatically called if it has access to the request.
    
    Set this using the following:
    
    request.setComponent(IRedirectAfterPost, "http://example.com/")
    """


def defaultBindingRenderer(context, data):
    return invisible(renderer=StringRenderer(data))


def calculateFullBindingName(context, data):
    configurable = context.locate(formless.IConfigurable, depth=-1)
    if configurable is None:
        postLocation = None
    else:
        postLocation = configurable.postLocation
    if postLocation is None:
        postLocation = context.locate(iwoven.IRequest).prepath[-1]
    return postLocation, context.key


def calculatePostURL(context, data, postLocation=None, fullBindingName=None):
    if postLocation is None or fullBindingName is None:
        postLocation, fullBindingName = calculateFullBindingName(context, data)
    if postLocation == '':
        return 'freeform_post:%s' % fullBindingName
    return "%s/freeform_post:%s" % (postLocation, fullBindingName)


class BaseInputRenderer(components.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        request = context.locate(iwoven.IRequest)

        try:
            withObject = context.locate(iwoven.IData, 2)
        except KeyError:
            print "withObject is None"
            withObject = None

        defaults = IFormDefaults(request)
        value = defaults.getDefault(context.key, context)
        try:
            _, fbn = calculateFullBindingName(context, context.locate(formless.IBinding))
        except KeyError:
            fbn = ''
        errors = IFormErrors(request.getSession())
        errormessage = errors.getError(fbn)
        if errormessage is None:
            errormessage = ''

        slot = span(_class="freeform-input")
        if not value and request.args.has_key(data.name):
            value = request.args[data.name][0]
        theslotfilled = self.input(context, slot, data, data.name, value)
        if theslotfilled is None:
            ## Breakout; we don't have any options to give the user, so don't render anything
            return ''

        return [
            label(_class="freeform-label")[ data.label ],
            theslotfilled, # freeform-input
            span(_class="freeform-error")[ errormessage ],
            div(_class="freeform-description")[ data.description ],
        ]

    def input(self, context, slot, data, name, value):
        raise NotImplementedError, "Implement in subclass"


class PasswordRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return [
            input(name=name, type="password"), " Again ", input(name="%s____2" % name, type="password"),
        ]


class StringRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return slot[
            input(type="text", name=name, value=value)
        ]


class TextRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return slot[
            textarea(name=name, _class="freeform-textarea")[
                value
            ]
        ]


class BooleanRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        ## The only difference here is the "checked" attribute; the value is still the same because
        ## we want true to be passed to the server when the checkbox is checked and the form
        ## is posted.
        if value:
            slot[input(type="checkbox", checked="checked", name=name, value='True')]
        else:
            slot[input(type="checkbox", name=name, value='True')]

        # HTML forms are so weak. If the checkbox is not checked, no value at all will be
        # in request.args with the name data.name. So let's force the value False to always
        # be in request.args[data.name]. If the checkbox is checked, the value True will
        # be first, and we will find that.
        slot[input(type="hidden", name=name, value="False")]
        return slot

class FileUploadRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        return slot[input(type="file", name=name)]


class ChoiceRenderer(BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        tv = data.typedValue
        if tv.choicesAttribute:
            choices = getattr(context.locate(formless.IConfigurable), tv.choicesAttribute)
        else:
            choices = tv.choices

        numChoices = len(choices)
        if numChoices in [0, 1]:
            return None

        selecter = select(name=name)
        stringify = tv.stringify

        for index, val in enumerate(choices):
            if val == value:
                selecter[option(value=str(index), selected="selected")[stringify(val)]]
            else:
                selecter[option(value=str(index))[stringify(val)]]
        return slot[selecter]


def isGrouped(context, depth):
    """Return True if the IBinding at the given depth is a Group binding.
    """
    try:
        bnd = context.locate(formless.IBinding, depth=depth)
    except KeyError:
        return False
    if isinstance(bnd, formless.GroupBinding):
        return True
    return False


class ButtonRenderer(components.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        if isGrouped(context, 3):
            return ''
        return input(type="submit")


class ObjectRenderer(components.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        configurable = context.locate(formless.IConfigurable)
        return div(data=getattr(configurable, data.name))[ configure ]


class RequestRenderer(components.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    def __call__(self, context, data):
        return ''


## Hoorj javascript
LOCATE_AND_CHANGE_CURRENT_SELECTION_INPUT = xml(\
"var sel = document.getElementById('freeform-action-selection:' + this.name);"\
"if (this.checked) { "\
"sel.value += this.value+'-';"\
"} else {"\
" split = sel.value.split('-'); "\
" sel.value = '';"\
" for(var i = 0; i<split.length; i++) {"\
"   if (split[i] && split[i] != this.value) { sel.value +=split[i]+'-' }"\
" }"\
"};")


class ListRenderer(components.Adapter):
    __implements__ = IRendererFactory, ITypedRenderer
    complexType = True
    styleClass = 'freeform-list'

    def items(self, context, data):
        request = context.locate(iwoven.IRequest)
        defaults = IFormDefaults(request)
        itemList = list(defaults.getDefault(context.key, context))
        configurableList = [
            locateIfPossible(context, item) for item in itemList
        ]
        postLocations = [None] * len(configurableList)
        for i, config in enumerate(configurableList):
            if True: #config.postLocation is None:
                config.postLocation = str(i)
            postLocations[i] = config.postLocation
        return zip(postLocations, itemList, configurableList)

    def __call__(self, context, data):
        errors = IFormErrors(context.locate(iwoven.IRequest).getSession())
        if isinstance(data, formless.Property):
            ## List properties are implemented
            actions = data.typedValue.actions or []

            if actions:
                postLocation, bindingName = calculateFullBindingName(context, data)
                #topConf = context.locate(formless.IConfigurable)
                actionForm = form(
                    _class="freeform-action-group", 
                    action="%s/freeform-action-post:%s" % (postLocation, context.key)
                )[
                    input(type="hidden", id="freeform-action-selection:%s" % data.name, name="freeform-action-selection", value=""),
                    [invisible(_class="freeform-action", data=act, renderer=IActionRenderer(act, persist=False)) for act in actions]
                ]
            else:
                actionForm = ''

            items = self.items(context, data)
            if not items:
                content = div(_class=self.styleClass) [ "An empty list." ]
            else:
                listItems = [None] * len(items)

                for i, (key, value, conf) in enumerate(items):
                    binding = formless.ElementBinding(str(i), formless.Object())
                    d = div(
                        _class="freeform-list-item",
                        key=key,
                        remember=binding
                    )
                    listItems[i] = d
                    confFound = not isinstance(conf, NotFoundConfigurable)
                    ## remove this; this is for the pool browser being embedded in freeform.
                    ## really, you want items embedded in freeform to automatically have
                    ## checkboxes.
                    if confFound and actions:
                        ## data.name makes it possible for the script to calculate 'freeform-action-selection:%s' % data.name
                        ## which is the id of the node which will keep track of which of these checkboxes are selected.
                        d[
                            input(
                                _class="freeform-list-item-checkbox",
                                type="checkbox",
                                name = data.name,
                                value = key,
                                onclick=LOCATE_AND_CHANGE_CURRENT_SELECTION_INPUT)]
                    ctx = context.with(d)
                    ctx.remember(binding, formless.IBinding)
                    ctx.remember(conf, formless.IConfigurable)
                    fake = formless.ElementBinding('', formless.Object())
                    _, fbn = calculateFullBindingName(ctx, fake)
                    #for errkey in errors.errors:
                    #    if errkey.startswith(fbn):
                    #        visibility = 'block'
                    #        break
                    #else:
                    #    visibility = 'none'
                    visibility = 'block'
                    if confFound:
                        d[
                            span(_class="freeform-form-label") [
                                ## TODO: Remove getattr by making sure every IConfigurable has a summary method
                                ## TODO: remove this by using ISummary instead
                                getattr(value, 'summary', lambda: str(value))()
                            ],
                            div(
                                _class="binding-content",
                                style="display: %s;" % visibility,
                                data=value
                            )[
                                configure
                            ]                
                        ]
                    else:
                        ## No configurable, let's just throw whatever it was in the dom
                        if iwoven.ISerializable(value, None):
                            try:
                                for x in range(100):
                                    theData = context.locate(iwoven.IData, x+1)
                                    if not isinstance(theData, formless.Binding):
                                        break
                            except KeyError:
                                print "keyeerror"
                            d[
                                invisible(data=theData)[ value ]
                            ]
                        else:
                            d[
                                "An unconfigurable, unserializable object: %r" % value
                            ]
    
                content = div(_class=self.styleClass) [ actionForm, data.typedValue.header, listItems, data.typedValue.footer ]
            return content
        else:
            ## List arguments aren't implemented yet
            return 'List argument input forms are not implemented yet.'


class DictionaryRenderer(ListRenderer):
    complexType = True
    styleClass = 'freeform-dictionary'

    __implements__ = IRendererFactory,
    def items(self, context, data):
        request = context.locate(iwoven.IRequest)
        defaults = IFormDefaults(request)
        theDict = defaults.getDefault(context.key, context)
        items = theDict.items()
        return [
            (x[0], "%s: %s" % (x[0], x[1]), locateIfPossible(context, x)) for x in items
        ]


def getError(context):
    errors = IFormErrors(context.locate(iwoven.IRequest).getSession())
    err = errors.getError(context.key)
    if err is not None:
        return div(_class="freeform-error")[ err ]
    return ''


class GroupBindingRenderer(components.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, data):
        context.remember(data, formless.IBinding)

        configurable = GroupConfigurable(data.boundTo, data.typedValue.interface)
        context.remember(configurable, formless.IConfigurable)

        bindingNames = configurable.getBindingNames(context)
        bindings = []

        for name in bindingNames:
            bnd = configurable.getBinding(context, name)
            bindings.append(
                invisible(
                    data = bnd,
                    renderer = IBindingRenderer(bnd, defaultBindingRenderer, persist=False),
                    key=name
                )
            )

        return getError(context), form(_class="freeform-group-binding", action=calculatePostURL(context, data))[
            bindings,
            input(type="submit")
        ]


class PropertyBindingRenderer(components.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, data):
        context.remember(data, formless.IBinding)
        typedRenderer = ITypedRenderer(data.typedValue, defaultBindingRenderer, persist=False)
        if getattr(typedRenderer, 'complexType', False):
            return invisible(data=data, renderer=typedRenderer)

        flat = flatten(iwoven.ISerializable(invisible(data=data, renderer=typedRenderer, key=data.name)).serialize(context, None))
        if not flat:
            return ''
        if isGrouped(context, 2):
            frm = div(_class="freeform-grouped-property-binding")
            subm = ''
        else:
            ## If we're not inside of a GroupBinding, we need a form and a submit button
            frm = form(action=calculatePostURL(context, data), _class="freeform-property-binding")
            subm = input(type="submit", name="change", value="Change")

        return getError(context), frm[
            flat,
            subm
        ]


class MethodBindingRenderer(components.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, data):
        if data.getAttribute('invisible'):
            return ''

        context.remember(data, formless.IBinding)
        args = data.getArgs()

        request = context.locate(iwoven.IRequest)

        postLocation, fullBindingName = calculateFullBindingName(context, data)
        errors = IFormErrors(request.getSession())
        if True: #data.getAttribute('revealed') or errors.getAllErrors(fullBindingName):
            visibility = 'visible'
        else:
            if len(args) <= 7:
                visibility = 'visible'
            else:
                visibility = 'none'

        if isGrouped(context, 2):
            frm = div(_class="freeform-grouped-method-binding")
        else:
            ## If we're not inside of a GroupBinding, we need a form
            frm = form(
                method="POST", 
                enctype="multipart/form-data", 
                action=calculatePostURL(context, data, postLocation, fullBindingName), 
                _class="freeform-method-binding")

        return getError(context), frm [
            div(_class="freeform-form-label") [
                data.label,
            ],
            fieldset(_class="binding-content", style="display: %s;" % visibility)[ ## This is the hidden form which can be revealed by clicking on the method label
                data.description, 
                [
                    div(key=argument.name, data = argument, renderer = ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist = False))
                    for argument in args
                ],
            ],
        ]


class ArgumentBindingRenderer(components.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, argument):
        context.remember(data, formless.IBinding)
        ## if need be add an extra div here with class freeform-argument-binding
        return div(_class="freeform-argument-binding")[
            invisible(renderer=ITypedRenderer(argument.typedValue, defaultBindingRenderer, persist=False))
        ]


class ActionRenderer(components.Adapter):
    __implements__ = IRendererFactory,

    def __call__(self, context, action):
        return input(type="submit", name="freeform-action", value=action.name)


def clearErrors(context, data):
    errs= IFormErrors(context.locate(iwoven.IRequest).getSession())
    #errs.errors = {}
    return ''


def createConfigurer(message):
    def locateConfigurable(context, data):
        if components.implements(data, formless.IConfigurable):
            return data
        configurable = None
        ## Only try the resource itself if we haven't already been configuring it.
        ## remove these "check the IResource for IConfigurable" hacks if possible
        try:
            existing = context.locate(formless.IConfigurable)
            if getattr(existing, 'original', None) is data:
                return existing
        except KeyError:
            ## First try the resource itself (hacky)
            res = context.locate(resource.IResource)
            configurable = formless.IConfigurable(res, None)
            if configurable is None:
                if components.implements(res, formless.TypedInterface):
                    configurable = TypedInterfaceConfigurable(res)
        if configurable is None:
            configurable = formless.IConfigurable(data, None)
            if configurable is None:
                if components.implements(data, formless.TypedInterface):
                    configurable = TypedInterfaceConfigurable(data)
                else:
                    return NotFoundConfigurable(message % {'data': data})

        return configurable

    def configure(context, data):
        configurable = locateConfigurable(context, data)
        context.remember(configurable, formless.IConfigurable)

        bindingNames = configurable.getBindingNames(context)
        bindings = []

        for name in bindingNames:
            bnd = configurable.getBinding(context, name)
            bindings.append(
                invisible(
                    data = bnd,
                    renderer = IBindingRenderer(bnd, defaultBindingRenderer, persist=False),
                    key = name
                )
            )

        return bindings
    return locateConfigurable, configure


locateConfigurable, configure = createConfigurer("An unconfigurable object (%(data)r)")
locateIfPossible, configureIfPossible = createConfigurer("")


class ImproperInputError(Exception):
    """Some malformed input was given to the form machinery;
    usually means a form was posted to a non-existant binding.
    """


class Configurable(object):
    __implements__ = formless.IConfigurable,

    bindingDict = None

    def __init__(self, original):
        self.original = original
        self.boundTo = self

    def getBindingNames(self, context):
        ## Todo: remove this getattr
        ifs = components.getInterfaces(getattr(self, 'boundTo', self))
        ifs = [
            x for x in ifs if x is not formless.IConfigurable and x is not formless.TypedInterface
        ]
        bindingNames = []
        self.bindingDict = bindingDict = {}
        for interface in ifs:
            ## TypedInterfaces have a __spec__ attribute which is a list of all Typed properties and
            ## autocallable methods
            for binding in getattr(interface, '__spec__', []):
                bindingDict[binding.name] = binding
                bindingNames.append(binding.name)
                if components.implements(binding.typedValue, formless.IActionableType):
                    acts = binding.typedValue.actions
                    if acts is None:
                        acts = []
                    for action in acts:
                        bindingDict[action.name] = action
        return bindingNames

    def getDefault(self, forBinding):
        name = forBinding.name
        if hasattr(self, name):
            return getattr(self, name)
        ## Todo: Only do this in ConfigurableAdapter instead of Configurable
        if hasattr(self.boundTo, name):
            return getattr(self.boundTo, name)
        if hasattr(self.original, name):
            return getattr(self.original, name)
        return forBinding.default

    def getBinding(self, context, name):
        if self.bindingDict is None:
            self.getBindingNames(context)
        if self.bindingDict is None:
            self.bindingDict = {}
        binding = getattr(self, 'bind_%s' % name, getattr(self.boundTo, 'bind_%s' % name, None))
        if binding is not None:
            binding = binding(context)
        else:
            try:
                binding = self.bindingDict[name]
            except KeyError:
                raise ImproperInputError, name
        binding.boundTo = self.boundTo
        return binding

    def postForm(self, ctx, bindingName, args):
        """Accept a form post to the given bindingName. The bindingName
        can be dotted to indicate an attribute of this Configurable, eg
        addresses.0.changeEmail. The post arguments are given in args.
        Return a Resource which will be rendered in response.
        """
        request = ctx.locate(iwoven.IRequest)
        pathSegs = bindingName.split('.')
        configurable = self

        for seg in pathSegs:
            binding = configurable.getBinding(ctx, seg)
            child = self.boundTo
            if not isinstance(binding, formless.GroupBinding):
                accessor = iwoven.IGettable(configurable.boundTo, None)
                if accessor is None:
                    child = getattr(configurable.boundTo, binding.name)
                else:
                    child = accessor.child(ctx, binding.name)
            ## If it's a groupbinding, we don't do anything at all for this path segment
            configurable = locateConfigurable(ctx, child)
            ctx = ctx.with(invisible(key=seg))
            ctx.remember(binding, formless.IBinding)
            ctx.remember(configurable, formless.IConfigurable)

        bindingValidator = IInputValidator(binding)
        try:
            result = bindingValidator.validate(ctx, binding.boundTo, args)
        except formless.ValidateError, e:
            sess = request.getSession()
            errors = IFormErrors(sess)
            ## Set the overall error for this form
            errors.setError(ctx.key, e.formErrorMessage)
            errors.updateErrors(bindingName, e.errors)
            IFormDefaults(sess).getAllDefaults(ctx.key).update(e.partialForm)
            result = None

        return WaitForPOSTResult(result, binding)

    def summary(self):
        return "An instance of %s" % self.__class__.__name__

    postLocation = None


class WaitForPOSTResult(resource.Resource):
    def __init__(self, result, binding):
        resource.Resource.__init__(self)
        self.result = result
        self.binding = binding

    def render(self, request):
        result = self.result; binding = self.binding
        if isinstance(self.result, Deferred):
            def _(x):
                self.result = x
                self.render(request)
                request.finish()
            self.result.addCallback(_)
            return server.NOT_DONE_YET
        else:
            redirectAfterPost = request.getComponent(IRedirectAfterPost, None)
            if redirectAfterPost is None:
                ref = request.getHeader('referer') or ''
                refpath = URLPath.fromString(ref)
                request.getSession().setComponent(iwoven.IHand, result)
                request.redirect(str(refpath))
            else:
                ## Use the redirectAfterPost url
                request.redirect(str(redirectAfterPost))

            return static.Data('You posted a form to %s' % binding, 'text/plain').render(request)


class NotFoundConfigurable(Configurable):
    def getBinding(self, context, name):
        raise ImproperInputError, self.original


class TypedInterfaceConfigurable(Configurable):
    def __init__(self, original):
        self.original = original
        self.boundTo = original

    def summary(self):
        return "An instance of %s" % self.original.__class__.__name__

    def __repr__(self):
        return "TypedInterfaceConfigurable(%r)" % self.original

    def _getPostLocation(self):
        if getattr(self.original, '_freeform_post_location', None):
            return self.original._freeform_post_location
        if hasattr(self.original, 'postLocation'):
            return self.original.postLocation
        return getattr(self.original, 'storeID', None)

    def _setPostLocation(self, new):
        try:
            self.original._freeform_post_location = new
        except AttributeError:
            pass

    ## Total hack. This is Quotient specific.
    postLocation = property(_getPostLocation, _setPostLocation)


class ListConfigurable(TypedInterfaceConfigurable):
    def getBinding(self, context, name):
        eb = formless.ElementBinding(name, formless.Object())
        eb.boundTo = self.original
        return eb


class GroupConfigurable(TypedInterfaceConfigurable):
    def __init__(self, original, groupInterface):
        TypedInterfaceConfigurable.__init__(self, original)
        self.groupInterface = groupInterface

    bindingDict = None

    def getBindingNames(self, context):
        bindingNames = []
        self.bindingDict = bindingDict = {}
        interface = self.groupInterface
        for binding in getattr(interface, '__spec__', []):
            bindingDict[binding.name] = binding
            bindingNames.append(binding.name)
            if components.implements(binding.typedValue, formless.IActionableType):
                acts = binding.typedValue.actions
                if acts is None:
                    acts = []
                for action in acts:
                    bindingDict[action.name] = action
        return bindingNames


class IInputValidator(components.Interface):
    """handle a post for a given binding
    """
    def validate(self, context, boundTo, data):
        """do something to boundTo in response to some data
        
        return a status message if everything's ok, and raise a
        formless.ValidateError if there is a problem
        """
    ## TODO should probably make a distinction between a Typed validator
    ## and a Binding validator; probably would make the code cleaner


faketag = html()


class GroupBindingValidator(components.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        ## THE SPEC: self.original.typedValue.interface.__spec__
        spec = self.original.typedValue.interface.__spec__
        resultList = [None] * len(spec)
        message = ''
        results = {}
        failures = {}
        failed = False
        for i, sub in enumerate(spec):
            try:
                resultList[i] = IInputValidator(sub).validate(context, boundTo, data, autoConfigure = False)
            except formless.ValidateError, e:
                _, fbn = calculateFullBindingName(context, self.original)
                if not fbn:
                    fbn = self.original.name
                errors = IFormErrors(context.locate(iwoven.IRequest).getSession())
                errors.updateErrors(fbn, e.errors)
                failed = True
                pf = e.partialForm
                err = e.errors
                msg = e.formErrorMessage
                ## Set an error message for this group of bindings
                errors.setError(fbn, msg)
        if not failed:
            for spec, result in zip(spec, resultList):
                spec.configure(boundTo, result)


class MethodBindingValidator(components.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data, autoConfigure = True):
        """Knows how to validate a dictionary of lists
        where the dictionary may contain a key with the same
        name as some of the arguments to the MethodBinding
        instance.
        """
        typedValue = self.original.typedValue
        results = {}
        failures = {}
        if data.has_key('----'):
            ## ---- is the "direct object", the one argument you can specify using the command line without saying what the argument name is
            data[typedValue.arguments[0].name] = data['----']
            del data['----']
        for binding in typedValue.arguments:
            name = binding.name
            try:
                context = context.with(faketag)
                context.remember(binding, formless.IBinding)
                results[name] = IInputValidator(binding.typedValue).validate(context, boundTo, data.get(name, ['']))
            except formless.InputError, e:
                failures[name] = e.reason

        if failures:
            raise formless.ValidateError(failures, "Error:", results)

        if autoConfigure:
            return self.original.configure(boundTo, results)
        return results


class PropertyBindingValidator(components.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data, autoConfigure = True):
        """Knows how to validate a dictionary of lists
        where the dictionary may contain a key with the
        same name as the property binding's name.
        """
        binding = self.original
        context.remember(binding, formless.IBinding)
        result = {}
        try:
            result[binding.name] = IInputValidator(binding.typedValue).validate(context, boundTo, data.get(binding.name, ['']))
        except formless.InputError, e:
            raise formless.ValidateError({binding.name: e.reason}, e.reason)

        ## atop-specific API; figure out how to remove this.
        if hasattr(boundTo, 'touch'):
            boundTo.touch()
        if autoConfigure:
            return self.original.configure(boundTo, result)
        return result


class TypedValidator(components.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        """data is a list of strings at this point
        """
        typed = self.original
        if hasattr(typed, 'coerceWithBinding'):
            return typed.coerceWithBinding(data[0], boundTo)
        return typed.coerce(data[0])


class PasswordValidator(components.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        """Password needs to look at two passwords in the data,
        """
        pw1 = data[0]
        if pw1 == '':
            raise formless.InputError("Please enter a non-empty password.")
        args = context.locate(iwoven.IRequest).args
        binding = context.locate(formless.IBinding)
        pw2 = args.get("%s____2" % binding.name, [''])[0]
        if pw1 != pw2:
            raise formless.InputError("Passwords do not match. Please reenter.")
        return self.original.coerce(data[0])


class BooleanValidator(components.Adapter):
    def validate(self, context, boundTo, data):
        if data == ['False']:
            data = ['']
        return self.original.coerce(data[0])


class ListValidator(components.Adapter):
    def validate(self, context, boundTo, data):
        request = context.locate(iwoven.IRequest)
        actee = request.args['freeform-actee']
        subsetIdentifiers = [
            x for x in request.args['freeform-action-selection'][0].split('-') if x
        ]

        subset = []

        fullList = getattr(boundTo, actee)

        for item in fullList:
            identifier = locateConfigurable(context, item).postLocation
            if identifier in subsetIdentifiers:
                subset.append(item)
                subsetIdentifiers.remove(identifier)
                if not subsetIdentifiers:
                    break
        ## TODO: Hack for the pool browser; give the method the store ids and 
        ## not the actual item references
        if subsetIdentifiers and not subset:
            return subsetIdentifiers

        return subset


class RequestValidator(components.Adapter):
    __implements__ = IInputValidator,

    def validate(self, context, boundTo, data):
        return context.locate(iwoven.IRequest)

class UploadValidator(components.Adapter):
    __implements__ = IInputValidator,
    def validate(self, context, boundTo, data):
        bind = context.locate(formless.IBinding)
        # TOTAL HACK: this comes from outer space
        fields = context.locate(iwoven.IRequest).fields
        return fields[bind.name]


from nevow import renderer


def instanceQual(instance):
    return reflect.qual(instance.__class__)


class DefaultTypedInterfaceToIResource(renderer.Renderer):
    document = html[
        head[
            title["Configure an object"],
            link(rel="stylesheet", href="/freeform_css")
        ],
        body[
            "Configure an instance of ", instanceQual, ":",
            p[ configure ]
        ]
    ]

