# -*- test-case-name: nevow.test -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import stat
import time
import types
import warnings
from time import time as now

from twisted.python.reflect import qual
from twisted.python import log

from twisted.web import resource, server, microdom
from twisted.internet.defer import Deferred
from twisted.protocols import http

from nevow.iwoven import ISerializable, IRendererFactory, IRequest, IData, IContainer
from nevow.stan import xml, sequence, mapping
from nevow import tags
from nevow.context import WovenContext


def flatten(gen):
    """
    I am a permissive flattener for precompilation.
    """
    results = []
    accumulator = ''
    for item in gen:
        if isinstance(item, types.StringTypes):
            accumulator += item
        else:
            if isinstance(item, types.GeneratorType):
                for sub in flatten(item):
                    if isinstance(sub, types.StringTypes):
                        accumulator += sub
                    else:
                        if accumulator:
                            results.append(xml(accumulator))
                        results.append(sub)
                        accumulator = ''
            elif isinstance(item, types.ListType):
                 if accumulator:
                     results.append(xml(accumulator))
                 results.extend(item)
                 accumulator=''
            else:
                # anything that can be adapted to ISerializable
                if accumulator:
                    results.append(xml(accumulator))
                results.append(item)
                accumulator = ''
    if accumulator:
        results.append(xml(accumulator))
    return results

def _strflatten(context, gen, buf):
    """
    I am a flattener which only allows strings.
    """
    for item in gen:
        if isinstance(item, types.StringTypes):
            buf(item)
        elif isinstance(item, Deferred):
            yield item
            for flat in _strflatten(context, ISerializable(item.result).serialize(context, None), buf):
                pass
        else:
            for sub in _strflatten(context, item, buf):
                yield sub

from cStringIO import StringIO
from twisted.python.context import get as getCtx

def _subrender(L, itr, context, writer, closer):
    for item in itr:
        if writer is not None:
            data = ''.join(L)
            writer(data)
            del L[:]
        cfac = getCtx("CursorFactory")
        if cfac:
            transact = cfac.store.transact
            item.addCallback(lambda ignored: transact(_subrender, L, itr, context, writer, closer))
        else:
            item.addCallback(lambda ignored: _subrender(L, itr, context, writer, closer))
        return item
    data = ''.join(L)
    if writer is not None:
        writer(data)
    if closer is not None:
        closer()
    return data

def render(gen, context=None, writer=None, closer=None):
    L = []
    itr = _strflatten(context, gen, L.append)
    return _subrender(L, itr, context, writer, closer)


class RenderFactory(object):
    __implements__ = IRendererFactory,

    def renderer(self, context, name):
        """Return a renderer with the given name.
        """
        callable = getattr(self, 'render_%s' % name, None)
        if callable is None:
            callable = lambda context, data: context.tag["The renderer named '%s' was not found in %r." % (name, self)]
        return callable

    render_sequence = lambda self, context, data: sequence(context, data)
    render_mapping = lambda self, context, data: mapping(context, data)
    render_string = lambda self, context, data: context.tag.clear()[str(data)]
    render_xml = lambda self, context, data: context.tag.clear()[xml(data)]


class DataFactory(object):
    __implements__ = IContainer,

    def child(self, context, name):
        callable = getattr(self, 'data_%s' % name, None)
        if callable is None:
            container = IContainer(self.original, None)
            if container is None:
                callable = lambda context, data: context.tag["The data named '%s' was not found in %r." % (name, self)]
            else:
                callable = container.child
        return callable


_documents = {}


class ChildPrefixMixin:
    def getChild(self, name, request):
        w = getattr(self, 'child_%s' %name, None)
        if w:
            return w(request)
        return self.getDynamicChild(name, request)

    def getDynamicChild(self, name, request):
        return FourOhFour(self.original)

    def child_(self, request):
        return self


class Renderer(ChildPrefixMixin, resource.Resource, RenderFactory, DataFactory):
    """Subclass me to implement a resource which uses a stan Tag tree
    to render a web page. Declare an html proto tag instance with children
    as a 'document' property. The document will be precompiled, optimizing
    out static strings, and when render is called, the document will be rendered
    with self as the main 'data' object.

    This class also remembers self as IResource and IRendererFactory, and
    remembers the web request as IRequest. So for example, you may
    locate the request from any object which has access to a context using
    context.locate(IRequest).

    This class also provides Deferred support.  If you want to allow Deferreds
    in your pages, but you have to set headers or do other such work from
    within a Deferred, set the 'buffered' attribute to True.
    """
    original = None
    buffered = False

    __implements__ = RenderFactory.__implements__, DataFactory.__implements__, resource.Resource.__implements__

    def __init__(self, original=None):
        if original is not None:
            self.original = original
        self.doc = self.precompile()
        resource.Resource.__init__(self)

    def precompile(self):
        klsnm = qual(self.__class__)
        if klsnm in _documents:
            return [hasattr(x, 'clone') and x.clone() or x for x in _documents[klsnm]]
        context = WovenContext(precompile=True)
        context.remember(self, resource.IResource)
        context.remember(self, IRendererFactory)
        _documents[klsnm] = rv = flatten(ISerializable(self.document).serialize(context, None))
        return rv

    def render(self, request):
        log.msg(http_render=None, uri=request.uri)
        context = WovenContext().remember(self, resource.IResource)
        context.remember(self, IRendererFactory)
        context.remember(request, IRequest)
        if self.original is None:
            context.remember(self, IData)
        else:
            context.remember(self.original, IData)
            context.remember(self.original)
        def clearErrors():
            from nevow import freeform
            sess = request.getSession()
            freeform.IFormErrors(sess).clearAll()
            freeform.IFormDefaults(sess).clearAll()
        if self.buffered:
            io = StringIO()
            writer = io.write
            def finisher():
                request.write(io.getvalue())
                clearErrors()
                return request.finish()
        else:
            writer = request.write
            def finisher():
                clearErrors()
                return request.finish()
        render(ISerializable(self.doc).serialize(context, request), context, writer, finisher)
        return server.NOT_DONE_YET


class HTMLRenderer(Renderer):
    """Subclass me to implement a resource which loads an xhtml template
    off disk.
    """
    templateDirectory = ''
    templateFile = ''
    def __init__(self, original=None, templateDirectory=None, templateFile=None):
        if templateDirectory is not None:
            self.templateDirectory = templateDirectory
        if templateFile is not None:
            self.templateFile = templateFile
        self.precompileTime = None
        Renderer.__init__(self, original)

    def precompile(self):
        context = WovenContext(precompile=True).remember(self, resource.IResource)
        dom = microdom.parse(os.path.join(self.templateDirectory, self.templateFile))
        doc = flatten(ISerializable(dom.childNodes[0]).serialize(context, None))
        # Precompiled. Record the time so we know when to reload the template.
        self.precompileTime = time.time()
        return doc

    def render(self, request):
        # Precompile (again) if the template file has been changed
        fullpath = os.path.join(self.templateDirectory, self.templateFile)
        if self.precompileTime is None or \
               os.stat(fullpath)[stat.ST_CTIME] > self.precompileTime:
            self.doc = self.precompile()
        return Renderer.render(self,request)


class DiskTemplateRenderer(HTMLRenderer):
    def __init__(self, *args, **kw):
        warnings.warn("DiskTemplateRenderer has been renamed HTMLRenderer.", stacklevel=2)
        HTMLRenderer.__init__(self, *args, **kw)


class DynamicRenderer(Renderer):
    """Instantiate me to create an IResource implementor which will render
    any object which implements ISerializable. This will usually be a stan Tag
    instance used by calling the html proto and using stan syntax to build a
    dom of child nodes. Use this renderer to create very dynamic pages whose
    template varies every time the page is rendered, or just to get some html
    onto a web page very quickly.

    This subclass of Renderer uses the same rememberances and can be used
    to provide render_ methods and data_ methods to your template in a
    similar way.
    """
    def __init__(self, doc):
        context = WovenContext(precompile=True).remember(self, resource.IResource)
        context.remember(self, IRendererFactory)
        self.doc = flatten(ISerializable(doc).serialize(context, None))
        ## Bypass Renderer init
        resource.Resource.__init__(self)


class FourOhFour(Renderer):
    isLeaf = True
    document = tags.html[
        tags.head[tags.title["Page Not Found"]],
        tags.body[
            "Sorry, but I couldn't find the object you requested."
        ]
    ]

    def render(self, request):
        request.setResponseCode(http.NOT_FOUND)
        return Renderer.render(self, request)
