# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


import types
import warnings

from twisted.python import components, log
from nevow.stan import Proto, Tag, xml, directive
from nevow.iwoven import ISerializable, IRendererFactory, IData
from nevow.renderer import flatten
from nevow.accessors import convertToData

class ProtoSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        yield xml('<%s />' % self.original)


class TagSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        visible = bool(self.original.tagName)
        singleton = not self.original.renderer and not self.original.children and not self.original.data
        special = context.precompile and self.original._specials
        if self.original.renderer:
            ## If we have a renderer function we want to render what it returns, not our tag
            visible = False
        if special:
            context = context.with(self.original)
            context.tag.children = flatten(ISerializable(context.tag.children).serialize(context, stream))
            yield context
        else:
            if visible:
                yield xml('<%s' % self.original.tagName)
                if self.original.attributes:
                    for (k, v) in self.original.attributes.items():
                        if v is None:
                            warnings.warn("An attribute value for key %r on tag %r was None; ignoring attribute" % (self.original.tagName, v))
                            continue
                        yield xml(' ')
                        yield xml('%s="' % k)
                        flat = flatten(ISerializable(v).serialize(context, stream))
                        if flat:
                            val = flat[0]
                            if isinstance(val, basestring):
                                val = val.replace('"', '&quot;')
                            yield xml(val)
                        yield xml('"')
            if singleton:
                if visible:
                    yield xml(' />')
            else:
                if visible:
                    yield xml('>')
                # TODO: Make this less buggy.
                try:
                    if context.locate(IData) != self.original.data:
                        context = context.with(self.original)                    
                except KeyError:
                    context = context.with(self.original)                    
                if self.original.renderer:
                    toBeRenderedBy = self.original.renderer
                    self.original.renderer = None
                    yield ISerializable(toBeRenderedBy, persist=False).serialize(context, stream)
                    self.original.wasRenderedBy = toBeRenderedBy
                elif self.original.children:
                    for child in self.original.children:
                        yield ISerializable(child).serialize(context, stream)
                if visible:
                    yield xml('</')
                    yield xml(self.original.tagName)
                    yield xml('>')


class StringSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        from twisted.xish.domish import escapeToXml
        ## quote it
        yield escapeToXml(self.original)


class NoneWarningSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        yield xml('<span style="position: relative; font-size: 100; font-weight: bold; color: red; border: thick solid red;">None</span>')


class StringCastSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        from twisted.xish.domish import escapeToXml
        ## quote it
        return escapeToXml(str(self.original))


class ListSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        for item in self.original:
            yield ISerializable(item).serialize(context, stream)


class XmlSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        return self.original


PASS_SELF = object()


class FunctionSerializer(components.Adapter):
    __implements__ = ISerializable,

    def nocontext(self):
        code = getattr(self.original, 'func_code', None)
        if code is None:
            return True
        argcount = code.co_argcount
        if argcount == 1:
            return True
        if argcount == 3:
            return PASS_SELF
        return False

    def serialize(self, context, stream):
        if context.precompile:
            yield self.original
        else:
            data = convertToData(context, context.locate(IData))
            try:
                nocontext = self.nocontext()
                if nocontext is True:
                    result = self.original(data)
                else:
                    if nocontext is PASS_SELF:
                        renderFactory = context.locate(IRendererFactory)
                        result = self.original(renderFactory, context, data)
                    else:
                        result = self.original(context, data)
            except StopIteration:
                log.err()
                raise RuntimeError, "User function %r raised StopIteration." % self.original
            yield ISerializable(result).serialize(context, stream)


class DeferredSerializer(components.Adapter):
    def serialize(self, context, stream):
        yield self.original


class MethodSerializer(FunctionSerializer):
    def nocontext(self):
        func = getattr(self.original, 'im_func', None)
        code = getattr(func, 'func_code', None)
        return code is None or code.co_argcount == 2


class CallableInstanceSerializer(FunctionSerializer):
    def nocontext(self):
        func = getattr(self.original.__call__, 'im_func', None)
        code = getattr(func, 'func_code', None)
        return code is None or code.co_argcount == 2


class DirectiveSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        rendererFactory = context.locate(IRendererFactory)
        renderer = rendererFactory.renderer(context, self.original)
        return ISerializable(renderer).serialize(context, stream)


class ContextSerializer(components.Adapter):
    __implements__ = ISerializable,

    def serialize(self, context, stream):
        originalContext = self.original.clone()
        originalContext.precompile = context.precompile
        originalContext.chain(context)
        try:
            yield flatten(ISerializable(originalContext.tag).serialize(originalContext, stream))
        except:
            from twisted.web import util
            from twisted.python import failure
            from twisted.internet import reactor, defer
            d = defer.Deferred()
            fail = failure.Failure()
            reactor.callLater(0, lambda: d.callback(xml(util.formatFailure(fail))))
            desc = str(fail.value)
            yield ISerializable([
                xml("""<div style="border: 1px dashed red; color: red; clear: both" onclick="this.childNodes[1].style.display = this.childNodes[1].style.display == 'none' ? 'block': 'none'">"""),
                desc,
                xml('<div style="display: none">'),
                d,
                xml('</div></div>')
            ]).serialize(context, stream)



