# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


class Proto(str):
    """Proto is a string subclass. Instances of Proto, which are constructed
    with a string, will construct Tag instances in response to __call__
    and __getitem__, delegating responsibility to the tag.
    """
    __slots__ = []

    def __call__(self, **kw):
        return Tag(self)(**kw)

    def __getitem__(self, children):
        return Tag(self)[children]


class xml(str):
    """Raw xml marker
    """
    __slots__ = []


class directive(str):
    """Marker for a directive in a template
    """
    __slots__ = []

    def __repr__(self):
        return "directive('%s')" % self


class Tag(object):
    """Tag instances represent html tags with a tag name, attributes,
    and children. Tag instances can be constructed using the Prototype
    tags in the 'tags' module, or may be constructed directly with a tag
    name. Tags have two special methods, __call__ and __getitem__,
    which make representing trees of XML natural using pure python
    syntax. See the docstrings for these methods for more details.
    """
    specials = ['data', 'renderer', 'observer', 'remember', 'pattern', 'slot', 'macro', 'fill-slot', 'key']

    produceContext = False

    def __init__(self, tag, attributes=None, children=None, specials=None):
        self.tagName = tag
        if attributes is None:
            self.attributes = {}
        else:
            self.attributes = attributes
        if children is None:
            self.children = []
        else:
            self.children = children
        if specials is None:
            self._specials = {}
        else:
            self._specials = specials

    def __call__(self, **kw):
        """Change attributes of this tag. This is implemented using
        __call__ because it then allows the natural syntax:
        
        table(width="100%", height="50%", border="1")
        
        Three magic attributes may
        have values other than string values, and will be remembered
        in the context stack for later retrieval:
        
        data
        renderer
        observer

        See WovenContext.with for details.
        
        pattern and slot attributes are special, and are not exported
        to the final page, but are used to locate nodes tagged with
        a given pattern or slot id.
        """
        if not kw:
            return self

        for name in self.specials:
            if kw.has_key(name):
                setattr(self, name, kw[name])
                del kw[name]
        for k, v in kw.items():
            if k[0] == '_':
                k = k[1:]
            self.attributes[k] = v
        return self

    def __getitem__(self, children):
        """Add children to this tag. Multiple children may be added by
        passing a tuple or a list. Children may be other tag instances,
        strings, functions, or any object which has a registered
        ISerializable adapter.
        
        This is implemented using __getitem__ because it then allows
        the natural syntax:
        
        html[
            head[
                title["Hello World!"]
            ],
            body[
                "This is a page",
                h3["How are you!"],
                div(style="color: blue")["I hope you are fine."]
            ]
        ]
        """
        if not isinstance(children, (list, tuple)):
            children = [children]
        self.children.extend(children)
        return self

    def clone(self, deep=True):
        """Return a clone of this tag. If deep is True, clone all of this
        tag's children. Otherwise, the children list of the clone will
        be empty.
        """
        if deep:
            newchildren = [
                hasattr(ch, 'clone') and ch.clone() or ch
                for ch in self.children
            ]
        else:
            newchildren = []
        return Tag(self.tagName, attributes=self.attributes.copy(), children=newchildren, specials=self._specials.copy())

    def clear(self):
        """Clear any existing children from this tag.
        """
        self._specials = {}
        self.children = []
        return self

    def __repr__(self):
        rstr = ''
        if self.attributes:
            rstr += ', attributes=%r' % self.attributes
        if self._specials:
            rstr += ', specials=%r' % self._specials
        if self.children:
            rstr += ', children=%r' % self.children
        return "Tag(%r%s)" % (self.tagName, rstr)


def makeAccessors(special):
    def getSpecial(self):
        return self._specials.get(special)

    def setSpecial(self, data):
        self._specials[special] = data

    return getSpecial, setSpecial


for name in Tag.specials:
    setattr(Tag, name, property(*makeAccessors(name)))
del name


def specials(tagOrContext, special):
    """Generate tags with special attributes regardless of attribute value.
    """
    tag = getattr(tagOrContext, 'tag', tagOrContext)
    if getattr(tag, special, None) is not None:
        yield tag
    else:
        for child in getattr(tag, 'children', []):
            for match in specials(child, special):
                yield match


def specialMatches(tagOrContext, special, pattern):
    """Generate special attribute matches starting with the given tag;
    if a match is found, do not look any deeper below that match.
    """
    tag = getattr(tagOrContext, 'tag', tagOrContext)
    if getattr(tag, special, None) == pattern:
        yield tag
    else:
        for child in getattr(tag, 'children', []):
            for match in specialMatches(child, special, pattern):
                yield match


def sequence(context, data):
    headers = specialMatches(context.tag, 'pattern', 'header')
    pattern = context.patterns('item')
    divider = context.patterns('divider', raiseIfMissing=False)
    content = [(pattern(data=element), divider(data=element)) for element in data]
    if not content:
        content = specialMatches(context.tag, 'pattern', 'empty')
    else:
        ## No divider after the last thing.
        content[-1] = content[-1][0]
    footers = specialMatches(context.tag, 'pattern', 'footer')
    return context.tag.clone(deep=False)[ headers, content, footers ]


def mapping(context, data):
    for k, v in data.items():
        for slot in context._locateSlots(k):
            slot[v]
    return context.tag

