# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


from nevow import renderer
from nevow import util

from twisted.trial import unittest
from twisted.internet import defer


class req(util.FakeRequest):
    def __init__(self):
        util.FakeRequest.__init__(self)
        self.d = defer.Deferred()
        self.accumulator = ''

    def write(self, data):
        util.FakeRequest.write(self, data)
        self.accumulator+=data

    def finish(self):
        util.FakeRequest.finish(self)
        self.d.callback(self.accumulator)


class TestHTMLRenderer(unittest.TestCase):
    def deferredRender(self, res):
        defres = req()
        res.render(defres)
        return unittest.deferredResult(defres.d)

    def test_diskTemplate(self):
        temp = self.mktemp()
        open(temp, 'w').write("""<html>
    <head>
        <title nevow:data="theTitle" nevow:render="string">This is some template data!</title>
    </head>
    <body>
        <h3 nevow:data="theHead" nevow:render="string">This is a header</h3>
        <span nevow:data="someDummyText" nevow:render="replaceNode">This node will be replaced entirely (including the span)</span>
    </body>
</html>""")

        class TemplateRenderer(renderer.HTMLRenderer):
            def data_theTitle(self, context, data):
                return "THE TITLE"

            def data_theHead(self, context, data):
                return "THE HEADER"

            def data_someDummyText(self, context, data):
                return "SOME DUMMY TEXT"

            def render_string(self, context, data):
                """The rule is, whatever a renderer returns *replaces* the node that came in
                so if we want this data to *fill* the node the directive was on, we have to
                do it explicitly
                """
                ## This could also be written as:
                # return context.tag.clear()[ data ]
                ## choose your poison
                tag = context.tag.clear()
                tag.children.append(data)
                return tag

            def render_replaceNode(self, context, data):
                """Render the current node by replacing whatever was there (including the
                node itself) with the current data.
                """
                return data

        tr = TemplateRenderer(templateFile=temp)

        result = self.deferredRender(tr)
        self.assertEquals(
            result, 
            '<html><head><title>THE TITLE</title></head><body><h3>THE HEADER</h3>SOME DUMMY TEXT</body></html>'
        )

    def test_sequence(self):
        """Test case provided by mg
        """
        temp = self.mktemp()
        open(temp, 'w').write("""<ol nevow:data="aList" nevow:render="sequence"><li nevow:pattern="item" nevow:render="string"></li></ol>""")

        class TemplateRenderer(renderer.HTMLRenderer):
            def data_aList(self,context,data):
                return ["one","two","three"]

        tr = TemplateRenderer(templateFile=temp)

        result = self.deferredRender(tr)
        self.assertEquals(
            result, '<ol><li>one</li><li>two</li><li>three</li></ol>',
            "Whoops. We didn't get what we expected!"
        )

    def test_sequence2(self):
        """Test case provided by radix
        """
        temp = self.mktemp()
        open(temp, 'w').write("""<ol nevow:data="aList" nevow:render="sequence"><li nevow:pattern="item"><span nevow:render="string" /></li></ol>""")

        class TemplateRenderer(renderer.HTMLRenderer):
            def data_aList(self,context,data):
                return ["one","two","three"]

        tr = TemplateRenderer(templateFile=temp)

        result = self.deferredRender(tr)
        self.assertEquals(
            result, '<ol><li><span>one</span></li><li><span>two</span></li><li><span>three</span></li></ol>',
            "Whoops. We didn't get what we expected!"
        )


    def test_slots(self):
        """test case provided by mg! thanks
        """
        temp = self.mktemp()
        open(temp,'w').write("""
        <table nevow:data="aDict" nevow:render="slots">
        <tr><td nevow:slot="1"></td><td nevow:slot="2"></td></tr>
        <tr><td nevow:slot="3"></td><td nevow:slot="4"></td></tr>
        </table>""")
    
        class Renderer(renderer.HTMLRenderer):
            def data_aDict(self,context,data):
                return {'1':'one','2':'two','3':'three','4':'four'}
            def render_slots(self,context,data):
                for name,value in data.items():
                    for slot in context._locateSlots(name):
                        slot[value]
                return context.tag

        result = self.deferredRender(Renderer(templateFile=temp))
        self.assertEquals(
            result,
            "<table><tr><td>one</td><td>two</td></tr><tr><td>three</td><td>four</td></tr></table>",
            "Whoops. We didn't get what we expected!")

    def test_keys(self):
        temp = self.mktemp()
        open(temp,'w').write("""<span nevow:render="foo">
			<span nevow:key="one">ONE</span>
			<span nevow:key="two">TWO</span>
			<span nevow:key="three">THREE</span>
		</span>""")

        class Mine(renderer.HTMLRenderer):
            def render_foo(self, context, data):
                return context.keyed(data)

        result = self.deferredRender(Mine("one", templateFile=temp))
        self.assertEquals(result, '<span>ONE</span>')
        result = self.deferredRender(Mine("two", templateFile=temp))
        self.assertEquals(result, '<span>TWO</span>')
        result = self.deferredRender(Mine("three", templateFile=temp))
        self.assertEquals(result, '<span>THREE</span>')
        


class TestSubclassAsRenderAndDataFactory(unittest.TestCase):
    def test_rendererSubclass(self):
        from nevow import tags
        class Subclass(renderer.Renderer):
            document = tags.html[
                tags.div(data=tags.directive("hello"))[
                    str
                ],
                tags.span(renderer=tags.directive("world"))
            ]
            def data_hello(self, context, data):
                self.helloCalled = True
                return "hello"
    
            def render_world(self, context, data):
                return "world"

        sr = Subclass()
        defres = req()
        sr.render(defres)
        result = unittest.deferredResult(defres.d)
        self.assertIn('hello', result)
        self.assertIn('world', result)
        self.assertEquals(result,
            "<html><div>hello</div>world</html>"
        )

