# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os

from twisted.trial import unittest

from nevow import formless

class Typed(unittest.TestCase):
    def testString(self):
        s = formless.String()
        self.assertEquals(s.coerce("Fooo"), "Fooo")
        self.assertEquals(s.coerce("This is a string"), "This is a string")
        self.assertEquals(s.coerce(""), "")
    
    def testText(self):
        s = formless.Text()
        self.assertEquals(s.coerce("Fooo"), "Fooo")
        self.assertEquals(s.coerce("This is a string"), "This is a string")
        self.assertEquals(s.coerce(""), "")

    def testPassword(self):
        s = formless.Password()
        self.assertEquals(s.coerce("Fooo"), "Fooo")
        self.assertEquals(s.coerce("This is a string"), "This is a string")
        self.assertEquals(s.coerce(""), "")

    def testHostname(self):
        s = formless.Hostname()
        self.assertEquals(unittest.deferredResult(s.coerce("127.0.0.1")), "127.0.0.1")
        self.assertRaises(formless.InputError, s.coerce, "")
        unittest.deferredError(s.coerce("utterly invalid domain name")).trap(formless.InputError)
    
    def testInteger(self):
        i = formless.Integer()
        self.assertEquals(i.coerce("0"), 0)
        self.assertEquals(i.coerce("3409823098"), 3409823098)
        self.assertRaises(formless.InputError, i.coerce, "")
        self.assertRaises(formless.InputError, i.coerce, "a string")
        self.assertRaises(formless.InputError, i.coerce, "1.5")
    
    def testReal(self):
        i = formless.Real()
        self.assertApproximates(i.coerce("0.0"), 0.0, 1e-10)
        self.assertApproximates(i.coerce("34098.23098"), 34098.23098, 1e-10)
        self.assertRaises(formless.InputError, i.coerce, "")
        self.assertRaises(formless.InputError, i.coerce, "a string")
        self.assertRaises(formless.InputError, i.coerce, "1.5j")

    def testBoolean(self):
        b = formless.Boolean()
        self.assertEquals(b.coerce("zoom"), True)
        self.assertEquals(b.coerce(True), True)
        self.assertEquals(b.coerce(54), True)
        self.assertEquals(b.coerce(""), False)
        self.assertEquals(b.coerce(0), False)
        self.assertEquals(b.coerce(False), False)
    
    def testFixedDigitInteger(self):
        d = formless.FixedDigitInteger(3)
        self.assertEquals(d.coerce("123"), 123)
        self.assertEquals(d.coerce("567"), 567)
        self.assertRaises(formless.InputError, d.coerce, "012")
        self.assertRaises(formless.InputError, d.coerce, "foo")
        self.assertRaises(formless.InputError, d.coerce, "   ")
    
    def testDirectory(self):
        p1 = self.mktemp()
        os.mkdir(p1)
        p2 = self.mktemp()
        
        d = formless.Directory()
        self.assertEquals(d.coerce(p1), p1)
        self.assertRaises(formless.InputError, d.coerce, p2)

    def testTypedInterfaceProperties(self):
        class Test(formless.TypedInterface):
            foo = formless.String()
            bar = formless.Text()
            baz = formless.Integer()

        self.assertEquals(Test.__properties__, Test.__spec__)

        bfoo, bbar, bbaz = Test.__properties__

        self.assertEquals(bfoo.name, 'foo')
        self.assertEquals(bbar.name, 'bar')
        self.assertEquals(bbaz.name, 'baz')

        self.assertEquals(bfoo.typedValue.__class__, formless.String)
        self.assertEquals(bbar.typedValue.__class__, formless.Text)
        self.assertEquals(bbaz.typedValue.__class__, formless.Integer)

    def testTypedInterfaceMethods(self):
        class IFoo(formless.TypedInterface):
            pass

        class Test2(formless.TypedInterface):
            def foo(self, bar=formless.String()):
                """This is a description of foo"""
                pass
            foo = formless.autocallable(foo)

            def bar(self, baz=formless.Integer(label="The Baz")):
                ## this has no docstring, make sure it doesn't fail
                return formless.String()
            bar = formless.autocallable(bar, someAttribute="Hello")

            def baz(self, foo=formless.Boolean(label="The Foo", description="The foo to baz.")):
                """The Label

                The description"""
                return IFoo
            baz = formless.autocallable(baz)

        self.assertEquals(Test2.__methods__, Test2.__spec__)

        bfoo, bbar, bbaz = Test2.__methods__

        self.assertEquals(bfoo.name, 'foo')
        self.assertEquals(bbar.name, 'bar')
        self.assertEquals(bbar.getAttribute('someAttribute'), "Hello")
        self.assertEquals(bbaz.name, 'baz')

        self.assertEquals(bfoo.label, 'Foo')
        self.assertEquals(bfoo.description, 'This is a description of foo')

        self.assertEquals(bbar.label, 'Bar')
        self.assertEquals(bbar.description, '')

        self.assertEquals(bbaz.label, 'The Label')
        self.assertEquals(bbaz.description, 'The description')

        def getArgTypes(mbinding):
            return [x.typedValue.__class__ for x in mbinding.arguments]

        self.assertEquals(getArgTypes(bfoo), [formless.String])
        self.assertEquals(bfoo.returnValue.interface, None)

        self.assertEquals(getArgTypes(bbar), [formless.Integer])
        self.assertEquals(bbar.returnValue.__class__, formless.String)

        self.assertEquals(getArgTypes(bbaz), [formless.Boolean])
        self.assertEquals(bbaz.returnValue.interface, IFoo)

        def firstArg(mbinding):
            return mbinding.arguments[0]

        self.assertEquals(firstArg(bfoo).label, 'Bar')
        self.assertEquals(firstArg(bfoo).description, '')

        self.assertEquals(firstArg(bbar).label, 'The Baz')
        self.assertEquals(firstArg(bbar).description, '')

        self.assertEquals(firstArg(bbaz).label, 'The Foo')
        self.assertEquals(firstArg(bbaz).description, 'The foo to baz.')


class IListWithActions(formless.TypedInterface):
    def actionOne(self, theSubset = formless.List()):
        pass
    def actionTwo(self, theSubset = formless.List()):
        pass

    theListOfStuff = formless.List(actions=[actionOne, actionTwo])


class TestListActions(unittest.TestCase):
    def test_listActionMetadata(self):
        ## IListWithActions only has one binding, a Property binding
        ## of theListOfStuff to a List with some actions.
        actions = IListWithActions.__spec__[0].typedValue.actions
        self.failUnless(reduce, (lambda x: x.name == 'actionOne', actions))
        self.failUnless(reduce, (lambda x: x.name == 'actionTwo', actions))


class TestPropertyGroups(unittest.TestCase):
    def test_nestedTypedInterfaces(self):
        class Outer(formless.TypedInterface):
            aSimpleProperty = formless.Object()

            class Inner(formless.TypedInterface):
                """Docstring
                
                This is a docstring.
                """
                anInnerProperty = formless.Integer()

        self.assertEquals(Outer.__spec__[1].typedValue.interface, Outer.Inner)
        inn = Outer.__spec__[1].typedValue.interface


class TestPosts(unittest.TestCase):
    pass

