

from nevow import tags
from nevow import formless
from nevow import freeform
from nevow import iwoven

from nevow.test.test_flatstan import Base
from nevow.test import test_formless


class Complete(Base):
    def test_configureProperty(self):
        class IStupid(formless.TypedInterface):
            foo = formless.String()

        class StupidThing(freeform.Configurable):
            __implements__ = IStupid,

            def __init__(self):
                self.foo = 'foo'

        dumb = StupidThing()

        val = self.render(tags.invisible(data=dumb) [ freeform.configure ])
        self.assertIn('freeform_post:foo', val)
        self.assertIn('foo', val)
        self.assertIn('type="text"', val)
        self.assertIn('<input type="submit"', val)


    def test_configureMethod(self):
        class IDumb(formless.TypedInterface):
            def foo(self, bar=formless.String()):
                return formless.String()
            foo = formless.autocallable(foo)

        class DumbThing(freeform.Configurable):
            __implements__ = IDumb,

            def foo(self, bar):
                return "baz"

        stupid = DumbThing(1)

        val = self.render(tags.invisible(data=stupid) [ freeform.configure ])

        self.assertIn('freeform_post:foo', val)
        self.assertIn('foo', val)
        self.assertIn('bar', val)


class BuildingBlocksTest(Base):
    def test_1_renderTyped(self):
        binding = formless.Property('hello', formless.String(
            label="Hello",
            description="Hello, world."))

        ## Look up a renderer specific to the type of our binding, typedValue;
        renderer = freeform.ITypedRenderer(binding.typedValue, None, persist=False)
        ## But render the binding itself with this renderer
        ## The binding has the ".name" attribute we need
        val = self.render(tags.invisible(data=binding, renderer=renderer))

        self.assertIn('hello', val)
        self.assertIn('Hello', val)
        self.assertIn('Hello, world.', val)
        self.failIfIn('</form>', val)
        self.failIfIn('<input type="submit"', val)
        
        #print val

    def test_2_renderPropertyBinding(self):
        binding = formless.Property('goodbye', formless.String(
            label="Goodbye",
            description="Goodbye cruel world"))

        ## Look up an IBindingRenderer, which will render the form and the typed
        renderer = freeform.IBindingRenderer(binding)
        val = self.render(tags.invisible(data=binding, renderer=renderer))
        self.assertIn('<form ', val)
        self.assertIn('<input type="submit"', val)
        self.assertIn('name="goodbye"', val)
        self.assertIn('Goodbye', val)
        self.assertIn('Goodbye cruel world', val)
        
        #print val

    def test_3_renderMethodBinding(self):
        binding = formless.MethodBinding('doit', formless.Method(
            returnValue=None,
            arguments=[formless.Argument('foo', formless.String(label="Foo"))],
            label="Do It",
            description="Do it to 'em all"))

        renderer = freeform.IBindingRenderer(binding)
        val = self.render(tags.invisible(data=binding, renderer=renderer))

        self.assertIn('<form action="', val)
        self.assertIn('Do It', val)
        self.assertIn("Do it to 'em all", val)
        self.assertIn("Foo", val)
        self.assertIn('name="foo"', val)
        #print val


class TestDefaults(Base):
    def test_1_renderWithDefaultValues(self):
        binding = formless.MethodBinding('haveFun', formless.Method(
            returnValue=None,
            arguments=[formless.Argument('funValue', formless.Integer(label="Fun Value", default=0))]
        ))

        def setupRequest(request):
            freeform.IFormDefaults(request).setDefault('funValue', 15)
            return request

        renderer = freeform.IBindingRenderer(binding)
        val = self.render(tags.invisible(data=binding, renderer=renderer), setupRequest=setupRequest)
        self.failIfIn('0', val)
        self.assertIn('15', val)

    def test_2_renderWithObjectPropertyValues(self):
        class IDefaultProperty(formless.TypedInterface):
            default = formless.Integer(default=2)

        class Foo(freeform.Configurable):
            __implements__ = IDefaultProperty,
            default = 54

        val = self.render(tags.invisible(data=Foo(None)) [ freeform.configure ])
        self.failIfIn('2', val)
        self.assertIn('54', val)

    def test_3_renderWithAdapteeAttributeValues(self):
        class IDefaultProperty(formless.TypedInterface):
            default = formless.Integer(default=2)

        class Adaptee(object):
            default = 69

        class Bar(freeform.Configurable):
            __implements__ = IDefaultProperty, freeform.Configurable.__implements__


        val = self.render(tags.invisible(data=Bar(Adaptee())) [ freeform.configure ])
        self.failIfIn('2', val)
        self.assertIn('69', val)

    def test_4_testBindingDefaults(self):
        class IBindingDefaults(formless.TypedInterface):
            def aMethod(self, foo=formless.String(default="The foo")):
                pass
            aMethod = formless.autocallable(aMethod)

            aProperty = formless.String(default="The property")

        class Implements(freeform.Configurable):
            __implements__ = IBindingDefaults,

        val = self.render(tags.invisible(data=Implements(None)) [ freeform.configure ] )

        self.assertIn("The foo", val)
        self.assertIn("The property", val)


class ComplexTypeTest(Base):
    def test_1_testList(self):
        class IListProperty(formless.TypedInterface):
            aList = formless.List()

        class IIntProperty(formless.TypedInterface):
            anInt = formless.Integer()

        class IntImpl(freeform.Configurable):
            __implements__ = IIntProperty
            def __init__(self, anInt):
                self.anInt = anInt

        class Implementor(freeform.Configurable):
            __implements__ = IListProperty,
            
            aList = map(IntImpl, range(5))

        val = self.render(tags.html(data=Implementor(None))[ freeform.configure ])
        for i in range(5):
            self.assertIn(str(i), val)
        #print val


class TestNonConfigurableSubclass(Base):
    def test_1_testSimple(self):
        class ISimpleTypedInterface(formless.TypedInterface):
            anInt = formless.Integer()
            def aMethod(self, aString = formless.String()):
                return None
            aMethod = formless.autocallable(aMethod)

        class ANonConfigurable(object): # Not subclassing Configurable
            __implements__ = ISimpleTypedInterface, # But implements a TypedInterface

        val = self.render(tags.html(data=ANonConfigurable())[ freeform.configure ])
        self.assertIn('anInt', val)
        self.assertIn('aMethod', val)


class TestRenderListWithActions(Base):
    def test_1_listWithActions(self):
        class Implementation(object):
            __implements__ = test_formless.IListWithActions,

            actionOneTaken = False
            def __init__(self):
                self.theListOfStuff = map(freeform.Configurable, [1, 2, "buckle", "my", "shoe"])

            def actionOne(self, theSubset):
                self.actionOneTaken = True
                #print "action one was taken on a subset of the list; ", theSubset

            def actionTwo(self, theSubset):
                print "action two was taken on a subset of the list; ", theSubset

        impl = Implementation()
        val = self.render(tags.invisible(data=impl)[ freeform.configure ])
        self.assertIn('actionOne', val)
        self.assertIn('actionTwo', val)
        #print val

        ctx = self.setupContext()
        configurable = freeform.locateConfigurable(ctx, impl)
        dct = {'freeform-actee': 'theListOfStuff', 'freeform-action-selection': ['0-1-']}
        ctx.locate(iwoven.IRequest).args = dct
        posted = configurable.postForm(ctx, "actionOne", dct)
        self.assertEquals(impl.actionOneTaken, True)


class TestPostAForm(Base):
    def test_1_failAndSucceed(self):
        class IAPasswordMethod(formless.TypedInterface):
            def password(self, pword = formless.Password()):
                pass
            password = formless.autocallable(password)

        class APasswordImplementation(object):
            __implements__ = IAPasswordMethod,
            matched = False
            def password(self, pword):
                self.matched = True
                return "password matched"

        theObj = APasswordImplementation()
        ctx = self.setupContext()
        configurable = freeform.locateConfigurable(ctx, theObj)

        result = configurable.postForm(ctx, "password", {"pword": ["these passwords"], "pword____2": ["don't match"]})

        self.assertEquals(theObj.matched, False)

        val = self.render(
            tags.invisible(data=theObj)[ freeform.configure ],
            setupRequest=lambda r:ctx.locate(iwoven.IRequest))

        self.assertIn("Passwords do not match. Please reenter.", val)


class TestRenderPropertyGroup(Base):
    def test_1_propertyGroup(self):
        class Outer(formless.TypedInterface):
            class Inner(formless.TypedInterface):
                one = formless.Integer()
                two = formless.Integer()

                def buckleMyShoe(self):
                    pass
                buckleMyShoe = formless.autocallable(buckleMyShoe)

                def buriedAlive(self):
                    pass
                buriedAlive = formless.autocallable(buriedAlive)

        class Implementation(object):
            __implements__ = Outer,
            one = 1
            two = 2
            buckled = False
            buried = False
            def buckleMyShoe(self):
                self.buckled = True
            def buriedAlive(self):
                self.buried = True

        impl = Implementation()
        ctx = self.setupContext()
        val = self.render(tags.invisible(data=impl)[ freeform.configure ])

        configurable = freeform.locateConfigurable(ctx, impl)

        posted = configurable.postForm(ctx, "Inner", {'one': ['Not an integer'], 'two': ['22']})

        self.assertEquals(impl.one, 1)
        self.assertEquals(impl.two, 2)
        self.assertEquals(impl.buckled, False)
        self.assertEquals(impl.buried, False)

        val = self.render(tags.invisible(data=impl)[ freeform.configure ], setupContext=lambda c: ctx)
        self.assertIn("is not an integer", val)

        posted = configurable.postForm(ctx, "Inner", {'one': ['11'], 'two': ['22']})

        self.assertEquals(impl.one, 11)
        self.assertEquals(impl.two, 22)
        self.assertEquals(impl.buckled, True)
        self.assertEquals(impl.buried, True)

    def test_2_renderWithList(self):
        class IListItem(formless.TypedInterface):
            class Inner(formless.TypedInterface):
                one = formless.Integer()
            outer = formless.Real()

        class ItemImplementation(object):
            __implements__ = IListItem,
            def __init__(self, one):
                self.one = one
                self.outer = 3.14

        class IItemWithList(formless.TypedInterface):
            theList = formless.List(header="HEADER", footer="FOOTER")

        class WithListImpl(object):
            __implements__ = IItemWithList,

            theList = map(ItemImplementation, range(2))

        inst = WithListImpl()
        val = self.render(tags.invisible(data=inst)[ freeform.configure ])

        for x in range(2):
            self.assertIn("freeform_post:theList.%s.Inner" % x, val)

        ctx = self.setupContext()
        configurable = freeform.locateConfigurable(ctx, inst)
        posted = configurable.postForm(ctx, "theList.0.Inner", {'one': ['Not an integer']})

        val = self.render(tags.invisible(data=inst)[ freeform.configure], setupContext=lambda c: ctx)
        self.assertIn("is not an integer", val)
        self.assertIn("display: block", val)

        posted = configurable.postForm(ctx, "theList.0.outer", {'outer': ['not a real']})
        val = self.render(tags.invisible(data=inst)[ freeform.configure ], setupContext=lambda c: ctx)
        ### TODO Make these cases pass
        #self.assertIn("is not a real", val)
        #self.assertIn("display: block", val)
        
        self.assertIn("HEADER", val, msg="Missing list header")
        self.assertIn("FOOTER", val, msg = "Missing list footer")

#        print val


class TestCustomTyped(Base):
    def test_typedCoerceWithBinding(self):
        class MyTyped(formless.Typed):
            passed = False
            wasBoundTo = None
            def coerceWithBinding(self, val, boundTo):
                self.passed = True
                self.wasBoundTo = boundTo
                return "passed"

        typedinst = MyTyped()

        class IMyInterface(formless.TypedInterface):
            def theFunc(self, test=typedinst):
                pass
            theFunc = formless.autocallable(theFunc)

        class Implementation(object):
            __implements__ = IMyInterface
            called = False
            def theFunc(self, test):
                self.called = True

        inst = Implementation()
        ctx = self.setupContext()
        configurable = freeform.locateConfigurable(ctx, inst)
        configurable.postForm(ctx, 'theFunc', {'test': ['a test value']})
        
        self.assertEquals(typedinst.passed, True)
        self.assertEquals(typedinst.wasBoundTo, inst)

        self.assertEquals(inst.called, True)


