# -*- test-case-name: quotient.test.test_blog -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.
#

import itertools

from atop import store

from nevow import tags
from nevow import freeform
from nevow import formless
from nevow import iwoven

from quotient import taglib
from quotient import provisioner
from quotient.web.root import Base
from quotient.web.webmail import messageBody, render_received
from quotient import searchup
from quotient.smarts.list import IRouteTarget

from twisted.python import components


class IBlogUp(formless.TypedInterface):
    def newPost(self, subject=formless.String(), body=formless.Text()):
        """New Post

        Make a new blog entry
        """
        pass
    newPost = formless.autocallable(newPost)


class IBlog(components.Interface):
    pass

class BlogUp(provisioner.ProvisionedPowerup):

    __implements__ = (IRouteTarget, provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable, store.IPipe, IBlog)

    dependencies = [provisioner.Provisioner, searchup.Searchup]
    powerupName = "Blog"

    tankID = provisioner.ETank.BLOG_POST
    startCost = 0

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)
        root = avatar.getRootPool()
        blog = store.Pool(avatar, name="Blog")
        blog.addIndex(store.DateIndex(store.RECEIVED))
        blog.addPipe(self.fwd(store.IPipe))
        root.addItem(blog)
        self.blogRef = blog.referenceTo()
        searcher = searchup.ISearch(avatar)
        searcher.indexPool(blog)
        avatar.setComponent(IBlog, self)

    # IPipe implementation

    def itemAdded(self, item):
        blogPool = store.Pool(self.store)
        blogPool.addIndex(store.DateIndex(store.RECEIVED))
        item.blogPoolRef = blogPool.referenceTo()

    def itemRemoved(self, item):
        p = item.blogPoolRef.getItem()
        del item.blogPoolRef
        for i in p.queryIndex():
            p.removeItem(i)
        # XXX CLEAN UP MORE

    def itemChanged(self, item, changeData):
        pass

    # IRouteTarget implementation

    def receive(self, route, item):
        if not route:
            # If route is empty, it's a new top-level post
            self.blogRef.getItem().addItem(item)
            return True
        else:
            # Otherwise it might be a reply to an existing post
            try:
                uid = int(route.pop())
            except ValueError:
                return False
            else:
                blog = self.blogRef.getItem()
                parent = blog.getIndexItem('uid', uid, None)
                if parent is None:
                    return False
                self.addReply(parent, item)
                return True

    def command(self, cmd, arg):
        if 'blog'.startswith(cmd):
            def _():
                n = self.newPost(arg)
            self.store.transact(_)

    def newPost(self, subject, body=''):
        import mimebakery
        r = mimebakery.Recipe(self.store, "me", subject=subject)
        r.text(body)
        blog = self.blogRef.getItem()
        blog.addItem(r.bake())

    # Handy stuff

    def addReply(self, parent, reply):
        assert parent in self.blogRef.getItem(), "Can't reply to posts not in the blog pool"
        parent.blogPoolRef.getItem().addItem(reply)

    def posts(self, parent=None):
        """Iterate over the top-level posts of this blog.

        Posts are ordered by date, starting at the most recent.
        """
        b = self.blogRef.getItem()
        if parent is None:
            seq = b.queryIndex
        else:
            seq = parent.blogPoolRef.getItem().queryIndex
        for o in seq(store.RECEIVED, result=store.ITEM, backwards=True):
            yield o


class IFoo(formless.TypedInterface):
    foo = formless.String()


class BlogPostDisplay(components.Adapter):
    __implements__ = iwoven.IRendererFactory, IFoo

    foo = "hello"

    renderExtra = True

    def blogControlLinks(self, context, data):
        if not self.renderExtra:
            return ''
        request = context.locate(iwoven.IRequest)
        url = request.URLPath()
        return tags.p[
            tags.a(href="%s#%s" % (url, self.original.storeID))["#"],
            " ",
            tags.a(href="%s/%s" % (url, self.original.storeID))["Comments"]
        ]

    def __call__(self, context, data):
        msg = self.original
        return tags.a(name="%s" % self.original.storeID)[
            tags.p(data=msg)[
                msg['subject'],
                tags.div[messageBody],
                self.blogControlLinks
            ]
        ]


# this is going away asap
N_DISPLAY = 10


def render_posts(self, context, data):
    return [
        tags.invisible(renderer=BlogPostDisplay(x)) for x in
        itertools.islice(data, N_DISPLAY)
    ]

class BlogDisplay(Base):
    __implements__ = (Base.__implements__, IBlogUp)

    ## IContainer needs to be made aware of methods
    def data_posts(context, data):
        return data.posts()

    document = taglib.page(title="Blog")[
        tags.p(data=data_posts)[render_posts],
        freeform.configure
    ]

    def newPost(self, subject, body):
        self.original.newPost(subject, body)

    def getChild(self, name, request):
        return Comments(self.original, self.original.store.getItemByID(int(name)))


class ICommentable(formless.TypedInterface):
    def addComment(
        self,
        name=formless.String(),
        email=formless.String(),
        url=formless.String(),
        body=formless.Text()
    ):
        """Add Comment

        Comment on this post.
        """
    addComment = formless.autocallable(addComment)


def calcHREF(context, message):
    fromWho = message.get('x-email')
    if fromWho:
        fromWho = "mailto:%s" % fromWho
    else:
        fromWho = message.get('x-url', None)
    fromName = message.get('x-name', "Anonymous")
    if fromWho:
        return tags.a(href=fromWho)[fromName]
    return fromName


class Comments(Base):
    __implements__ = Base.__implements__, ICommentable
    def __init__(self, original, message):
        Base.__init__(self, original)
        self.message = message

    def render_message(self, context, data):
        it = BlogPostDisplay(self.message)
        it.renderExtra=False
        return tags.invisible(
            renderer=it
        )

    def data_comments(self, context, data):
        return data.posts(self.message)

    ## this is stupid, make it better
    def format_comments(self, context, data):
        return [
            tags.p(data=message)[
                messageBody,
                tags.div(style="font-size: xx-small")[
                    "Posted by ", calcHREF,
                    " at ", render_received
                ]
            ] for message in data
        ]

    def render_comments(self, context, data):
        return tags.p(data=self.data_comments)[
            self.format_comments
        ]

    document = taglib.page(title="")[
        render_message,
        tags.hr,
        render_comments,
        freeform.configure
    ]

    def addComment(self, name, email, url, body):
        import mimebakery
        r = mimebakery.Recipe(self.original.store, "%s <%s>" % (name, email), subject=url)
        r.text(body)
        baked = r.bake()
        baked['x-name'] = name
        baked['x-email'] = email
        baked['x-url'] = url
        self.original.addReply(self.message, baked)
