# -*- test-case-name: quotient.test.test_spambayesup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import time
import rfc822

from nevow import tags
from nevow import formless
from nevow import freeform

from atop.powerup import Powerup
from atop.store import IPipe, Pool
from atop.tpython import Forgetter

from twisted.python import log
from twisted.internet import threads, reactor
from twisted.python.components import Interface

from quotient.stats import Statoscope
from quotient.mimemessage import MIMEMessage, EmailPowerup, ICleanPoint, IMIMEDelivery, newPoolForEMail
from quotient.searchup import Searchup, ISearch
from quotient.proto.smtpout import IOutgoingMail, OutgoingMailup
from quotient import taglib
from quotient.web.root import Base
from quotient.prefup import touchNForward, IPrefs, Prefup

# Hack to get around isinstance(email.Message.Message)
import spambayes.mboxutils as mbu
mbu.get_message = lambda x:x
from spambayes import hammie
from spambayes.classifier import Bayes
from spambayes.Options import options


class PreferenceGrouping(formless.TypedInterface):
    cleanThresh = formless.Integer(
        label="Clean Threshold",
        description="Above this threshold, messages are marked as clean.")
    spamThresh = formless.Integer(
        label="Spam Threshold",
        description="Below this threshold, messages are marked as spam.")

class ISpamPref(formless.TypedInterface):
    PreferenceGrouping = PreferenceGrouping

    incinerateSpam = formless.Boolean(
        label="Incinerate Spam",
        description="Destroy spam after the configured holding period, instead of moving it to trash.")


class SpambayesupDisplay(Base):
    __implements__ = Base.__implements__, ISpamPref

    # XXX TODO - Get rid of this somehow
    postLocation = property(lambda self: self.original.storeID)

    incinerateSpam = touchNForward("incinerateSpam")

    def spamThreshGet(self):
        return self.original.spamThresh

    def spamThreshSet(self, thresh):
        self.original.spamThresh = thresh
        self.original.setTrainedTime()
    spamThresh = property(spamThreshGet, spamThreshSet)

    def cleanThreshGet(self):
        return self.original.cleanThresh

    def cleanThreshSet(self, thresh):
        self.original.cleanThresh = thresh
        self.original.setTrainedTime()
    cleanThresh = property(cleanThreshGet, spamThreshSet)

    def render_msgsProcessed(self, context, data):
        return self.original.msgsProcessed

    def render_nspam(self, context, data):
        return self.original.guesser.bayes.nspam

    def render_nham(self, context, data):
        return self.original.guesser.bayes.nham

    document = taglib.page(title="Spambayes Home")[
        ## Todo: Add "editable" flag to formless, so the following can just be described in the TypedInterface
        tags.p["Messages Processed: ", render_msgsProcessed],
        tags.p["Trained Spam: ", render_nspam],
        tags.p["Trained Clean: ", render_nham],
        freeform.configure
    ]

    def render(self, request):
        getByID = self.original.store.getItemByID
        if request.args.has_key('spam'):
            msgID = request.args['spam'][0]
            msg = getByID(int(msgID))
            self.original.train(msg, 'spam')
        elif request.args.has_key('clean'):
            msgID = request.args['clean'][0]
            msg = getByID(int(msgID))
            self.original.train(msg, 'clean')
        elif request.args.has_key('file'):
            poolToFileIn = request.args['file'][0]
            itemsToFile = request.args['items']
            if poolToFileIn == 'Clean':
                removeFrom = self.original.spamRef.getItem()
                addTo = self.original.cleanRef.getItem()
            elif poolToFileIn == 'Spam':
                removeFrom = self.original.cleanRef.getItem()
                addTo = self.original.spamRef.getItem()
            else:
                raise RuntimeError, "Misuse of spam powerup"
            def _():
                for itemid in itemsToFile:
                    item = getByID(int(itemid))
                    removeFrom.removeItem(item)
                    addTo.addItem(item)
            self.original.store.transact(_)
        else:
            return Base.render(self, request)
        request.redirect(request.getHeader('referer'))
        return Base.render(self, request)


class ISpambayes(Interface):
    """
    @ivar incinerateSpam: If true, indicates that spam should be
    destroyed after the holding period.  Otherwise, spam should
    be moved to the trash pool for manual deletion.
    """

    def train(self, item, pool):
        pass

class Spambayesup(Powerup):
    __implements__ = ISpambayes, Powerup.__implements__, IPipe
    dependencies = [EmailPowerup]
    powerupName = 'SpamBayes'
    persistenceForgets = ['guesser']

    # TODO - This should all be moved into Cleanup
    # I wrote this first and then cleanup and did
    # want to change too much at once - AAB

    incinerateSpam = False

    spamThresh = 20
    cleanThresh = 90
    msgsProcessed = 0

    def setUpPools(self, avatar):
        sbDir = avatar.newDirectory('spambayes')
        self.sbPath = os.path.join(sbDir, 'spambayes.db')
        self.activate()
        avatar.setComponent(ISpambayes, self)
        self.msgsProcessed = 0
        self.setTrainedTime()
        self.pools = {}

    def activate(self):
        self.guesser = hammie.open(self.sbPath, useDB=False, mode='c')

    def setTrainedTime(self):
        self.lastTrainedTime = time.time()
        self.touch()

    def train(self, item, pool):
        def _():
            if pool.lower() == 'spam':
                self.guesser.train(item, True)
            else:
                self.guesser.train(item, False)
            self.setTrainedTime()
            self.touch()
        self.store.transact(_)
        self.guesser.store()

    def match(self, item):
        pool, prob = self.guessPool(item)
        if pool.lower() == 'spam':
            dest = self.spamRef
        else:
            dest = self.cleanRef
        dest.getItem().addItem(item)
        self.msgsProcessed += 1
        self.touch()

    def guessPool(self, item):
        # prob = self.guesser.score(item)
        finfo = getattr(item, 'filingInfo', None)
        if finfo:
            pool, prob, lTime = finfo
            if pool is None:
                # Message filed via a rule
                return (None, prob)
            else:
                if not self.lastTrainedTime > lTime:
                    # cache hit
                    return (pool, prob)

        toks = hammie.tokenize(item)
        # Could iterate tokens in a reactor before calling
        # spamprob
        prob = self.guesser.bayes.spamprob(toks, False)
        prob = 1.0 - prob
        if prob > self.cleanThresh/100.0:
            pool = 'Clean'
        elif prob < self.spamThresh/100.0:
            pool = 'Spam'
        else:
            pool = 'Unknown'
        item.filingInfo = (pool, prob, time.time())
        item.touch()
        return (pool, prob)

    def setCleanAndSpam(self, clean, spam):
        """
        This needs to be generalized for Reverend, which can
        handle multiple aribtrary pools
        """
        self.cleanRef = clean.referenceTo()
        self.spamRef = spam.referenceTo()
        spam.addPipe(self.fwd())

    ##
    ## IPipe implementation
    ##

    def itemAdded(self, item):
        pool, prob = self.guessPool(item)
        if pool != 'Spam':
            self.train(item, 'spam')

    def itemRemoved(self, item):
        pool, prob = self.guessPool(item)
        if pool == 'Spam':
            self.train(item, 'clean')

    def itemChanged(self, item, notificationData):
        pass


class IBasicWhitelist(formless.TypedInterface):
    defaultFrom = formless.String(label="Add", description="")


class WhitelistDisplay(Base, freeform.Configurable):
    __implements__ = IBasicWhitelist,
    def render_whitelist(self, context, data):
        addys = self.original.whitelist.list.keys()
        addys.sort()
        disp = []
        for a in addys:
            disp.append(tags.p["%s" % a])
        if disp:
            return tags.p["Addresses in your whitelist:"], disp
        return tags.p["There are no addresses in your whitelist."]

    document = taglib.page(title="Whitelist Home")[
        tags.p["Add an address to your whitelist:"],
        tags.form(action="")[
            tags.input(type="text", name="add"),
            tags.input(type="submit")
        ],
        render_whitelist
    ]

    def render(self, request):
        if request.args.has_key('add'):
            addy = request.args['add'][0]
            def _():
                self.original.addToWhitelist(addy)
            self.original.store.transact(_)
            request.redirect(request.getHeader('referer'))
            return Base.render(self, request)
        return Base.render(self, request)

class IRuler(Interface):
    def match(self, item):
        """Return True or False"""

class IRule(Interface):
    def match(self, item):
        """Return True or False"""

class Whitelist(object):
    __implements__ = IRule

    def __init__(self):
        self.list = {}

    def match(self, item):
        name, address = item.nameAndAddress()
        address = address.lower()
        if address in self.list:
            return True
        return False

    def addToList(self, address):
        self.list[address.lower()] = True

class Ruler(Powerup):
    __implements__ = IRuler, Powerup.__implements__
    dependencies = [OutgoingMailup, Prefup]
    powerupName = 'Whitelist'

    def setUpPools(self, avatar):
        avatar.setComponent(IRuler, self)
        self.rules = {}
        self.whitelist = Whitelist()
        outmail = IOutgoingMail(avatar)
        outmail.sentRef.getItem().addPipe(self.fwd(IPipe))

    def match(self, item):
        if self.whitelist.match(item):
            item.filingInfo = (None, 'whitelist', None)
            item.touch()
            return True
        for name, rule in self.rules.items():
            if rule.match(item):
                item.filingInfo = (None, name, None)
                item.touch()
                return True
        return False

    def addRule(self, ruleName, rule):
        self.touch()
        self.rules[ruleName] = rule

    def addToWhitelist(self, address):
        self.touch()
        self.whitelist.addToList(address)

    def itemAdded(self, item):
        """
        Automatically add folks that you sende
        a message to, to your whitelist.
        """
        if IPrefs(self.store).whitelistOnSend:
            addys = ''
            for field in ['to','cc','bcc']:
                addy = item[field]
                if addy:
                    addys += addy + ','
            addresses = [x[1] for x in rfc822.AddressList(addys)]
            assert addresses, 'There must be a to, cc or bcc header'
            for wl in addresses:
                self.addToWhitelist(wl)


class IClean(Interface):
    pass

class Cleanup(Powerup):
    __implements__ = IClean, IPipe, Powerup.__implements__
    dependencies = [Spambayesup, Ruler, Searchup]
    powerupName = 'Cleaner'

    def activate(self):
        pass

    def setUpPools(self, avatar):
        arrRef = IMIMEDelivery(avatar).getArrivalRef()
        arrRef.getItem().addPipe(self.fwd(IPipe))
        avatar.setComponent(IClean, self)
        searcher = ISearch(avatar)
        self.rulerRef = IRuler(avatar).referenceTo()
        sb = ISpambayes(avatar)
        self.sbRef = sb.referenceTo()
        spam = newPoolForEMail(avatar, 'Spam')
        clean = newPoolForEMail(avatar, 'Clean')
        clean.invisible = True
        self.spamRef = spam.referenceTo()
        self.cleanRef = clean.referenceTo()
        avatar.getRootPool().addItem(clean)
        avatar.getRootPool().addItem(spam)
        sb.setCleanAndSpam(clean, spam)
        searcher.indexPool(clean)
        # Move the clean point
        cp = ICleanPoint(avatar)
        cp.getItem().setPoint(clean)
        self.greyedOut = True
        self.invisible = True
        # IPrefs(avatar).addToPrimaryNav(spam)
        # HACK - It's as good a place as any
        IPrefs(avatar).addToPrimaryNav(avatar.getRootPool())

    def itemAdded(self, item):
        if not isinstance(item, MIMEMessage):
            return
        if self.rulerRef.getItem().match(item):
            self.cleanRef.getItem().addItem(item)
        else:
            self.sbRef.getItem().match(item)


class HeaderRule(object):
    __implements__ = (IRule,)

    def __init__(self, headerName, headerValue):
        self.name = headerName
        self.value = headerValue

    def match(self, item):
        return self.op(item[self.name], self.value)

class HeaderEquals(HeaderRule):
    def op(self, orig, spec):
        return orig.lower() == spec.lower()

class HeaderDoesntEqual(HeaderRule):
    def op(self, orig, spec):
        return orig.lower() != spec.lower()

class HeaderContains(HeaderRule):
    def op(self, orig, spec):
        return orig.lower().find(spec.lower()) != -1

class HeaderDoesntContain(HeaderRule):
    def op(self, orig, spec):
        return orig.lower().find(spec.lower()) == -1

class HeaderMatchesExpression(HeaderRule):
    _cache = None
    def op(self, orig, spec):
        if self._cache is None:
            self._cache = re.compile(spec)
        return self._cache.match(orig) is not None

class HeaderDoesntMatchExpression(HeaderRule):
    _cache = None
    def op(self, orig, spec):
        if self._cache is None:
            self._cache = re.compile(spec)
        return self._cache.match(orig) is not None

def HeaderRuleDisplay(context, data):
    return tags.form(action=str(data.storeID))[
        "Header:",
        tags.input(type="text", name="headerName"),
        "Operation:",
        tags.input(type="text", name="operation"),
        "Value:",
        tags.input(type="text", name="headerValue"),
        ]
