# -*- test-case-name: quotient.test.test_mimemessage -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from tempfile import TemporaryFile
import datetime
import types
import time

from atop import tpython

from twisted.persisted.styles import Versioned
from twisted.internet import defer, reactor, protocol, ssl as sslModule
from twisted.python.components import implements, Interface
from twisted.python import log, failure
from twisted.protocols import pop3

from twisted.protocols import imap4

from twisted.python.failure import Failure
from twisted.internet.error import ConnectionDone

from nevow import formless
from nevow import freeform
from atop import store
from atop import sched
from atop import powerup
from atop.fwd import ItemForwardingComponent

from quotient import stats
from quotient import mimemessage
from quotient import provisioner
from quotient import sysup

class GrabStatus:
    """Formattable object that describes status of a grabber.
    """

    message = "***ERROR***"
    successful = 1

    def __init__(self, message='***ERROR***', successful=1):
        self.lastChanged = time.time()
        self.message = message
        self.successful = successful

    def __str__(self):
        return self.message

    def __repr__(self):
        return '<%s: %r>' % (self.__class__, self.message)

class NeverChecked(GrabStatus):
    def __init__(self, message="Never checked this account."):
        GrabStatus.__init__(self, message)

class CheckingPaused(GrabStatus):
    def __init__(self, message="Checking paused."):
        GrabStatus.__init__(self, message, 0)

class CheckSuccessful(GrabStatus):
    def __init__(self, messageCount):
        self.messageCount = messageCount
        self.curTime = time.localtime()
        GrabStatus.__init__(self, "Last check %s; Retrieved %s messages" % (
            time.asctime(self.curTime), messageCount
        ))

class CheckInProgress(GrabStatus):
    def __init__(self, message="Check in progress..."):
        GrabStatus.__init__(self, message)


class FetchingMessage(GrabStatus):
    def __init__(self, xOf, y):
        GrabStatus.__init__(self, "Retrieving message %s of %s" % (xOf + 1, y))

class AuthFailed(GrabStatus):
    def __init__(self, message):
        GrabStatus.__init__(self, "Server said: %s" % (message,), 0)

class RetrieveFailed(GrabStatus):
    def __init__(self, message):
        GrabStatus.__init__(self, "Retrieval Failed: %s" % (message,), 0)

class ExamineFailed(GrabStatus):
    def __init__(self, message):
        GrabStatus.__init__(self, "IMAP4 EXAMINE Failed: %s" % (message,), 0)

class LoggingIn(GrabStatus):
    def __init__(self, message="Logging in..."):
        GrabStatus.__init__(self, message)

class SelectingMailbox(GrabStatus):
    def __init__(self, message="Selecting mailbox"):
        GrabStatus.__init__(self, message)

class POP3MessageGrabber(pop3.POP3Client):
    # connect to server
    # authenticate
    # retrieve UIDL message listing
    # compare to persistent message listing from this host
    # retrieve new messages

    messageCount = 0
    sawLast = 0

    def connectionMade(self):
        self.retrievalStats = stats.Statoscope(name="popgrabber")

    def logout(self):
        delete = self.factory.getDeletionList()
        if not delete:
            self.transport.loseConnection()
            return
        # Yeaaaaaaaarg
        self.installSpecialUIDLHandlers()
        self.uidsToDelete = dict.fromkeys(delete)
        self.uidl()

    def installSpecialUIDLHandlers(self):
        self.handle_UIDL_continue = self.deletionPhase_UIDL_continue
        self.handle_UIDL_end = self.deletionPhase_UIDL_end
        self.uidl = []

    def deletionPhase_UIDL_continue(self, data):
        cinfo = data.split(None, 1)
        if len(cinfo) == 2:
            if cinfo[1] in self.uidsToDelete:
                self.uidl.append(c[0])

    def deletionPhase_UIDL_end(self):
        for uid in self.uidls:
            self.dele(uid)
        self.transport.loseConnection()

    def handle_WELCOME(self, line):
        pop3.POP3Client.handle_WELCOME(self, line)
        self.user(self.factory.username)
        self.factory.setStatus(CheckInProgress())

    def handle_USER(self, data):
        cinfo = data.split(' ', 1)
        code, info = cinfo[0], ' '.join(cinfo[1:])
        if code == '+OK':
            self.pass_(self.factory.password)
        else:
            self.factory.setStatus(AuthFailed(info))
            self.transport.loseConnection()

    def handle_PASS(self, data):
        cinfo = data.split(' ', 1)
        code, info = cinfo[0], ' '.join(cinfo[1:])
        if code == '+OK':
            self.idxs = []
            self.uidls = []
            self.uidl()
        else:
            self.factory.setStatus(AuthFailed(info))
            self.transport.loseConnection()

    def handle_UIDL_continue(self, data):
        "UIDL response"
        cinfo = data.split(' ', 1)
        idx, uid = cinfo[0], ' '.join(cinfo[1:])
        if uid == self.lastUid:
            self.sawLast = 1
            self.idxs = []
            self.uidls = []
        else:
            self.idxs.append(idx)
            self.uidls.append(uid)

    def handle_UIDL_end(self):
        "UIDL complete"
        self.ptr = 0
        self.getNext()

    def getNext(self):
        index = self.ptr
        self.ptr += 1
        if len(self.idxs) <= index:
            self.factory.setStatus(CheckSuccessful(index))
            self.logout()
        else:
            idx = self.idxs[index]
            uid = self.uidls[index]
            self.msgReceiver = self.factory.createMIMEReceiver(True)
            self.msgReceiver.makeConnection(None)
            self.curIdx = idx
            self.curUid = uid
            self.retr(idx)
            self.factory.setStatus(FetchingMessage(index, len(self.idxs)))
            self.rtime = time.time()
            self.bytes = 0

    def handle_RETR(self, data):
        "'+OK xxx octets' line"
        cinfo = data.split(' ', 1)
        code, data = cinfo[0], ' '.join(cinfo[1:])
        if code == '-ERR':
            self.factory.setStatus(
                RetrieveFailed(data + " " + self.curIdx + " " + self.curUid))
            self.getNext()

    def handle_RETR_continue(self, data):
        "incoming message data"
        self.msgReceiver.lineReceived(data)

        self.bytes += len(data)

    def handle_RETR_end(self):
        "message data end"
        def _():
            self.msgReceiver.messageDone()
            self.lastUid = self.curUid
            self.factory.setLastUid(self.lastUid)
            self.messageCount += 1

            self.factory.messageReceived(self.msgReceiver.message, self.bytes, self.curUid)

            # I don't like how this makes me record stat info on self
            # Perhaps "m = stats.startingMessage()" later followed by
            # multiple calls to "m.addBytes(b)" and a final call to
            # "m.finished()"?
            self.retrievalStats.record(self.bytes, time.time() - self.rtime)
        self.factory.avatar.transact(_)
        self.bytes = 0
        self.rtime = None
        ###
        log.msg('POP3: %s'% self.retrievalStats)

        if self.factory.paused:
            self.logout()
            return

        if self.maxMsgs is not None:
            self.maxMsgs -= 1
            if self.maxMsgs <= 0:
                self.logout()
                return
        self.getNext()

    def connectionLost(self, reason):
        if self.messageCount > 0:
            self.factory.setStatus(CheckSuccessful(self.messageCount))
            self.factory.successfulCheck = True
        elif self.factory.paused:
            self.factory.setStatus(CheckingPaused("Checking paused: User requested"))
            self.factory.successfulCheck = True

class FileWrapper:
    # XXX - This is ghetto, clean it up
    def __init__(self, messageReceiver, stats = None):
        self.mr = messageReceiver
        self.mr.makeConnection(None)
        self.buffer = ''
        self.stats = stats
        if stats:
            stats.start()
        self.bytes = 0

    def write(self, bytes):
        self.bytes += len(bytes)
        self.buffer += bytes
        lines = self.buffer.splitlines(True)
        for l in lines:
            if l.endswith('\n') or l.endswith('\r\n'):
                l = l.rstrip('\r\n')
                self.mr.lineReceived(l)
            else:
                self.buffer = l
                break

    def seek(self, off, from_):
        pass

    def tell(self):
        return self.bytes

    def close(self):
        self.mr.messageDone()
        self.mr.connectionLost(Failure(ConnectionDone()))
        if self.stats:
            self.stats.finish(self.bytes)

    messageDone = close

MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
          "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

class IMAP4MessageGrabber(imap4.IMAP4Client):

    allowInsecureLogin = False

    ###
    ### Begin Monkey Patching
    ###
    def rawDataReceived(self, data):
        self._pendingSize -= len(data)
        if self._pendingSize > 0:
            self._pendingBuffer.write(data)
        else:
            passon = ''
            if self._pendingSize < 0:
                data, passon = data[:self._pendingSize], data[self._pendingSize:]
            self._pendingBuffer.write(data)
            rest = self._pendingBuffer
            self._pendingBuffer = None
            self._pendingSize = None
            size = rest.tell()
            rest.seek(0, 0)
            self._parts.append('ELIDED')
            self.elided = rest

            # XXX TRANSACT ME together with the transaction in _cbFetch
            def _():
                self.elided.messageDone()
                self.factory.messageReceived(self.elided.mr.message, size)
            self.factory.avatar.transact(_)

            self.setLineMode(passon.lstrip('\r\n'))

    def _setupForLiteral(self, rest, octets):
        self._pendingBuffer = self.messageFile(octets)
        self._pendingSize = octets

        # Chop of the {..} bit
        rest = rest[:rest.find('{')]

        self._parts = [rest, '\r\n']
        self.setRawMode()

    def _IMAP4Client__cbFetch(self, (lines, last)):
        flags = {}
        for line in lines:
            parts = line.split(None, 2)
            if len(parts) == 3:
                if parts[1] == 'FETCH':
                    try:
                        id = int(parts[0])
                    except ValueError:
                        raise imap4.IllegalServerResponse, line
                    else:
                        data = imap4.parseNestedParens(parts[2])
                        while len(data) == 1 and isinstance(data, types.ListType):
                            data = data[0]
                        while data:
                            if len(data) < 2:
                                raise imap4.IllegalServerResponse("Not enough arguments", data)
                            if data[1] == 'ELIDED':
                                data[1] = self.elided
                                self.elided = None
                            flags.setdefault(id, {})[data[0]] = data[1]
                            del data[:2]
        return flags

    ###
    ### End Monkey Patching
    ###

    ###
    ### Callbacks
    ###

    def messageFile(self, octets):
        mr = mimemessage.IMIMEDelivery(self.factory.avatar).createMIMEReceiver(True)
        f = FileWrapper(mr, self.stats)
        self.onMessage += 1
        self.setFetching()
        return f

    def serverGreeting(self, caps):
        self.factory.setStatus(LoggingIn())
        self.stats = stats.Statoscope(name="imapgrabber")
        self.registerAuthenticator(imap4.CramMD5ClientAuthenticator(self.factory.username))
        self.registerAuthenticator(imap4.LOGINAuthenticator(self.factory.username))
        d = self.authenticate(self.factory.password)
        d.addCallbacks(self._cbLogin, self._ebLogin).addErrback(self._ebDoDisconnect)

        # reactor.callLater(60, self.nooploop)

    def newMessages(self, exists, recent):
        if self.factory.paused:
            return

        if recent:
            if self.lastUid is None:
                msgs = imap4.MessageSet(1, None)
            else:
                msgs = imap4.MessageSet(self.lastUid, None)
            d = self.fetchUID(msgs, uid=True)
            d.addCallbacks(self._cbFetchUids, self._ebDoDisconnect)

    ###
    ### End callbacks
    ###

    def _ebDoDisconnect(self, failure):
        log.err(failure)
        self.factory.setStatus(AuthFailed(str(failure.value)))
        self.logout().addBoth(lambda result: self.transport.loseConnection())

    def _cbLogin(self, result):
        self.factory.setStatus(SelectingMailbox())
        self.select(self.factory.mailbox).addCallbacks(self._cbSelect, self._ebSelect)

    def _ebLogin(self, failure):
        failure.trap(imap4.NoSupportedAuthentication)
        if self.startedTLS or self.allowInsecureLogin or self.transport.getHost()[0] == 'SSL':
            return self.login(self.factory.username, self.factory.password
                ).addCallback(self._cbLogin
                )
        else:
            self.factory.setStatus(AuthFailed("No secure authentication available"))
            self.factory.paused = True
            self.logout().addErrback(log.msg)

    def _ebSelect(self, failure):
        self.examine(self.factory.mailbox).addCallbacks(self._cbSelect, self._ebDoDisconnect)

    def _cbSelect(self, details):
        self.factory.setStatus(CheckInProgress())
        try:
            uidvalidity = str(details['UIDVALIDITY'])
        except KeyError:
            # This is totally bunk, the server is probably broken
            self.factory.setStatus(ExamineFailed("Server is probably broken - missing UIDVALIDITY"))
            self.logout().addBoth(lambda result: self.transport.loseConnection())
        else:
            if not self.factory.uidvalidity:
                self.factory.setUIDValidity(uidvalidity)      # First time we've checked this box
            elif self.factory.uidvalidity != uidvalidity:
                self.factory.setStatus(RetrieveFailed("POD PEOPLE: Mailbox was replaced"))
                self.logout().addBoth(lambda result: self.transport.loseConnection())
                return

            if details.get('EXISTS', 0):
                # Fetch everything newer than the last UID we saw.
                if self.lastUid is None:
                    msgs = imap4.MessageSet(1, None)
                else:
                    msgs = imap4.MessageSet(self.lastUid, None)
                d = self.fetchUID(msgs, uid=True)
                d.addCallbacks(self._cbFetchUids, self._ebDoDisconnect)
            else:
                self.factory.setStatus(CheckSuccessful(0))

    def _cbFetchUids(self, flags):
        v = [int(v['UID']) for v in flags.itervalues()]
        low = min(v)
        high = max(v)
        if self.maxMsgs is not None:
            high = min(high, self.maxMsgs)
        if high <= self.lastUid:
            self.factory.setStatus(CheckingPaused("No new messages"))
        else:
            if self.lastUid is None:
                msgs = imap4.MessageSet(low, high)
            else:
                msgs = imap4.MessageSet(max(low, self.lastUid + 1), high)
            self.onMessage = -1
            self.totMessages = len([x for x in flags.itervalues() if int(x['UID']) in msgs])
            self.setFetching()
            d = self.fetchMessage(msgs, uid=True)
            d.addCallbacks(self._cbFetch, self._ebDoDisconnect)

    def setFetching(self):
        self.factory.setStatus(FetchingMessage(self.onMessage, self.totMessages))

    def _cbFetch(self, messages):
        filterUID = self.lastUid
        def _():
            # XXX TRANSACT ME with the transaction in rawDataReceived
            for result in messages.values():
                if int(result['UID']) > self.lastUid:
                    self.factory.setLastUid(int(result['UID']))
        self.factory.avatar.transact(_)

        self.factory.setStatus(CheckSuccessful(len(messages)), True)
        self.cleanupOldMessages()

    def cleanupOldMessages(self):
        if self.factory.age is not None:
            # Retrieve the dates of all the messages in this folder
            self.fetchInternalDate('1:*', True
                                   ).addCallbacks(self._cbCleanup, self._ebCleanup
                                                  )
        else:
            self.logout()

    def _ebCleanup(self, failure):
        log.err(failure)
        return self.logout()

    def _cbCleanup(self, result):
        delete = self.filterOld(result, self.factory.age)
        return self.setFlags(','.join(map(str, delete)), ['\\Deleted'], uid=True
                      ).addCallback(self._cbDelete, self._ebDelete
                                    )

    def _ebDelete(self, failure):
        self.log(failure)
        return self.logout()

    def _cbDelete(self, result):
        return self.logout()

    def filterOld(self, result, age, now=None):
        if now is None:
            now = datetime.date.today()
        delete = []
        for (uid, date) in result.iteritems():
            try:
                day, month, year = date.split(None, 1)[0].split('-', 2)
                day = int(day)
                year = int(year)
                month = MONTHS.index(month)
                date = datetime.date(year, month + 1, day)
            except:
                log.msg("Junk response from FETCH INTERNALDATE query")
                log.err()
            else:
                if now.toordinal() - date.toordinal() > age:
                    delete.append(uid)
        return delete

    _pendingCall = None
    def nooploop(self, call=None):
        def setNoop(call):
            self._pendingCall = call
        self.noop(
            ).addCallback(lambda _: reactor.callLater(60, self.nooploop)
            ).addCallback(setNoop
            )

    def connectionLost(self, reason):
        if self._pendingCall and self._pendingCall.active():
            self._pendingCall.cancel()
        self._pendingCall = None


def check(self):
    """Check

    Check mail right now.
    """
    pass
check = formless.autocallable(check)


username = formless.String(label="Username", description="Your Username.")
password = formless.Password(label="Password", description="Your Password.")
# Don't actually check to see if the host exists until freeform supports deferreds
#host = formless.Hostname(label="Host", description="The address of the Host, eg mail.mac.com")
host = formless.String(label="Host", description="The address of the Host, eg mail.mac.com")
port = formless.Integer(label="Port", description="The Port to connect to")

msgsPer = formless.Integer(
    label = "Messages Per Grab",
    description = "Maximum number of messages to retrieve per connection (lets you train SpamBayes a little so your messages are classified properly when they do arrive). Set to 0 to grab all messages from the server at once.",
    default = 0,
)
paused = formless.Boolean(
    label = "Pause Grabber",
    description = "Suspend checking of this account",
)

age = formless.Integer(
    label = "Deletion Age",
    description = "Minimum age of messages to delete from this account's server.  Set to 0 to disable this feature.",
    default = 0)

popport = formless.Integer(label="Port", description="The POP3 Port to connect to", default=110)
imapport = formless.Integer(label="Port", description="The IMAP Port to connect to", default=143)

ssl = formless.Boolean(
    label = "SSL",
    description = "Use a secure (encrypted) socket layer for connections",
)

address = formless.String(
    label="Email Address",
    description = "The email address associated with this account"
)

class IGrabberFactory(formless.TypedInterface):
    check = check
    password = password
    class GrabberAttributes(formless.TypedInterface):
        """A grouping of all the grabber attributes, so that
        they can all be changed in one transaction.

        Except for password. That has to be changed
        in a different way.
        """
        interval = formless.Integer(
            label = "Interval",
            description = "The amount of time to wait between checks for new mail.")
        username = username
        host = host
        port = port
        address = address
        ssl = ssl
        msgsPer = msgsPer
        paused = paused


class GrabberFactory(store.Item, protocol.ClientFactory, sched.Schedulable, object):
    __implements__ = IGrabberFactory, sched.Schedulable.__implements__,

    persistenceForgets = ['currentProtocol', 'doneDeferred', 'successfulCheck']

    _lastGrabTime = 0
    interval = 60 * 5

    age = None

    debug = False

    _paused = False

    def __init__(self, avatar, username, password, host, port, address, msgsPer=0, paused=False, interval=None, ssl=False, age=None):
        """
        @param username: The username with which to log into this account.
        @param password: The password with which to authenticate for this account.

        @param host: The hostname of the server to which to connect.
        @param port: The port number of the server to which to connect.

        @param address: The email address to associated with messages received from this account.

        @param msgsPer: The maximum number of messages to retrieve per connection.
        @param interval: The number of seconds between checks.
        @param ssl: Indicate whether or not to use SSL for this connection.
        @param age: The age, in days, of messages to delete from the server.
        """
        store.Item.__init__(self, avatar)
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.address = address
        if interval is not None:
            self.interval = interval
        self.msgsPer = msgsPer
        self.paused = paused
        self.ssl = ssl
        self.age = age
        self.setStatus("Initializing...")

    def __repr__(self):
        return "%s://%s@%s" % (self.protoName, self.username, self.host)

    def delete(self):
        IGrabberAdder(self.store).delete(self)

    def ssl():
        def get(self):
            return self._ssl
        def set(self, value):
            self.touch()
            self._ssl = value
            if self._ssl:
                if self.port == self.DEFAULT_PORT:
                    self.port = self.SSL_PORT
                if 'protoName' not in vars(self):
                    self.protoName = self.protoName + 's'
            else:
                if self.port == self.SSL_PORT:
                    self.port = self.DEFAULT_PORT
                if 'protoName' in vars(self):
                    del self.protoName
        return get, set
    ssl = property(*ssl())

    def paused():
        def get(self):
            if self._paused:
                return True
            cp = sysup.IControlPanel(self.store.store)
            perm = sysup.IControlPanel.EXTERNAL_GRABBING
            return not cp.actionAllowed(perm)
        def set(self, value):
            def setPaused():
                self.touch()
                self._paused = value
            self.store.transact(setPaused)
        return get, set
    paused = property(*paused())

    def avatar():
        def _getAvatar(self):
            return self.store
        def _setAvatar(self):
            raise NotImplementedError("WHAT THE HECK, YO")
        return _getAvatar, _setAvatar
    avatar = property(*avatar())

    def createMIMEReceiver(self, trustDate=False):
        return mimemessage.IMIMEDelivery(self.store).createMIMEReceiver(trustDate)

    doneDeferred = None
    currentProtocol = None
    lastUid = None
    successfulCheck = False
    statusSet = False

    def setLastUid(self, lastUid):
        """Call me from within a transaction.
        """
        def _():
            self.touch()
            self.lastUid = lastUid
        self.avatar.transact(_)

    def setStatus(self, stat, success=False):
        def _():
            # TODO: move this elsewhere
            self.status = stat
            self.statusSet = True
            self.successfulCheck = success
        self.avatar.transact(_)

    def messageReceived(self, message, bytes):
        self.touch()

        # self.avatar.realm.service.stats.processedBytes += bytes
        # self.avatar.realm.service.stats.receivedMessages += 1

        args = (self.protoName, self.username, self.host, self.port)
        message.setMessageOrigin('%s://%s@%s:%s/' % args)
        message.setMessageAddress(self.address or ('@'.join((self.username, self.host))))

        try:
            provisioner.IProvisioner(self.avatar).deplete(self.tankID, bytes)
        except provisioner.NotEnoughFuel:
            self.paused = True
            self.setStatus(CheckingPaused("Quota exceeded"))

    def startedConnecting(self, connector):
        log.msg('Started %s Connecting' % (self.__class__.__name__,))
        args = (self.protoName, self.username, self.host, self.port)
        log.msg('%s://%s@%s:%s/' % args)

    def clientConnectionFailed(self, connector, reason):
        def _():
            self.reset(self.interval)
        self.avatar.transact(_)

        log.err(reason)
        self.currentProtocol = None
        self.setStatus(CheckingPaused(str(reason.value)))
        if self.doneDeferred:
            self.doneDeferred.errback(reason)
            self.doneDeferred = None

    def clientConnectionLost(self, connector, reason):
        self.currentProtocol = None
        if not self.successfulCheck and not self.statusSet:
            self.setStatus(CheckingPaused(str(reason.value)))
        if self.doneDeferred:
            self.doneDeferred.callback(None)
            self.doneDeferred = None

    def buildProtocol(self, addr):
        assert not self.currentProtocol, \
               "GrabberFactory can manage only one connection at a time"
        p = self.protocol()
        p.factory = self
        p.lastUid = self.lastUid
        p.maxMsgs = self.msgsPer or None
        self.currentProtocol = p
        self.successfulCheck = False
        self.statusSet = False
        return self.debug and tpython.DebugShimTransport(p) or p

    _webUI_nextRun = 0
    _webUI_forceCount = 0

    def check(self):
        self.touch()

        # This restriction is a hack to be replaced at a later date
        # with a real provisioning-based solution.
        if self._webUI_nextRun < time.time():
            self._webUI_forceCount += 1
            self._webUI_nextRun = time.time() + (2 ** self._webUI_forceCount)
            self.run()
            return "Checking"
        else:
            return "Working!  Be patient."

    def run(self):
        """This function must be run in a storage context.
        """
        self.reset(self.interval)

        if self.paused or self.doneDeferred is not None:
            return self.doneDeferred

        self.doneDeferred = defer.Deferred()
        if self.ssl:
            if not sslModule.supported:
                log.err(failure.Failure(Exception("SSL connection requested, but SSL not available")))
                return None
            else:
                reactor.connectSSL(self.host, self.port, self, sslModule.ClientContextFactory())
        else:
            reactor.connectTCP(self.host, self.port, self)

        return self.doneDeferred

    def format(self):
        """Return a summary of this grabber as a nicely formatted string.
        """
        ## This should be removed and replaced with the IPreview adapter for grabber factories.
        return "%s@%s" % (self.username, self.host)


class POP3GrabberFactory(GrabberFactory):
    protoName = 'pop3'
    protocol = POP3MessageGrabber
    status = NeverChecked()

    DEFAULT_PORT = 110
    SSL_PORT = 995

    def __init__(self, *a, **kw):
        GrabberFactory.__init__(self, *a, **kw)
        p = store.Pool(self.store)
        p.addIndex(store.DateIndex())
        self.poolRef = p.referenceTo()

    def messageReceived(self, message, bytes, uid):
        GrabberFactory.messageReceived(self, message, bytes)
        if self.age:
            message.touch()
            message.pop3uid = uid
            self.poolRef.getItem().addItem(message)

    def getDeletionList(self, now=None):
        """Returns a list of UIDs which should be deleted

        UIDs returned will have their corresponding messages
        removed from the pop3-uid pool.
        """
        if now is None:
            now = time.gmtime()
        if not self.age:
            return []
        p = self.poolRef.getItem()
        r = []
        then = time.mktime(now) - (60 * 60 * 24 * self.age)
        def _():
            for m in p.queryIndex('received', endKey=then):
                r.append(m.pop3uid)
                p.removeItem(m)
            return r
        return self.avatar.transact(_)

class IMAP4GrabberFactory(GrabberFactory):
    protoName = 'imap4'

    # To be subsumed by, say, IMAP4MessageSynchronizer
    protocol = IMAP4MessageGrabber
    status = NeverChecked()

    DEFAULT_PORT = 143
    SSL_PORT = 993

    mailbox = 'inbox'
    uidvalidity = None

    def setUIDValidity(self, validity):
        def _():
            self.touch()
            self.uidvalidity = validity
        self.avatar.transact(_)

class IGrabber(Interface):
    def format(self):
        pass

    def startChecking(self, deliver):
        pass

    def stopChecking(self):
        pass

    def setStatus(self, status):
        pass

    def update(self, **args):
        pass

    def getInitSignature(self):
        pass

IGrabberObject = formless.Object(
    IGrabber,
    label="Grabber",
    description="A grabber instance.")

class IGrabberAdder(formless.TypedInterface):
    def importDirectory(self,
        dirname = formless.Directory()):
        """Import Directory

        Import a directory full of messages on the server into this account.
        """
        return None
    importDirectory = formless.autocallable(importDirectory, invisible=True)

    def addPopGrabber(self,
        username = username, password = password,
        host = host, port = popport, address=address, msgsPer = msgsPer,
        ssl = ssl, age = age):
        """Add Pop Grabber

        Add a POP3 email account to be checked by this Quotient account.
        """
        return IGrabberObject
    addPopGrabber = formless.autocallable(addPopGrabber)

    def addIMAPGrabber(self,
        username = username, password = password,
        host = host, port = imapport, address=address, msgsPer = msgsPer,
        ssl = ssl, age = age):
        """Add IMAP Grabber

        Add an IMAP email account to be checked by this Quotient account.
        """
        return IGrabberObject
    addIMAPGrabber = formless.autocallable(addIMAPGrabber)

    def delete(self, grabbers=formless.List()):
        """Delete a list of grabbers
        """
        return None

    grabbers = formless.List(actions=[delete], label="Existing accounts")


class GrabberPowerup(provisioner.ProvisionedPowerup):
    __implements__ = (IGrabberAdder, provisioner.IProvisionable, provisioner.ProvisionedPowerup.__implements__)
    dependencies = [mimemessage.EmailPowerup, sched.SubScheduler]
    storeDependencies = [sysup.ControlPanel]
    powerupName = "External Accounts"

    tankID = provisioner.ETank.INCOMING_MAIL
    startCost = 1

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        g = store.Pool(self.store, 'grabbers')
        self.grabbersRef = g.referenceTo()
        avatar.setComponent(IGrabberAdder, self.fwd())

    def delete(self, grabbers):
        for grabber in grabbers:
            grabber.paused = True
            grabber.unschedule()
            self.grabbersRef.getItem().removeItem(grabber)

    def getGrabbers(self):
        return list(self.grabbersRef.getItem().queryIndex())
    grabbers = property(getGrabbers)

    def addGrabber(self, grabberFactory, username, password, host, port, address, msgsPer, ssl, age):
        gf = grabberFactory(self.store, username, password, host, port, address, msgsPer, ssl=ssl, age=age)
        gf.reset(1)
        self.grabbersRef.getItem().addItem(gf)
        gf.tankID = self.tankID
        return gf

    def addPopGrabber(self, username, password, host, port, address, msgsPer, ssl, age):
        return self.addGrabber(POP3GrabberFactory, username, password, host, port, address, msgsPer, ssl, age)

    def addIMAPGrabber(self, username, password, host, port, address, msgsPer, ssl, age):
        return self.addGrabber(IMAP4GrabberFactory, username, password, host, port, address, msgsPer, ssl, age)

    def importDirectory(self, dirname):
        from quotient import tap2
        tap2.importmessages(self.store, [dirname], [], [], [])
