# -*- test-case-name: quotient.test.test_mimebakery -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import quopri
import time
import rfc822
from email.Generator import _make_boundary

from twisted.protocols.smtp import messageid

from quotient.copyright import version
from quotient import prefup
from quotient import mimemessage

def _addrtup2str(t):
    if isinstance(t, str):
        display = None
        email = ''
    else:
        display, email = t
    if display:
        tstr = '"%s" <%s>' % (display, email)
    else:
        tstr = "<%s>" % email
    return tstr

def _addrlist2str(l):
    return ', '.join(map(_addrtup2str, l))

def _fmthdrs(part, io, d):
    part.headersOffset = io.tell()
    for k, v in d.iteritems():
        part[k] = v
        io.write("%s: %s\n" % (k, v.replace('\n', '\n\t')))
    part.headersLength = part.headersOffset - io.tell()
    io.write('\n')
    part.bodyOffset = io.tell()

def _closepart(part, io):
    part.bodyLength = io.tell() - part.bodyOffset

class Recipe:

    """A simple wrapper API for creating MIME messages based on simple
    requirements, like attachments and text parts.

    TODO: multipart-related, content-disposition: inline, nevow integration
    """

    def __init__(self, avatar, to,
                 sender=None, cc=None, bcc=None, subject='(No Subject)'):
        """
        @param avatar: an avatar.

        @param to: a list of 2-tuples or strings a-la the sender argument.

        @param sender: a 2-tuple (displayname, email) or string (email) which
        represents the originator of the message.  If left blank, this will
        default to the avatar's default name and address.

        @param cc: similar to 'to'

        @param bcc: similar to 'to'.

        @param subject: a string to go in the subject header.
        """
        self.avatar = avatar
        if sender is None:
            sender = prefup.IPrefs(avatar).replyto
        senderstr = _addrtup2str(sender)
        self.protoheaders = {
            'Sender': senderstr,
            'From': senderstr,
            'Subject': subject,
            'Date': rfc822.formatdate(),
            'MIME-Version': '1.0',
            'X-Mailer': 'Divmod-Quotient/%s' % version
            }

        for (h, v) in (('To', to), ('Cc', cc)):
            if v:
                self.protoheaders[h] = _addrlist2str(to)
        self.bcc = bcc
        self.alternatives = []
        self.attachments = []

    def text(self, data, subtype='plain'):
        try:
            data = data.encode('ascii')
            encoding = '7bit'
        except UnicodeDecodeError:
            data = quopri.encodestring(data)
            encoding = 'quoted-printable'
        self.alternatives.append((data, subtype, encoding))

    def attach(self, data,
               mimetype="application/octet-stream",
               filename="unknown.bin"):
        self.attachments.append((data.encode('base64'), mimetype, filename))

    def _mainpart(self, part, io, partheaders):
        if len(self.alternatives) == 0:
            data, subtype, encoding = '', 'plain', '7bit'
        elif len(self.alternatives) == 1:
            data, subtype, encoding = self.alternatives[0]
        else:
            raise NotImplementedError("No multipart/alternative support yet.")
##             boundary = _make_boundary()
##             partheaders['Content-type'] = 'multipart/alternative; boundary="%s"' % boundary
        partheaders['Content-type'] = 'text/'+subtype
        partheaders['Content-transfer-encoding'] = encoding
        _fmthdrs(part, io, partheaders)
        io.write(data)
        _closepart(part, io)

    def bake(self):
        io = self.avatar.newFile()
        root = mimemessage.MIMEMessage()
        rootheaders = self.protoheaders.copy()
        
        # Not sure if this is the best place for this.
        # Perhaps it should go in MIMEMessage?
        root.filingInfo = (None, 'Me', None)

        # Copying the should be safe to call 'bake' multiple times, but perhaps this
        # should be refactored to take avatar as an arg to bake rather than
        # __init__ if we actually need to do so?

        rootheaders['Message-ID'] = messageid(uniq="divmod quotient")
        root.bcc = self.bcc
        if self.attachments:
            alltext = '\n'.join([self.alternatives[0][0]] + [data for data, mimetype, filename in self.attachments])
            boundary = _make_boundary(alltext)
            rootheaders['Content-type'] = 'multipart/mixed; boundary="%s"' % boundary
            _fmthdrs(root, io, rootheaders)
            nbound = '\n--%s\n' % boundary
            fbound = '\n--%s--\n' % boundary
            io.write(nbound)
            self._mainpart(root.newChild(), io, {})
            for data, mimetype, filename in self.attachments:
                io.write(nbound)
                attpart = root.newChild()
                _fmthdrs(attpart, io, {'Content-type': '%s' % mimetype,
                                       'Content-disposition': 'attachment; filename="%s"' % filename,
                                       'Content-transfer-encoding': 'base64'})
                io.write(data)
                _closepart(attpart, io)
            io.write(fbound)
            _closepart(root, io)
        else:
            self._mainpart(root, io, rootheaders)
        gmtDate = time.gmtime()

        # the following lines are probably a method in mimemessage, since they
        # always have to happen at once.

        root.size = io.tell()
        root.addToStore(self.avatar)
        root.filename = io.close(root.setReceived(gmtDate))

        return root
