# -*- test-case-name: quotient.test.test_mimemessage -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import re
import quopri
import base64
import rfc822
import binascii
import time

from cStringIO import StringIO

from twisted.internet.error import ConnectionDone
from twisted.persisted.styles import Versioned
from twisted.python import components
from twisted.python.components import Interface
from twisted.python.failure import Failure
from twisted.python import log

from atop.tpython import iterateInReactor, Forgetter
from atop.store import Item, Pool, HardPoint
from atop.store import DateIndex, NameIndex, IPoolIndex, INDEX
from atop.store import IPipe
from atop.filepile import symlink

from quotient.stats import Statoscope
from quotient import provisioner
#from quotient.searchup import IIndexable

def unquote(st):
    if len(st) > 1:
        if st[0] == st[-1] == '"':
            return st[1:-1].replace('\\\\', '\\').replace('\\"', '"')
        if st.startswith('<') and st.endswith('>'):
            return st[1:-1]
    return st


def headerParams(headerstr):
    """
    @param header: a str containing a header in the format foo; bar=\"baz\",
    etc.
    @return: 2-tuple of str(header-value), dict(header-params)
    """
    typeinfo = headerstr.split(';')
    headerval = typeinfo[0].strip().lower()
    params = {}
    for t in typeinfo[1:]:
        kv = t.split('=', 1)
        if len(kv) == 2:
            k = kv[0].strip().lower()
            v = kv[1].strip().strip('"')
            params[k] = v
    return headerval, params


class HeaderBodyParser:
    def __init__(self, part, parent):
        self.parent = parent
        self.parsingHeaders = 1
        self.prevheader = None
        self.prevvalue = None
        self.warnings = []
        self.part = part
        self.bodyMode = 'body'
        self.gotFirstHeader = False

    def close(self):
        if self.parent:
            self.parent.close()

    def startBody(self, linebegin, lineend):
        self.parsingHeaders = 0
        self.part.headersLength = linebegin - self.part.headersOffset
        self.part.bodyOffset = lineend

    def lineReceived(self, line, linebegin, lineend):
        if self.parsingHeaders:
            if not self.gotFirstHeader:
                self.part.headersOffset = linebegin
                self.gotFirstHeader = True
            return self.parseHeaders(line, linebegin, lineend)
        else:
            return self.parseBody(line, linebegin, lineend)

    def warn(self, text):
        self.warnings.append(text)

    def finishHeader(self):
        if self.prevheader is not None:
            self.part[self.prevheader] = self.prevvalue
        self.prevheader = self.prevvalue = None

    def parseHeaders(self, line, linebegin, lineend, hdrValueDelim=re.compile(":[ \t]")):
        if not line:
            self.finishHeader()
            self.startBody(linebegin, lineend)
            return self
        if line[0] in ' \t':
            self.prevvalue += '\n' + line
            return self
        h = hdrValueDelim.split(line, 1)
        if len(h) == 2:
            self.finishHeader()
            header, value = h
            self.prevheader = header
            self.prevvalue = value
        elif line and line[-1] == ':':
            # is this even a warning case?  need to read the rfc... -glyph
            self.prevheader = line[:-1]
            self.prevvalue = ''
        else:
            self.warn("perhaps a body line?: %r" % line)
            self.finishHeader()
            self.startBody(linebegin, lineend)
            self.lineReceived(line, linebegin, lineend)
        return self

    def parseBody(self, line, linebegin, lineend):
        return getattr(self, "parse_" + self.bodyMode)(line, linebegin, lineend)

class MIMEMessageParser(HeaderBodyParser):
    bodyFile = None
    def startBody(self, linebegin, lineend):
        HeaderBodyParser.startBody(self, linebegin, lineend)
        self.boundary = self._calcBoundary()
        if self.boundary:
            self.finalBoundary = self.boundary + '--'
            self.bodyMode = 'preamble'
            return
        ctyp = self.part['content-type']
        if ctyp and ctyp.split()[0].strip().lower() == 'message/rfc822':
            self.bodyMode = 'rfc822'
            return
        self.bodyMode = 'body'
        # self.bodyFile = self.part.getBody("wb")
        # ^ was only used for on-the-fly decoding

    def close(self):
        if self.bodyFile:
            self.bodyFile.close()
        HeaderBodyParser.close(self)

    def _calcBoundary(self):
        ctype = self.part['content-type']
        if ctype and ctype.strip().lower().startswith('multipart'):
            parts = ctype.split(';')
            for part in parts:
                ps = part.split('=', 1)
                if len(ps) == 2:
                    key, val = ps
                    key = key.strip().lower()
                    if key.lower() == 'boundary':
                        return '--' + unquote(val.strip())
            return None
        else:
            return None

    def parse_body(self, line, b, e):
        # TODO: on-the-fly decoding
        return self

    def parse_rfc822(self, line, b, e):
        np = self.subpart(parent=self, factory=MIMEMessageParser)
        np.lineReceived(line, b, e)
        return np

    def subpart(self, parent=None, factory=None):
        if parent is None:
            parent = self
        if factory is None:
            factory = MIMEPartParser
        newpart = self.part.newChild()
        nmp = factory(newpart, parent)
        return nmp

    def parse_preamble(self, line, b, e):
        if line.strip('\r\n') == self.boundary:
            self.bodyMode = 'nextpart'
            return self.subpart()
        return self

    def parse_nextpart(self, line, b, e):
        if line.strip('\r\n') == self.boundary:
            # If it's a boundary here, that means that we've seen TWO
            # boundaries, one right after another!  I can only assume that the
            # sub-human cretins who have thusly encoded their MIME parts are
            # attempting to convey the idea that the message *really* has a
            # part-break there...
            return self
        nmp = self.subpart()
        nmp.lineReceived(line, b, e)
        return nmp

    def parse_postamble(self, line, b, e):
        return self

class MIMEPartParser(MIMEMessageParser):
    def parseBody(self, line, linebegin, lineend):
        if line.strip('\r\n') == self.parent.boundary:
            # my body is over now - this is a boundary line so don't count it
            self.part.bodyLength = linebegin - self.part.bodyOffset
            return self.parent
        elif line == self.parent.finalBoundary:
            self.parent.bodyMode = 'postamble'
            self.part.bodyLength = linebegin - self.part.bodyOffset
            return self.parent
        else:
            return MIMEMessageParser.parseBody(self, line, linebegin, lineend)

    def parse_rfc822(self, line, linebegin, lineend):
        np = self.subpart(parent=self.parent)
        np.lineReceived(line, linebegin, lineend)
        return np

class MIMEPart:
    def __init__(self, parent=None):
        self.parent = parent
        self.children = []
        self.headers = []

    # email.Message compat: note non-coding-standard-compliant method names

    def walk(self):
        yield self
        for child in self.children:
            for part in child.walk():
                yield part

    def get_all(self, field, failObj):
        return self.get(field, failObj)

    def get_filename(self, failObj=None):
        return self.get_param('filename', failObj, 'content-disposition')

    def get_param(self, param, failObj=None, header='content-type', un_quote=True):
        h = self[header]
        if not h:
            return failObj
        param = param.lower()
        for pair in [x.split('=', 1) for x in h.split(';')[1:]]:
            if pair[0].strip().lower() == param:
                r = len(pair) == 2 and pair[1].strip() or ''
                if un_quote:
                    return unquote(r)
                return r
        return failObj

    def newChild(self):
        c = MIMEPart(self)
        self.children.append(c)
        return c

    # email.Message compat

    def __setitem__(self, key, val):
        self.headers.append((key, val))

    def __getitem__(self, key, failobj=None):
        for k,v in self.headers:
            if key.lower() == k.lower():
                return v
        return failobj
    get = __getitem__

    def __contains__(self, name):
        return not not self.get(name)

    def has_key(self, name):
        return name in self

    def items(self):
        return self.headers

    def get_charset(self):
        return None

    def get_type(self, failobj=None):
        return self.get('content-type', failobj)

    def get_payload(self, decode=False):
        """Get the message payload.
        """
        f = self.openFile()
        offt = self.bodyOffset
        leng = self.bodyLength
        f.seek(offt)
        data = f.read(leng)
        if decode:
            ctran = self['content-transfer-encoding']
            if ctran:
                ct = ctran.lower().strip()
                if ct == 'quoted-printable':
                    return quopri.decodestring(data)
                elif ct == 'base64':
                    try:
                        return base64.decodestring(data)
                    except binascii.Error:
                        return data
                elif ct == '7bit':
                    return data
        return data

    def _uberparent(self):
        o = self
        while o.parent:
            o = o.parent
        return o

    def openFile(self):
        return open(self._uberparent().filename, 'rb')

    def get_default_type(self):
        return 'text/plain'

    def get_content_type(self):
        missing = object()
        value = self.get('content-type', missing)
        if value is missing:
            return self.get_default_type()
        ctype = value.split(';', 1)[0].lower().strip()
        if ctype.count('/') != 1:
            return 'text/plain'
        return ctype

    def get_content_maintype(self):
        ctype = self.get_content_type()
        return ctype.split('/')[0]

    def get_content_subtype(self):
        ctype = self.get_content_type()
        return ctype.split('/')[1]

    def get_main_type(self, failobj=None):
        """Return the message's main content type if present."""
        missing = object()
        ctype = self.get_type(missing)
        if ctype is missing:
            return failobj
        if ctype.count('/') != 1:
            return failobj
        return ctype.split('/')[0]

    def is_multipart(self):
        return bool(self.children)

    def getdate(self, name):
        data = self.get(name)
        if data:
            return rfc822.parsedate(data)

    def getHeaderParams(self, hdrname):
        ctype = self[hdrname]
        if ctype is None:
            return {}
        return headerParams(ctype)[1]

    def getAttachmentName(self):
        params = self.getHeaderParams("content-disposition")
        for fnk in 'name', 'filename':
            if params.has_key(fnk):
                return params[fnk]
        else:
            gtl = self.get_type().split(';')[0].lower()
            ext = {'text/html':  'html',
                   'text/plain': 'plain',
                   'image/jpeg': 'jpeg',
                   'image/png':  'png',
                   'image/gif':  'gif'}.get(gtl, 'bin')
            return 'Unknown.'+ext

    def inferType(self):
        """Infer a content-type.  This will attempt to do something with
        garbage data that isn't properly typed.
        """
        ctype = self['content-type']
        if not ctype:
            return 'text/plain'
        if ctype.lower().startswith("application/octet-stream"):
            ext = self.getAttachmentName().strip().split(".")[-1].lower()
            if exts.has_key(ext):
                return exts[ext]
        return ctype

    def getTypedParts(self, *types):
        for part in self.walk():
            # possible change: rather than get_content_type, use inferType to
            # catch parts which are malformed MIME-ly but still valid data.
            if part.get_content_type() in types:
                yield part

    def getAttachments(self):
        for part in self.walk():
            ctyp = part.get_content_type()
            if ((not (ctyp.startswith('text') or ctyp.startswith('multipart'))) or
                ((part['content-disposition'] or '').startswith('attachment'))):
                yield part

    def keys(self):
        return [k for k, v in self.headers]

    # STUBBED METHODS: these will prevent spambayes et. al. from raising
    # exceptions, but we should look into how far we want to support them.

    def __delitem__(self, thing):
        pass

    def add_header(self, header, value):
        pass

    def get_charsets(self, failObj=None):
        return []

    # end stubbed methods


_flag = (1).__lshift__

class MIMEMessage(MIMEPart, Item, Versioned):
    #__implements__ = ISearchable

    # Yay flags
    # For now, when you add soemthing here, add it to imapup.py:173 too
    SEEN, ANSWERED, FORWARDED, REDIRECTED, DELETED = map(_flag, range(5))

    # Flags set on this message
    flags = 0

    parent = None
    smtpInfo = None

    # A string describing how this message came to us
    receivedVia = None

    addressedTo = None

    # Reference to the contact who sent this message
    contactRef = None

    # The manner in which this message arrived at its present location.
    # This is a list of something like ["todo"] or ["38721", "blog"]
    route = None

    # Don't call Item.__init__ - we don't want to initialize the item part of
    # ourselves until we're filled out enough to exist in the database.

    def assignIDs(self):
        mimeID = 0
        for part in self.walk():
            part.mimeID = mimeID
            mimeID += 1

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.assignIDs()

    def getPartByID(self, mimeID):
        w = self.walk()
        c = 0
        for p in w:
            if mimeID == c:
                return p
            c += 1

    def addToStore(self, store):
        self.assignIDs()
        Item.__init__(self, store)

    transmissionInfo = None

    def markForTransmission(self, from_, localpart, domain):
        self.touch()
        self.transmissionAttempts = 0
        self.transmissionInfo = (from_, localpart, domain)

    def changeFlag(self, flag, set=True):
        if set:
            newFlags = self.flags | flag
        else:
            newFlags = self.flags & ~flag
        if newFlags != self.flags:
            self.setIndexKey('flags', newFlags)

    def stringFlags(self):
        result = []
        for fname in ('SEEN', 'ANSWERED', 'FORWARDED', 'REDIRECTED'):
            if self.flags & getattr(self, fname):
                result.append(fname)
        return result

    def setMessageOrigin(self, s):
        """Indicate the source from which this message came to us."""
        self.touch()
        self.receivedVia = s

    def setMessageAddress(self, a):
        self.touch()
        self.addressedTo = a

    def index_sender(self, pool):
        self.nameAndAddress()[0].lower()

    def index_subject(self, pool):
        subj = self['subject'] or '.'
        subj = subj.lower()
        if subj.startswith('re:') or subj.startswith('fw:'):
            subj = subj[len('re:'):]
        elif subj.startswith('fwd:'):
            subj = subj[len('fwd:'):]
        subj = subj.strip()
        return subj

    def index_received(self, pool):
        return self.dateReceived

    def index_pop(self, pool):
        return self.storeID

    def index_thread(self, pool):
        return (self.threadedBeneathRef.storeID, self.storeID)

    def index_flags(self, pool):
        return self.flags

    def setindex_flags(self, newValue):
        self.flags = newValue
        return newValue

    def index_size(self, pool):
        return self.size

    def getDisplayPart(self):
        # TODO: StopIteration is a lame exception, but what else should I do
        # here?
        supported = ('text/plain', 'text/html', 'text/rtf')
        try:
            return self.getTypedParts(*supported).next()
        except StopIteration:
            if self.get_content_type() == 'multipart/related':
                start = self.get_param('start')
                for p in self.children:
                    if p['content-id'] == start:
                        return p
                try:
                    return self.children[0]
                except IndexError:
                    raise ValueError("Multipart/Related is missing subparts")
            elif self.get_content_type() == 'multipart/alternative':
                for p in self.children:
                    if p['content-type'] in self.supported:
                        return p
        raise ValueError("No displayable parts found")

    def setReceived(self, rdate):
        """Take a date, and return an extra file-path to add to the avatar's filespace to uniquely identify this message.
        """
        self.dateReceived = rdate
        dplist = [str(x) for x in rdate[:3]]    # Y/M/D
        dplist.append(str(self.storeID))
        fn = os.path.join('mail', *dplist)
        return fn

    def nameAndAddress(self):
        """
        Return a name address pair of who this is from
        """
        fheader = self['from'] or 'Nobody'
        try:
            name, address = rfc822.parseaddr(fheader)
            if not name:
                name = address
            return (name, address)
        except:
            return (fheader, fheader)


def messageFromStructure(avatar, headers, parts):
    """Create a MIMEMessage from a structure.

    @param headers: A list of [(string, string)]
    @param parts: A list OR a string.  If it's a list, a list in the form [(part-headers), data]
    XXX ONLY A STRING IS CURRENTLY SUPPORTED
    """
    def _():
        f = avatar.newFile()
        msg = MIMEMessage()
        msg.headersOffset = 0
        # format headers: PUT THIS IN MIMEMessage
        for k, v in headers:
            f.write(k+': '+(v.replace('\n', '\n\t'))+'\n')
        msg.headersLength = f.tell()
        f.write('\n')
        msg.bodyOffset = f.tell()
        msg.headers.extend(headers)
        if isinstance(parts, str):
            f.write(parts)
        else:
            raise NotImplementedError('This interface makes no sense.')
        msg.size = msg.bodyLength = f.tell()
        msg.addToStore(avatar)
        msg.filename = f.close(msg.setReceived(time.gmtime()))
        return msg
    return avatar.transact(_)


# message started - headers begin (begin of line)

# headers ended - headers end (begin of line), body begins (end of line)

# boundary hit - body ends for previous child (begin of line) headers begin for
#      next child (end of line)

# "rfc822-begin" - headers begin for sub-rfc822-message

# subpart headers ended - headers end for child (begin of line), body begins
#      for child (end of line)

# subpart ended - body

# message ended (body ends)

class MIMEMessageReceiver:
    def __init__(self, avatar, deliver, trustDateHeaders=False):
        self.avatar = avatar
        self.deliver = deliver
        self.trustDateHeaders = trustDateHeaders
        self.done = False
        self.lineReceived = self.firstLineReceived

    def makeConnection(self, t):
        # rhg protocol
        self.bytecount = 0
        self.connectionMade()

    def connectionMade(self):
        self.message = MIMEMessage()
        self.file = self.avatar.newFile()
        #         self.message._currentsize = self.file.tell
        # ^ causes problems with pickle, obviously
        self.parser = MIMEMessageParser(self.message, None)

    def firstLineReceived(self, line):
        del self.lineReceived
        if line.startswith('From '):
            return
        return self.lineReceived(line)

    def lineReceived(self, line):
        linebegin = self.bytecount
        self.bytecount += (len(line) + 1)
        lineend = self.bytecount
        self.file.write(line+'\n')
        newParser = self.parser.lineReceived(line, linebegin, lineend)
        oldParser = self.parser
        if newParser is not oldParser:
            self.parser = newParser

    def connectionLost(self, reason):
        if self.done:
            return
        self.file.abort()

    def messageDone(self):
        self.done = True
        localNow = time.time()
        gmtDate = time.gmtime(localNow)
        self.parser.part.bodyLength = (self.bytecount - self.parser.part.bodyOffset)
        if self.trustDateHeaders:
            try:
                rdate = time.struct_time(rfc822.parsedate(self.message['received'].split(';')[-1]))
            except:
                rdate = gmtDate
        else:
            rdate = gmtDate
        self.message['x-divmod-processed'] = rfc822.formatdate(localNow)

        def _():
            self.file.flush()
            size = self.file.tell()
            self.message.size = size
            self.message.addToStore(self.avatar)
            # store/avatarid/Y/M/D/msgid
            fpath = self.message.setReceived(rdate)
            self.message.filename = self.file.close(fpath)
            self.deliver(self.message)
            log.msg(storage='used', type='message-text', amount=size, avatar=self.avatar)
        self.avatar.transact(_)

    # utility methods

    def feedFile(self, f):
        """Feed a file in.
        """
        return iterateInReactor(self._deliverer(f)).addCallback(
            lambda x: self.message)

    def feedString(self, s):
        """Feed a string in.
        """
        return self.feedFile(StringIO(s))

    def feedFileNow(self, f):
        for x in self._deliverer(f):
            pass
        return self.message

    def feedStringNow(self, s):
        return self.feedFileNow(StringIO(s))

    def _deliverer(self, f):
        self.makeConnection(None)
        try:
            while True:
                line = f.readline()
                if not line:
                    break
                line = line.strip('\r\n')
                self.lineReceived(line)
                yield None
        except:
            self.connectionLost(Failure())
            raise
        else:
            self.messageDone()
            self.connectionLost(Failure(ConnectionDone()))

class IMIMEDelivery(components.Interface):
    """I am a MIME delivery object.  I can wrap a storage avatar.
    """

    def createMIMEReceiver(self, trustReceivedHeaders):
        """Create a MIME receiver.  'trustReceivedHeaders' is an option to
        specify the primary date index: if it is True, it will use the last
        'Received' header.  If False, it will use the current time of the
        message's delivery.  This is dependent upon the delivery mechanism.
        For example, SMTP should NOT trustReceivedHeaders, because although the
        message was received from another mail server whose clock is probably
        correct, the message is not *finished* being 'received' until the
        system the user uses to check their mail has got its hands on it (in
        this case, us).  POP3, on the other hand, SHOULD trustReceivedHeaders,
        because the mail hosting system on the other end of the POP connection
        has arguably already added a header as to when it arrived at the
        address represented by the POP account.

        Finally, file imports should always, always trustReceivedHeaders,
        otherwise you will end up with a date index with all the imported
        messages clustered within 5 seconds of each other.
        """

class MIMEDeliverator:
    __implements__ = IMIMEDelivery
    def __init__(self, avatar, arrivalRef):
        self.avatar = avatar
        self.arrivalRef = arrivalRef

    def getArrivalRef(self):
        "get arrival reference"
        return self.arrivalRef

    def createMIMEReceiver(self, trustReceivedHeaders):
        arrivalPool = self.arrivalRef.getItem()
        return MIMEMessageReceiver(self.avatar, arrivalPool.addItem, trustReceivedHeaders)


class FlagsIndex:
    __implements__ = (IPoolIndex,)

    def getIndexName(self):
        return "flags"

    def keyToSortableString(self, key):
        return '%010d' % (key,)

    def sortableStringToKey(self, sstr):
        return int(sstr)

def numberFlagged(pool, flag):
    """Return the number of items with the given flag in the given pool.

    @type pool: C{Pool}
    @param pool: The Pool to query.

    @type flag: C{int}
    @param flag: The flag count.

    @rtype: C{int}
    """
    return len([k for k in pool.queryIndex('flags', result=INDEX) if k & flag])

class SizeIndex:
    __implements__ = (IPoolIndex,)

    def getIndexName(self):
        return "size"

    def keyToSortableString(self, key):
        return '%010d' % (key,)

    def sortableStringToKey(self, sstr):
        return int(sstr)

class ICleanPoint(Interface):
    pass

class CleanPoint(HardPoint):
    __implements__ = ICleanPoint

class IEmailPowerup(Interface):
    pass

class EmailPowerup(provisioner.ProvisionedPowerup):
    __implements__ = (IEmailPowerup, provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable)
    powerupName = 'Email'

    startCost = 1
    tankID = provisioner.ETank.INCOMING_MAIL

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        p = Pool(avatar, name='Arrival')
        p.invisible = True
        avatar.getRootPool().addItem(p)
        p.addPipe(self.fwd(IPipe))
        avatar.setComponent(IMIMEDelivery, MIMEDeliverator(avatar, p.referenceTo()))
        cp = CleanPoint(avatar, 'Clean Point')
        cp.setPoint(p)
        avatar.setComponent(ICleanPoint, cp.referenceTo())
        self.msgScope = Statoscope()
        self.msgScope.start()
        self.invisible = True

    def itemAdded(self, item):
        self.msgScope.finish()
        self.msgScope.start()
        self.touch()

def newPoolForEMail(avatar, name):
    pool = Pool(avatar, name=name)
    pool.addIndex(DateIndex())
    pool.suggestIndex('received', backwards=True)
    pool.addIndex(NameIndex('subject'))
    pool.addIndex(NameIndex('sender'))
    pool.addIndex(FlagsIndex())
    pool.addIndex(SizeIndex())
    return pool

exts = {
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "png": "image/png",
    "gif": "image/gif"
    }
