# -*- test-case-name: quotient.test.test_imap -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import bisect
import operator
import weakref
import time
import rfc822

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from atop import store
from atop import tpython

from twisted.python import reflect
from twisted.python import log
from twisted.python.util import OrderedDict
from twisted.protocols import imap4
from twisted.python import components
from twisted.internet import protocol
from twisted.internet import defer

from quotient import mimemessage
from quotient import provisioner
from quotient import smartenup

MESSAGE_FLAGS = [
    '\\Answered',
    '\\Deleted',
    '\\Draft',
    '\\Flagged',
    '\\Recent',
    '\\Seen',
    ]

class IMAPAccountUp(provisioner.ProvisionedPowerup):
    __implements__ = (imap4.IAccount, provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable)
    powerupName = 'IMAP4 Server'

    dependencies = [smartenup.Smartenup]

    H_DELIM = '/'

    startCost = 1
    tankID = provisioner.ETank.INCOMING_MAIL

    # Cred haxor
    def logout(self):
        pass

    def setUpPools(self, substore):
        provisioner.ProvisionedPowerup.setUpPools(self, substore)

        self.subscriptions = {}

        self.rootRef = substore.rootPoolRef
        substore.setComponent(imap4.IAccount, self)
        self.greyedOut = True
        self.invisible = True


    def _createPool(self, old, path, pathspec):
        exists = True
        newPool = old
        for part in path:
            exists = False
            try:
                newPool = old.getIndexItem('name', part)
                exists = True
            except store.NoSuchIndex:
                # This pool has no name index.  We can't use it.
                raise imap4.MailboxException("Cannot create folders there")
            except KeyError:
                newPool = store.Pool(self.store, part)
                old.addItem(newPool)
            old = newPool
        if exists:
            # The terminal component of the path specification
            # already existed.  Call this a failure.
            raise imap4.MailboxCollision(self.H_DELIM.join(pathspec))

    def create(self, pathspec):
        # XXX - Get a case insensitive version of NameIndex
        path = pathspec.lower()
        path = path.split(self.H_DELIM)
        oldPool = self.rootRef.getItem()
        if path[0] == 'inbox':
            oldPool = oldPool.getIndexItem('name', 'Arrival')
            del path[0]
        self.store.transact(self._createPool, oldPool, path, pathspec)
        return True

    def addMailbox(self, name, mbox = None):
        raise RuntimeError("Don't call this")

    def select(self, name, rw=True):
        parts = name.split(self.H_DELIM)
        def trashyhack():
            pool = self.rootRef.getItem()
            for p in parts:
                for nextPool in pool.queryIndex('name', result=store.INDEX):
                    if nextPool.lower() == p.lower():
                        pool = pool.getIndexItem('name', nextPool)
                        break
                else:
                    return None
            return Poolbox(self.store, pool, rw)
        try:
            return self.store.transact(trashyhack)
        except store.NoSuchIndex:
            # raise imap4.MailboxException("Cannot select that folder")
            return None

    def delete(self, name):
        raise imap4.MailboxException("Use the web UI to delete things, please.")

    def rename(self, oldname, newname):
        raise imap4.MailboxException("Use the web UI to rename things, please.")

    def isSubscribed(self, name):
        return name.lower() in self.subscriptions

    def subscribe(self, name):
        def _():
            self.touch()
            self.subscriptions[name.lower()] = True
        self.store.transact(_)
        return True

    def unsubscribe(self, name):
        def _():
            self.touch()
            del self.subscriptions[name.lower()]
        try:
            self.store.transact(_)
        except KeyError:
            raise imap4.MailboxException("Not subscribed to that mailbox.")
        return True

    def _getAllPools(self, root, seen=None):
        if seen is None:
            seen = {}
        if root in seen:
            return
        seen[root] = None
        for x in root.queryIndex():
            if isinstance(x, store.Pool):
                yield x.name, x
                for (name, pool) in self._getAllPools(x, seen):
                    yield self.H_DELIM.join((x.name, name)), pool

    def listMailboxes(self, ref, wildcard):
        L = []
        if ref == "":
            re = imap4.wildcardToRegexp(wildcard, None)
            def _():
                for (name, pool) in self._getAllPools(self.rootRef.getItem()):
                    if re.match(name):
                        L.append((name, Poolbox(self.store, pool, False)))
            self.store.transact(_)
        return L

    def getPersonalNamespaces(self):
        return [["", None]]

    def getSharedNamespaces(self):
        return None

    def getUserNamespaces(self):
        return None

class Poolbox(components.Adapter, tpython.Forgetter):

    FLAGS = {
        '\\Seen': mimemessage.MIMEMessage.SEEN,
        '\\Answered': mimemessage.MIMEMessage.ANSWERED,
        '\\Forwarded': mimemessage.MIMEMessage.FORWARDED,
        '\\Redirected': mimemessage.MIMEMessage.REDIRECTED,
        '\\Deleted': mimemessage.MIMEMessage.DELETED,
    }

    temporaryAdapter = True

    exists = None
    recent = None

    ITEM_TYPES = (mimemessage.MIMEMessage)

    def __init__(self, avatar, original, rw):
        components.Adapter.__init__(self, original)
        self.avatar = avatar
        msgs = list(self.original.queryIndex('uid', result=store.ITEM))
        msgs = filter(self.ok, msgs)
        self.sequence = []
        self.references = []
        for m in msgs:
            self.sequence.append(self.uidByItem(m))
            self.references.append(m.referenceTo())
        self.listeners = []
        self.rw = rw
        self.original.addPipe(self, ephemeral=True)

    def ok(self, item):
        return isinstance(item, self.ITEM_TYPES)

    def skip(self, item):
        return not isinstance(item, self.ITEM_TYPES)

    # IPipe implementation
    def itemAdded(self, item):
        if self.skip(item):
            return
        self.sequence.append(self.uidByItem(item))
        self.references.append(item.referenceTo())
        exists = len(self.sequence)
        recent = self.getUnseenCount()
        for L in self.listeners:
            L.newMessages(exists, recent)

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, changeData):
        pass

    # IMailbox implementation

    def getUIDValidity(self):
        return self.original.storeID

    def getUIDNext(self):
        return self.original.nextUID()

    def getUID(self, message):
        return self.sequence[message - 1]

    def getFlags(self):
        return self.FLAGS.keys()

    def getMessageCount(self):
        return len(self.sequence)

    def getRecentCount(self):
        return self.getUnseenCount()

    def getUnseenCount(self):
        f = mimemessage.MIMEMessage.SEEN
        unseen = 0
        for ref in self.references:
            o = ref.getItem()
            if not o.flags & f:
                unseen += 1
        return unseen

    def isWriteable(self):
        return self.rw

    def destroy(self):
        raise imap4.MailboxException("Please delete folders using the web UI.")

    def getHierarchicalDelimiter(self):
        return "/"

    def requestStatus(self, names):
        return imap4.statusRequestHelper(self, names)

    def addListener(self, listener):
        self.listeners.append(listener)

    def removeListener(self, listener):
        self.listeners.remove(listener)

    def addMessage(self, message, flags = (), date = None):
        def cbFeedFile(msg):
            f = 0
            if '\\Seen' in flags:
                f |= msg.SEEN
            if '\\Answered' in flags:
                f |= msg.ANSWERED
            def doIt(pool, msg, flags):
                msg.changeFlag(flags)
                pool.addItem(msg)
            self.avatar.transact(doIt, self.original, msg, f)
        mr = mimemessage.MIMEMessageReceiver(self.avatar, cbFeedFile, True)
        return mr.feedFile(message).addCallback(self.uidByItem)

    def expunge(self):
        delete = []
        def collectDeleted(rate=50):
            limit = rate
            for (i, ref) in enumerate(self.references):
                msg = ref.getItem()
                if msg.flags & msg.DELETED:
                    delete.append((i, msg))
                limit -= 1
                if limit == 0:
                    limit = rate
                    yield None
        def removeDeleted():
            for (i, msg) in delete:
                try:
                    self.original.removeItem(msg)
                except KeyError:
                    # Maybe someone beat us to it, no big deal
                    pass
            ids = [i + 1 for (i, msg) in delete]
            ids.sort()
            ids.reverse()
            return ids
        g = collectDeleted()
        d = tpython.iterateInReactor(g)
        d.addCallback(lambda r: self.avatar.transact(removeDeleted))
        return d

    def _getnext(self, last):
        x = bisect.bisect(self.sequence, last)
        if x == len(self.sequence):
            raise StopIteration()
        return self.sequence[x]

    def _seqAndUIDByEither(self, msgset, uid):
        if not self.sequence:
            return

        if uid:
            msgset.last = self.sequence[-1]
            msgset.getnext = self._getnext
            f = lambda m: (self.sequence.index(m) + 1, m)
        else:
            msgset.last = len(self.sequence)
            f = lambda m: (m, self.sequence[m - 1])

        for msg in msgset:
            yield f(msg)

    def fetch(self, msgset, uid):
        for s, u in self._seqAndUIDByEither(msgset, uid):
            msg = self.itemByUID(u)
            yield s, Message(msg, u)

    def store(self, msgset, flags, mode, uid):
        if not self.isWriteable():
            raise imap4.ReadOnlyMailbox("Cannot store")

        state = []

        flags = reduce(operator.or_, [self.FLAGS[f] for f in flags])

        for (seqNum, uid) in self._seqAndUIDByEither(msgset, uid):
            msg = self.itemByUID(uid)
            seen = msg.SEEN & msg.flags
            answered = msg.ANSWERED & msg.flags

            f = 0
            if mode == -1:
                for bv in self.FLAGS.values():
                    if bv & flags & msg.flags:
                        f |= bv
                if f:
                    state.append((msg, f, False, seqNum))
            elif mode == 0:
                state.append((msg, msg.flags, False, seqNum))
                state.append((msg, flags, True, seqNum))
            elif mode == 1:
                for bv in self.FLAGS.values():
                    if bv & flags & ~msg.flags:
                        f |= bv
                if f:
                    state.append((msg, f, True, seqNum))

        result = {}
        if state:
            def _():
                for (msg, flags, set, seqNum) in state:
                    msg.changeFlag(flags, set)
                    result[seqNum] = ['\\' + x for x in msg.stringFlags()]
            self.avatar.transact(_)
        return result

    # Helper-o
    def itemByUID(self, uid):
        return self.original.getIndexItem('uid', uid - 1)

    def uidByItem(self, item):
        return item.index_uid(self.original) + 1

class _MsgMixin:
    def getHeaders(self, negate, *names):
        r = OrderedDict()
        if negate:
            opt = dict.fromkeys(map(str.lower, names))
            for (n, v) in self.msg.items():
                n = n.lower()
                if n not in opt:
                    r[n] = v
        else:
            for n in names:
                n = n.lower()
                v = self.msg[n]
                if v:
                    r[n] = v
        return r

    def isMultipart(self):
        return self.msg.is_multipart()

    def getSubPart(self, part):
        return SubMessage(self.msg.children[part])

    def getSize(self):
        return self.msg.size

class Message(_MsgMixin):

    __implements__ = (imap4.IMessage, imap4.IMessageFile)

    def __init__(self, msg, uid):
        assert isinstance(msg, mimemessage.MIMEMessage)
        self.msg = msg
        self.uid = uid

    def getUID(self):
        return self.uid

    def getFlags(self):
        return [k for (k, v) in Poolbox.FLAGS.items() if self.msg.flags & v]

    def getInternalDate(self):
        return self.msg['x-divmod-processed']

    def getBodyFile(self):
        f = self.msg.openFile()
        f.seek(self.msg.bodyOffset)
        return f

    def open(self):
        return self.msg.openFile()

class SubMessage(_MsgMixin):

    __implements__ = (imap4.IMessagePart,)

    def __init__(self, mimepart):
        self.msg = mimepart

    def getBodyFile(self):
        return StringIO(self.msg.get_payload(decode=False))

class IMAP4Server(imap4.IMAP4Server):
    def newMessages(self, exists, recent):
        if hasattr(imap4.IMAP4Server, 'do_IDLE'):
            imap4.IMAP4Server.newMessages(self, exists, recent)

class IMAP4Factory(protocol.ServerFactory):
    protocol = IMAP4Server

    def __init__(self, portal, challengers = None, contextFactory = None):
        self.portal = portal
        if challengers is None:
            challengers = {}
        self.challengers = challengers
        self.contextFactory = contextFactory

    def startFactory(self):
        self.connectedProtocols = weakref.WeakKeyDictionary()

    def stopFactory(self):
        for p in self.connectedProtocols:
            try:
                p.transport.loseConnection()
            except:
                log.err()
        self.connectedProtocols = None

    def buildProtocol(self, addr):
        p = self.protocol(self.challengers, self.contextFactory)
        p.portal = self.portal
        p.factory = self
        self.connectedProtocols[p] = True
        return p
