# -*- test-case-name: quotient.test.test_popup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os, md5
import weakref

from twisted.protocols.pop3 import IMailbox as IPOP3Mailbox
from twisted.python.components import registerAdapter, Adapter
from twisted.internet.protocol import Factory

from atop.store import UIDIndex, STORE_ID

from atop import tpython
from quotient.mimemessage import EmailPowerup
from quotient.mimemessage import IMIMEDelivery, ICleanPoint

from quotient import provisioner

class AlreadyLocked(Exception):
    """Raised when a mailbox is locked.
    """
    def __init__(self):
        Exception.__init__(self, "Can't get lock.  Mailbox in use")

class SizeListing:
    def __init__(self, box):
        self.box = box
    def __getitem__(self, i):
        return self.box.getMessageSize(i)

class POPIndex(UIDIndex):
    name = "pop"

class Popup(provisioner.ProvisionedPowerup):

    __implements__ = (IPOP3Mailbox, provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable)

    dependencies = [EmailPowerup]
    persistenceForgets = ('locked', 'msgs', 'deleted')
    locked = False
    powerupName = 'POP3 Service'

    tankID = provisioner.ETank.INCOMING_MAIL
    startCost = 1

    greyedOut = True
    invisible = True

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        cp = ICleanPoint(avatar)
        self.arrivalRef = cp.getItem().sourcePoolRef
        self.arrivalRef.getItem().addIndex(POPIndex())
        avatar.setComponent(IPOP3Mailbox, self)

    def login(self):
        if self.locked:
            raise AlreadyLocked()
        self.locked = True
        self.deleted = []
        self.msgs = self.store.transact(list, self.arrivalRef.getItem().queryIndex("pop", result=STORE_ID))
        return self

    def listMessageSizes(self):
        return SizeListing(self)

    def getMessageSize(self, i):
        f = self.store.getItemByID(self.msgs[i]).openFile()
        return os.fstat(f.fileno()).st_size

    def getMessage(self, i):
        return self.store.getItemByID(self.msgs[i]).openFile()

    def getUidl(self, i):
        return self.msgs[i]

    def listMessages(self, index=None):
        """Retrieve the size of one or more messages.

        @type index: C{int} or C{None}
        @param index: The number of the message for which to retrieve the size,
        or None to retrieve the size of all messages.

        @rtype: C{int} or C{list} of C{int}
        @return: The number of octets in the specified message, or a
        list of integers representing the number of octets in all the
        messages.
        """
        if index is None:
            return self.listMessageSizes()
        return self.getMessageSize(index)

    def deleteMessage(self, index):
        self.deleted.append(self.msgs[index])

    def undeleteMessages(self):
        self.deleted = []

    def logout(self):
        self.sync()
        self.locked = False

    def sync(self):
        def _():
            avatar = self.store
            for sid in self.deleted:
                avatar.getItemByID(sid).setIndexKey("pop", None)
        self.store.transact(_)
        self.deleted = []

from twisted.protocols.pop3 import POP3
class LoggingInPOP3(POP3):
    def _cbMailbox(self, ial, user):
        result = POP3._cbMailbox(self, ial, user)
        self.mbox = self.mbox.login()
        return result

    def do_XSENDER(self, *a):
        # We could actually implement this, but we won't for now
        self.failResponse("XSENDER is unsupported")


class POP3Factory(Factory):
    protocol = LoggingInPOP3
    debug = False

    def __init__(self, portal, auths = None):
        if auths is None:
            auths = {}
        self.portal = portal
        self.challengers = auths

    def startFactory(self):
        self.connectedProtocols = weakref.WeakKeyDictionary()

    def stopFactory(self):
        for p in self.connectedProtocols:
            try:
                p.transport.loseConnection()
            except:
                log.err()
        self.connectedProtocols = None

    def buildProtocol(self, addr):
        p = self.protocol()
        p.factory = self
        p.portal = self.portal
        self.connectedProtocols[p] = True
        return self.debug and tpython.DebugShimTransport(p) or p
