# -*- test-case-name: quotient.test.test_smtp -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""Incoming ESMTP server for delivery to users."""

import weakref

from twisted.protocols import smtp
from twisted.internet import protocol, defer
from twisted.python import log
from twisted.cred.checkers import ICredentialsChecker

from atop import tpython

from quotient import provisioner
from quotient.copyright import longversion
from quotient.mimemessage import IMIMEDelivery

def validateTo(store, user, tankID = provisioner.ETank.QUOTA):
    """Verify email is valid Quotient user."""
    user, dom = user.dest.local, user.dest.domain
    route = user.split('.')
    user = route.pop()
    try:
        avatar = ICredentialsChecker(store).getAvatar(user, dom)
    except KeyError:
        raise smtp.SMTPBadRcpt(user)
    def _():
        return provisioner.IProvisioner(avatar).deplete(tankID, 1, True)
    if not avatar.transact(_):
        raise smtp.SMTPBadRcpt(user, 552, "Requested mail action aborted: exceeded storage allocation")
    if not route:
        return lambda: Message(avatar, route=None)
    else:
        return lambda: Message(avatar, route=route)

class QuotientMessageDelivery(tpython.Forgetter):
    """Message delivery into Quotient."""
    
    __implements__ = smtp.IMessageDelivery,

    tankID = provisioner.ETank.QUOTA

    def __init__(self, store, avatar = None):
        """Create a QuotientMessageDelivery object.
        
        @type store: Object adaptable to C{ICredentialsChecker}
        @param store: The realm in which recipients reside.
        
        @type avatar: C{QuotientAvatar}
        @param avatar: The avatar sending the mail, or None.
        """
        self.store = store
        self.avatar = avatar
    
    def validateFrom(self, helo, origin):
        """Accept all Froms by default."""
        return origin

    def validateTo(self, user):
        return validateTo(self.store, user, self.tankID)

    def receivedHeader(self, helo, origin, recipients):
        # XXX Hack?
        domain = str(recipients[0]).split('@', 1)[1]
        
        heloStr = ""
        if helo[0]:
            heloStr = " helo=%s" % (helo[0],)
        auth = ''
        if self.avatar:
            auth = ' AUTH=' + self.avatar.replyTo.encode('xtext')
        from_ = "from %s ([%s]%s)%s" % (helo[0], helo[1], heloStr, auth)
        by = "by %s with ESMTP (%s)" % (domain, longversion)
        for_ = "for %s; %s" % (' '.join(map(str, recipients)), smtp.rfc822date())
        return "Received: %s\n\t%s\n\t%s" % (from_, by, for_)

class Message:
    """A message being delivered by the ESMTP server."""
    
    __implements__ = smtp.IMessage,

    def __init__(self, avatar, route=None):
        self.avatar = avatar
        self.route = route
        self.receiver = IMIMEDelivery(avatar).createMIMEReceiver(False)
        self.receiver.makeConnection(None)

    def lineReceived(self, line):
        return self.receiver.lineReceived(line)

    def eomReceived(self):
        self.receiver.message.route = self.route
        self.receiver.messageDone()
        # If we want to delay notification to the peer of successful reception of the
        # message, this is the place to do it.
        return defer.succeed("deliverated")
    
    def connectionLost(self):
        self.receiver.connectionLost(None)

class ESMTPInFactory(protocol.ServerFactory):
    """Server protocol for ESMTP."""
    protocol = smtp.ESMTP
    debug = False

    connectedProtocols = None

    def __init__(self, store, portal, auths = None, ctxFactory = None):
        self.delivery = QuotientMessageDelivery(store)
        self.portal = portal
        self.auths = auths
        self.ctxFactory = ctxFactory

    def startFactory(self):
        self.connectedProtocols = weakref.WeakKeyDictionary()

    def stopFactory(self):
        for p in self.connectedProtocols:
            try:
                p.transport.loseConnection()
            except:
                log.err()
        self.connectedProtocols = None

    def buildProtocol(self, addr):
        p = self.protocol(self.auths, self.ctxFactory)
        p.portal = self.portal
        p.factory = self
        p.delivery = self.delivery
        self.connectedProtocols[p] = True
        return self.debug and tpython.DebugShimTransport(p) or p
