# -*- test-case-name: quotient.test.test_smtp -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import sys
import rfc822
import math
import time
from email.Generator import Generator

from atop import credup
from atop import sched
from atop import store
from atop.tpython import Forgetter

from twisted.mail import bounce
from twisted.protocols import smtp
from twisted.python import log
from twisted.python import components
from twisted.internet import error
from twisted.internet import defer

from twisted.protocols import smtp
from twisted.names.client import Resolver
from twisted.mail.relaymanager import MXCalculator

from quotient.copyright import longversion
from quotient.proto import smtpin
from quotient import mimemessage
from quotient import mimebakery
from quotient import provisioner
from quotient import prefup

def logpassthru(err):
    log.err(err)
    return err

def backoff(T, N, q=math.e):
    def intervals():
        a = T / ((q ** (N + 1) - 1) / (q - 1))
        for i in range(N + 1):
            yield a * (q ** i)
    return intervals()

def makeResolver():
    if sys.platform == 'win32':
        return Resolver(servers=[('204.91.10.94', 53)])
    return None

def msgToFile(msg):
    """This function is Death Made Source Code.  Invoke it at your own peril.
    """
    s = StringIO()
    Generator(s).flatten(msg)
    s.seek(0, 0)
    return s


class IOutgoingMail(components.Interface):
    def sendMIMEMessage(self, to, msg):
        pass

class OutgoingMail(Forgetter):
    __implements__ = (IOutgoingMail,)
    persistenceForgets = ['mx']
    mx = None

    # Exponential backoff tuned to a 5 day retry period
    # Skip the first couple, though, because they're
    # vanishingly small.
    RETRANS_TIMES = list(backoff(60 * 60 * 24 * 5, 20))[2:]

    tankID = provisioner.ETank.OUTGOING_MAIL

    def sendmail(self,*a,**kw):
        return smtp.sendmail(*a,**kw)

    def __init__(self, avatar, outRef, transRef, sentRef):
        self.avatar = avatar
        self.outRef = outRef
        self.transRef = transRef
        self.sentRef = sentRef

    def sendString(self, messageText, attachments, to, fromAddress, cc='', bcc='', subject='(No Subject)'):
        """Take some user input and create a message and send it.

        @param messageText: A convenience string for the text/plain part.
        @param to: rfc822-format address list.
        @param fromAddress: rfc822-format address.
        @param cc: rfc822-format address list.
        @param bcc: rfc822-format address list.
        @param subject: str.  the message's subject.
        @param attachments: A list of 3-tuples of (data, mimetype, filename).
        """
        avatar = self.avatar
        to = list(rfc822.AddressList(to))
        cc = list(rfc822.AddressList(cc))
        bcc = list(rfc822.AddressList(bcc))
        sender = list(rfc822.AddressList(fromAddress))[0]
        rec = mimebakery.Recipe(avatar, to, sender=sender, cc=cc or None, bcc=bcc or None, subject=subject)
        rec.text(messageText)
        if attachments:
            for data, mimetype, filename in attachments:
                rec.attach(data, mimetype, filename)
        msg = rec.bake()
        dl = []
        for localpart, domain in [x[1].split('@') for x in to+cc+bcc]:
            dl.append(avatar.transact(self.sendMIMEMessage, fromAddress, localpart, domain, msg))
        return defer.DeferredList(dl)

    def sendMIMEMessage(self, from_, localpart, domain, msg):
        """
        @param from_: The address from which to send this message.
        @param localpart: The local part of the recipient of this message.
        @param domain: The domain part of the recipient of this message.
        @param msg: A MIMEMessage to send to the specified recipient.
        """
        self.transRef.getItem().addItem(msg)
        msg.markForTransmission(from_, localpart, domain)
        return self.send(msg)

    def send(self, msg):
        if self.mx is None:
            self.mx = MXCalculator(makeResolver())
        d = self.mx.getMX(msg.transmissionInfo[2])
        d.addCallback(self._cbGotExchange, msg)
        d.addErrback(self._ebGotExchange, msg)
        return d

    def flush(self):
        # Call this in a transaction
        out = self.outRef.getItem()
        g = out.queryIndex()
        try:
            now = sched.IScheduler(self.avatar).clock()
            m = g.next()
            while m is self.transRef.getItem() or m.nextAttempt < now:
                m = g.next()
        except StopIteration:
            return
        self._addAndRemove(self.transRef, self.outRef, m)
        self.send(m)

    def _cbGotExchange(self, mx, msg):
        from_, lpart, dname = msg.transmissionInfo
        d = self.sendmail(str(mx.exchange), from_, ['@'.join((lpart, dname))], msgToFile(msg))
        d.addCallback(self._cbMailsent, msg)
        return d

    def _ebGotExchange(self, failure, msg):
        # Check the retry count on this message
        if msg.transmissionAttempts >= len(self.RETRANS_TIMES) - 1:
            if failure.check(smtp.SMTPDeliveryError):
                return self.generateBounce(msg, failure.value.log)
            return self.generateBounce(msg)

        def _():
            msg.touch()
            msg.transmissionAttempts += 1
            offset = self.RETRANS_TIMES[msg.transmissionAttempts]
            msg.nextAttempt = sched.IScheduler(self.avatar).clock() + offset
        self.avatar.transact(_)

        if failure.check(smtp.SMTPDeliveryError):
            if 400 <= failure.value.code < 500:
                return self._transient(msg).addErrback(logpassthru)
            else: # 500 <= failure.value.code < 600:
                # Permanent failure, bounce it.
                return self.generateBounce(msg, failure.value.log)
        elif failure.check(error.DNSLookupError):
            return self._transient(msg).addErrback(logpassthru)
        else:
            return self.generateBounce(msg, failure.getTraceback())

    def _transient(self, msg):
        self.avatar.transact(self._addAndRemove, self.outRef, self.transRef, msg)
        fmt = "Transient delivery error.  Rescheduled %s for delivery at %s."
        args = (msg.getItemPath(), time.ctime(msg.nextAttempt))
        log.msg(fmt % args)
        return defer.succeed(None)

    def _cbBounce(self, result, msg):
        def _():
            self.transRef.getItem().removeItem(msg)
        self.avatar.transact(_)

    def _ebBounce(self, failure, msg):
        log.err(failure)
        log.msg("Bounce generation failed!!  Putting message back in outbox.")
        def _():
            self._addAndRemove(self.outRef, self.transRef, msg)
        self.avatar.transact(_)

    def _addAndRemove(self, addTo, removeFrom, msg):
        removeFrom.getItem().removeItem(msg)
        addTo.getItem().addItem(msg)

    def _cbMailsent(self, result, msg):
        def _():
            provisioner.IProvisioner(self.avatar).deplete(self.tankID, msg.size)
            self._addAndRemove(self.sentRef, self.transRef, msg)
        self.avatar.transact(_)
        return result

    def generateBounce(self, msg, transcript):
        f = msgToFile(msg)
        ffrom = msg.transmissionInfo[0]
        fto = '@'.join((msg.transmissionInfo[1:]))
        _, _, bounceStr = bounce.generateBounce(msgToFile(msg), ffrom, fto, transcript)
        mr = mimemessage.IMIMEDelivery(self.avatar).createMIMEReceiver(True)
        return mr.feedString(bounceStr
                             ).addCallbacks(self._cbBounce, self._ebBounce,
                                            callbackArgs=(msg,),
                                            errbackArgs=(msg,))

class Message:
    """A message being delivered to the ESMTP server for relaying."""

    __implements__ = smtp.IMessage,

    def __init__(self, avatar, sender):
        self.avatar = avatar
        m = mimemessage.MIMEMessageReceiver(avatar, sender, False)
        m.makeConnection(None)
        self.receiver = m

    def lineReceived(self, line):
        return self.receiver.lineReceived(line)

    def eomReceived(self):
        self.receiver.messageDone()
        # If we want to delay notification to the peer of successful reception of the
        # message, this is the place to do it.
        return defer.succeed("deliverated")

class TeeMessage:
    def __init__(self, one, two):
        self.one = one
        self.two = two

    def lineReceived(self, line):
        self.one.lineReceived(line)
        self.two.lineReceived(line)

    def eomReceived(self):
        return defer.DeferredList([self.one.eomReceived(), self.two.eomReceived()])

class MessageDelivery(store.Item):
    def getDomain(self):
        def _():
            return credup.ICredentialsChecker(self.store.store).getDomains().next()
        return self.store.transact(_).name

    def getDomains(self):
        def _():
            return [d.name for d in credup.ICredentialsChecker(self.store.store).getDomains()]
        return self.store.transact(_)

    def receivedHeader(self, helo, origin, recipients):
        # XXX violate that abstraction!  *violate* it!
        domain = self.getDomain()
        heloStr = ""
        if helo[0]:
            heloStr = " helo=%s" % (helo[0],)
        auth = " AUTH=" + self.store.name.encode("xtext")
        from_ = "from %s ([%s]%s)%s" % (helo[0], helo[1], heloStr, auth)
        by = "by %s with ESMTP (%s)" % (domain, longversion)
        for_ = "for %s; %s" % (" ".join(map(str, recipients)), smtp.rfc822date())
        return "Received: %s\n\t%s\n\t%s" % (from_, by, for_)

    def validateFrom(self, helo, origin):
        # The only from address we accept is that of the authenticated user
        # XXX How to verify the domain?
        if origin.local == self.store.name:
            return origin
        raise smtp.SMTPBadSender(origin)

    def validateTo(self, user):
        omail = IOutgoingMail(self.store)

        for d in self.getDomains():
            if user.dest.domain == d:
                remote = smtpin.validateTo(self.store.store, user)()
                local = Message(self.store, omail.sentRef.getItem().addItem)
                return lambda: TeeMessage(remote, local)

        # All other addresses are accepted
        f = self.store.name + '@' + self.getDomain()
        l = user.dest.local
        d = user.dest.domain
        def makeSender():
            return lambda m: omail.sendMIMEMessage(f, l, d, m)
        return lambda: Message(self.store, makeSender())

class OutgoingMailup(provisioner.ProvisionedPowerup, sched.Schedulable):
    __implements__ = (provisioner.ProvisionedPowerup.__implements__, sched.Schedulable.__implements__, provisioner.IProvisionable)

    dependencies = [provisioner.Provisioner, sched.SubScheduler]
    storeDependencies = [sched.SystemScheduler]

    powerupName = 'SMTP Client'

    interval = 60

    startCost = 1
    tankID = provisioner.ETank.OUTGOING_MAIL

    greyedOut = True
    invisible = True

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        outbox = store.Pool(avatar, name="Outbox")
        transit = store.Pool(avatar, name="In Transit")
        outbox.addItem(transit)
        transit.invisible = True

        sent = store.Pool(avatar, name="Sent")

        avatar.getRootPool().addItem(outbox)
        avatar.getRootPool().addItem(sent)

        avatar.setComponent(IOutgoingMail,
                            OutgoingMail(avatar,
                                         outbox.referenceTo(),
                                         transit.referenceTo(),
                                         sent.referenceTo()))
        avatar.setComponent(smtp.IMessageDelivery, MessageDelivery(avatar))
        self.reset(self.interval)

    def run(self):
        self.reset(self.interval)

        o = IOutgoingMail(self.store)
        self.store.transact(o.flush)
