# -*- test-case-name: quotient.test.test_reset -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


import struct
import random

from atop import store
from atop import sched
from atop import powerup

from twisted.python import log
from twisted.python import components

from quotient import prefup
from quotient import mimemessage
from quotient.proto import smtpout

class ResetError(Exception):
    pass

class ResetAlreadyPending(ResetError):
    pass

class ResetRecentlyRequested(ResetError):
    pass

class NoContactAddressDefined(ResetError):
    pass

class IPasswordResetter(components.Interface):
    def resetPasswordFor(self, key, password):
        """Change the password for the avatar associated with the given key.
        
        @raise KeyError: if the specified key has timed out or never existed.
        """

    def requestReset(self, avatar):
        """Request that a key be mailed to the given avatar.
        
        The key can be used to change the given avatar's password for
        a limited period of time.
        """

class PasswordResetter(powerup.Powerup, sched.Schedulable):

    __implements__ = (powerup.Powerup.__implements__, sched.Schedulable.__implements__, IPasswordResetter)

    dependencies = [sched.SubScheduler, mimemessage.EmailPowerup, smtpout.OutgoingMailup]
    powerupName = 'Password Resetter'

    MESSAGE = ('<html>'
               '<a href="%(baseURL)s/reset/?key=%(key)s">Click here</a>'
               '</html>')

    # How long requested reset links last, in seconds
    TIMEOUT = 60 * 60

    # How long a time must pass between reset requests, in seconds
    REREQUEST_DELAY = 60 * 60 * 24 * 7
    
    # How long between cleanups, in seconds
    interval = 60 * 60
    
    # Mapping of request keys to  (request time, avatar references)
    # for all outstanding password reset requests.
    outstanding = None
    
    # Sorted list of (expiration time, request key)
    outstandingSorted = None
    
    # Mapping of storeIDs to times when another reset
    # request will be allowed.
    recent = None

    # Sorted list of (time to allow another reset request, store ID)
    recentSorted = None
    
    def __init__(self, store):
        powerup.Powerup.__init__(self, store)
        self.outstanding = {}
        self.outstandingSorted = []
        self.recent = {}
        self.recentSorted = []

    def setUpPools(self, substore):
        substore.setComponent(IPasswordResetter, self)
    
    def resetPasswordFor(self, key, password):
        self.touch()
        storeID = self.outstanding.pop(key)
        orig = self.store.store.getItemByID(storeID)
        orig.setInsecurePassword(password)
        orig.setSecurePassword(password)

    def requestReset(self, avatar):
        if avatar.storeID in self.outstanding.values():
            raise ResetAlreadyPending()
        elif avatar.storeID in self.recent:
            raise ResetRecentlyRequested()

        self.touch()

        key = self.newResetRequest(avatar.storeID)
        msg = self.formatMessage(key)
        recip = self.determineDestination(avatar)
        origin = self.determineOrigin(avatar)
        subj = "Password Reset Request"
        
        # XXX - If the avatar has a public key we know about, encrypt the message with it.
        out = smtpout.IOutgoingMail(self.store)
        d = out.sendString(msg, (), recip, origin, subject=subj)
        d.addCallbacks(self._cbResetRequest, self._ebResetRequest)
        return d

    def _cbResetRequest(self, result):
        # print 'Reset Request Success', result
        return result

    def _ebResetRequest(self, failure):
        # print 'Reset Request Failure'
        # log.err(failure)
        return failure

    def newResetRequest(self, storeID, size=8):
        # self.touch()
        rands = [random.randrange(2 ** 31) for i in range(size)]
        bytes = struct.pack('%di' % size, *rands)
        key = bytes.encode('hex')
        self.outstanding[key] = storeID
        now = sched.IScheduler(self.store).clock()
        self.outstandingSorted.append((now + self.TIMEOUT, key))
        self.recent[storeID] = now + self.REREQUEST_DELAY
        self.recentSorted.append((now + self.REREQUEST_DELAY, storeID))
        return key

    def formatMessage(self, key):
        return self.MESSAGE % {'baseURL': 'http://localhost:8485', 'key': key}
    
    def determineDestination(self, avatar):
        dest = prefup.IPrefs(avatar).getPref('emergencyContact')
        if not dest:
            raise NoContactAddressDefined()
        return dest
    
    def determineOrigin(self, avatar):
        return "postmaster@divmod.org"

    def run(self):
        self.touch()
        self.reset(self.interval)
        
        now = sched.IScheduler(self.store).clock()
        while self.outstandingSorted and self.outstandingSorted[0][0] <= now:
            try:
                del self.outstanding[self.outstandingSorted[0][1]]
            except KeyError:
                pass
            del self.outstandingSorted[0]
        while self.recentSorted and self.recentSorted[0][0] <= now:
            try:
                del self.recent[self.recentSorted[0][1]]
            except KeyError:
                pass
            del self.recentSorted[0]

from atop import credup

from nevow import tags
from nevow import formless
from nevow import freeform

from quotient.web import root
from quotient.taglib import page

class IResetter(formless.TypedInterface):
    def requestReset(self, username = formless.String()):
        """Send a password reset token to the specified user.
        """
    requestReset = formless.autocallable(requestReset)

class ResetterDisplay(root.Base, freeform.Configurable):
    __implements__ = (IResetter,)

    document = page(title="Password Reset Request", anonymous=True)[ freeform.configure ]

    def requestReset(self, username):
        av = credup.preauth(self.original.store.store, username)
        try:
            d = self.original.requestReset(av)
        except ResetRecentlyRequested:
            return ("A password reset was recently requested for this account. "
                    "Passwords may only be reset at infrequent intervals. "
                    "Please try again in a few days.")
        except ResetAlreadyPending:
            return ("A token has already been issued to reset this account's "
                    "password.  Another may not be issued until the first expires "
                    "or is used.")
        except NoContactAddressDefined:
            return ("No emergency contact address has been defined for this "
                    "account.  Unfortunately, this means the password may not "
                    "be reset automatically.")
        def ok(_):
            return "Reset token sent"
        def fail(_):
            log.err(_)
            return "Reset token NOT sent"
        d.addCallbacks(ok, fail)
        return d

    postLocation = '/reset/'
    # postLocation = property(lambda self: self.original.storeID)
    
    def render(self, request):
        if request.args.has_key("key"):
            key = request.args["key"][0]
            newPassword = genNewPassword()
            try:
                self.original.store.transact(self.original.resetPasswordFor, key, newPassword)
            except KeyError:
                return "<html>Invalid key</html>"
            return '<html>Your password is now %s<br><a href="%s">Return to the login page</a></html>' % (
                newPassword, '/'+('/'.join(request.sitepath))
            )
            # return page(title="Password Reset")[
            #     tags.p["Your password is now ", newPassword],
            #     tags.a(href='/'.join([''] + request.sitepath)),
            # ]
        return root.Base.render(self, request)

import string
def genNewPassword(pickFrom=string.letters+string.digits):
    return ''.join([random.choice(pickFrom) for i in range(8)])

