# -*- test-case-name: quotient.test.test_searchup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import time

from atop.powerup import Powerup, IPowerStation

from atop.tpython import Forgetter
from atop.store import IPipe, Pool, NameIndex
from atop import sched

from lupy.indexer import Index

from twisted.python import log
from twisted.internet import threads, reactor
from twisted.internet import defer
from twisted.python.components import Interface, implements

from stats import Statoscope

from quotient import taglib
from quotient.web.root import Base
from quotient.prefup import touchNForward
from quotient.provisioner import ProvisionedPowerup, ETank, NotEnoughFuel, IProvisioner, IProvisionable
from quotient import sysup

from nevow import tags
from nevow import formless
from nevow import freeform

class IBasicSearch(formless.TypedInterface):
    paused = formless.Boolean(
        label = "Pause indexing",
        description = 'Stops the indexer and queues new requests.'
        )

def hitRenderer(data, avatar):
    results = []
    for x in data:
        uid = x.get('uid')
        item = avatar.getItemByID(int(uid))
        try:
            summary = item['subject']
        except:
            summary = str(item)
        results.append(tags.p[
            tags.a(href=uid)[
                summary
            ]
        ])
    if not results:
        return tags.p["No results."]
    return results

def hitFailure(failure):
    return failure

class SearchDisplay(Base, freeform.Configurable):
    __implements__ = IBasicSearch,
    document = taglib.page(title="Search Home")[
        tags.p["Search:"],
        tags.form(action="")[
            tags.input(type="text", name="query"),
            tags.input(type="submit")
        ],
        freeform.configure
    ]

    paused = touchNForward('paused')

    def render(self, request):
        if request.args.has_key('query'):
            query = request.args['query'][0]
            # CLI HACK
            if query.startswith('/'):
                query = query[1:]
                cmd, arg = query.split(' ', 1)
                cmd = cmd.strip().lower()
                argFirst = arg.split(' ', 1)[0].lower()
                if cmd in ('j', 'p'):
                    if cmd == 'j':
                       pool = self.original.store.getRootPool()
                    elif cmd == 'p':
                       pool = IPowerStation(self.original.store).getPowerupsPool()
                    def _():
                        for p in pool.queryIndex('name', startKey=argFirst.capitalize()):
                            if p.name.lower().startswith(argFirst):
                                request.redirect(str(p.storeID))
                                break
                            else:
                                request.redirect(request.getHeader('referer'))
                    self.original.store.transact(_)
                else:
                    ps = IPowerStation(self.original.store)
                    def _():
                        for pup in ps.getPowerupsPool().queryIndex('name'):
                            if hasattr(pup, 'command'):
                                pup.command(cmd, arg)
                    self.original.store.transact(_)
                    self.doc = self.document
                    request.redirect(request.getHeader('referer'))
            else:
                hits = self.original.search(query)
                self.doc = taglib.page(title="Search Results")[
                    "Search results for: %s" % query,
                    hits.addCallback(hitRenderer, self.avatar
                        ).addErrback(hitFailure
                        )
                ]
        else:
            self.doc = self.document
        return Base.render(self, request)

class IIndexable(Interface):
    def indexableData(self):
        """
        Return a dictionary with strings as keys and values.
        For instance {'topic':'spam sandwich'} would index
        'spam sandwich' in a Lupy Field named 'topic'.
        See lupy.indexer.Index for more details.
        The searcher automatically adds the items storeID
        before saving it to the index. This allows you to
        retrieve your item from search results.
        """

class ISearch(Interface):
    def search(self, query):
        pass

class Searchup(ProvisionedPowerup, sched.Schedulable):
    __implements__ = sched.Schedulable.__implements__, IPipe, ISearch, IProvisionable, ProvisionedPowerup.__implements__,

    dependencies = [sched.SubScheduler]
    storeDependencies = [sysup.ControlPanel]

    powerupName = 'Search Engine'
    persistenceForgets = ['indexer', 'checkpointDefer']

    indexing = False
    invisible = True
    paused = False

    checkpointDefer = None

    tankID = ETank.INDEXING
    startCost = 1
    messageIndexCost = 1
    messageSearchCost = 1

    # TODO - Provisioning and Statoscope
    def setUpPools(self, avatar):
        ProvisionedPowerup.setUpPools(self, avatar)

        avatar.setComponent(ISearch, self)
        idxDir = avatar.newDirectory('indexer')
        create = not os.path.exists(os.path.join(idxDir, 'segments'))
        self.indexer = Index(idxDir, create=create)
        self.indexDir = idxDir
        self.indexer.setMergeFactor(25)
        self.scope = Statoscope()
        p = Pool(avatar, name='Indexing Queue')
        p.invisible = True
        self.indexQueuePoolRef = p.referenceTo()
        avatar.getRootPool().addItem(p)
        self.pendingQueries = []
        self.pendingDeletes = []

    def restoreStateFromStore(self, lastItem):
        cp = self.checkpointDefer
        ProvisionedPowerup.restoreStateFromStore(self, lastItem)
        self.checkpointDefer = cp

    def run(self):
        # self.unschedule()
        self._indexMore()

    def activate(self):
        self.indexer = Index(self.indexDir, create=False)
        self._setCheckpointTimer()

    def checkpoint(self):
        self.checkpointDefer = None
        self.indexer.close()

    def _setCheckpointTimer(self):
        if self.checkpointDefer:
            self.checkpointDefer.reset(15)
        else:
            self.checkpointDefer = reactor.callLater(15, self.checkpoint)

    def indexPool(self, pool):
        pool.touch()
        pool.addPipe(self.fwd(IPipe))

    def itemAdded(self, item):
        indexable = IIndexable(item, default=None)
        if indexable is not None:
            self.indexQueuePoolRef.getItem().addItem(item)
            if not self.paused:
                self._indexMore()

    def _threadIndexItem(self, item, delete=False):
        """
        Index an item. Should only be called if we are
        not already in the process of indexing an item.
        """
        def threadedDeletion(uid):
            self.indexer.delete(uid=uid)
        def threadedIndexing(**kw):
            before = time.clock()
            self.indexer.index(**kw)
            return time.clock() - before
        def deplete(p, id, amt):
            self.store.transact(p.deplete, id, amt)

        if delete:
            return threads.deferToThread(threadedDeletion, str(item.storeID)
                                         ).addCallback(lambda _: item
                                                       )
        cost = 0
        indexable = IIndexable(item, default=None)
        kw = indexable.indexableData()
        if kw:
            kw['_uid'] = str(item.storeID)
            p = IProvisioner(self.store)
            deplete(p, self.tankID, 1)
            d = threads.deferToThread(threadedIndexing, **kw)
            d.addCallback(lambda cost: deplete(p, self.tankID, cost - 1))
            d.addCallback(lambda _: item)
        else:
            d = defer.succeed(item)
        return d

    def _cbDoneIndexing(self, item):
        """
        Successfully indexed this item.
        """
        self.indexing = False
        if self.pendingQueries:
            # Pop from the *end* of the list because there is no reason not to.
            q, d = self.pendingQueries.pop()
            d.callback(self.indexer.find(q))
        elif self.pendingDeletes:
            d = defer.Deferred()
            d.callback(self.indexer.deleteAll(uid=self.pendingDeletes))
            d.addCallback(lambda r: (self._setCheckpointTimer(), r)[1])
            self.pendingDeletes = []

        def _():
            self.indexQueuePoolRef.getItem().removeItem(item)
            self._indexMore()

        self.store.transact(_)

    def systemBlockedIndexing(self):
        # System-wide block on indexing, try again later
        self.reset(60 * 30)

    def _indexMore(self):
        """
        Index any unindexed items in the indexQueue.
        """
        if self.indexing:
            # We are already busy indexing something
            return
        cp = sysup.IControlPanel(self.store.store)
        perm = sysup.IControlPanel.MESSAGE_INDEXING
        if not cp.actionAllowed(perm):
            return self.systemBlockedIndexing()
        if not reactor.running:
            self.reset(1)
            return
        for item in self.indexQueuePoolRef.getItem().queryIndex():
            self.indexing = True
            # Find and item that needs indexing and index it in a thread
            self._setCheckpointTimer()
            zindexing = self._threadIndexItem(item)
            zindexing.addCallback(self._cbDoneIndexing)
            zindexing.addErrback(self._ebIndexing)
            break
        else:
            self.indexing = False

    def _ebIndexing(self, failure):
        if failure.check(NotEnoughFuel):
            self.reset(60 * 60 * 24)
        else:
            self.indexing = False
            log.msg("FATAL ERROR INDEXER FAILED")
            log.err(failure)

    def search(self, query):
        if not query:
            return defer.succeed([])
        elif not self.indexing:
            return defer.succeed(self.indexer.find(query))
        else:
            self.pendingQueries.append((query, defer.Deferred()))
            return self.pendingQueries[-1][-1]

    def itemRemoved(self, item):
        indexable = IIndexable(item, default=None)
        if indexable is not None:
            if not self.indexing:
                self._setCheckpointTimer()
                self.indexer.delete(uid=str(item.storeID))
            else:
                self.pendingDeletes.append(str(item.storeID))
                self.touch()

    def itemChanged(self, item, changes):
        pass
