# -*- test-case-name: quotient.test.test_extract -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from atop import store
from atop import tpython

from nevow import tags

from quotient import cleanup
from quotient import provisioner
from quotient import mimemessage
from quotient import taglib
from quotient.web.root import Base
from quotient.smarts.note import INotebook, Notebook
from quotient.prefup import IPrefs

from twisted.python.components import Interface

class FromMessageIndex:
    __implements__ = (store.IPoolIndex,)

    def getIndexName(self):
        return "fmsg"

    def keyToSortableString(self, key):
        return '%010d' % (key,)

    def sortableStringToKey(self, sstr):
        return int(sstr)


class IExtractable(Interface):
    pass

class ISmartenup(Interface):
    def smartsPool(self):
        """Return the pool containing extraction smarts items"""

    def subscribeItem(self, smart):
        """Subscribe the given extraction smarts item to the item added
        event notification list.
        """

    def subscribePart(self, smart, type='text/plain', decode=True):
        """Subscribe the given extraction smarts item to the notification
        list for message parts which match the given type.
        """

class IInbox(Interface):
    inboxRef = property(doc="A reference to the inbox pool")

class Smartenup(provisioner.ProvisionedPowerup):

    __implements__ = (provisioner.ProvisionedPowerup.__implements__,
                      provisioner.IProvisionable, ISmartenup, IInbox)

    dependencies = [provisioner.Provisioner, cleanup.Cleanup, Notebook]
    invisible = True
    totalExtracted = 0
    tankID = provisioner.ETank.EXTRACTION
    startCost = 1
    byteCost = 0.0001
    powerupName = "Smartenup"

    def setUpPools(self, avatar, name=None):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)

        cleaner = cleanup.IClean(avatar)
        cleaner.cleanRef.getItem().addPipe(self.fwd(store.IPipe))

        notey = INotebook(avatar)
        notey.notesPool().addPipe(self.fwd(store.IPipe))

        avatar.setComponent(ISmartenup, self)
        avatar.setComponent(IInbox, self)

        smartsPool = store.Pool(avatar, "Collections")
        self.smartsPoolRef = smartsPool.getRef()
        # avatar.getRootPool().addItem(smartsPool)

        IPrefs(avatar).addToPrimaryNav(smartsPool)

        # dict of smartRef to type string
        self.partSmarts = {}
        # list of smartRefs that want the whole item
        self.itemSmarts = []

        # Arg special case arg
        inbox = mimemessage.newPoolForEMail(avatar, 'Inbox')
        self.inboxRef = inbox.referenceTo()
        avatar.getRootPool().addItem(inbox)
        IPrefs(avatar).addToPrimaryNav(inbox)

    def smartsPool(self):
        return self.smartsPoolRef.getItem()

    def subscribeItem(self, smart):
        self.smartsPool().addItem(smart.extractPool())
        self.itemSmarts.append(smart.referenceTo())

    def subscribePart(self, smart, type='text/plain', decode=True):
        self.smartsPool().addItem(smart.extractPool())
        self.partSmarts[smart.referenceTo()] = (type, decode)

    def itemAdded(self, item):
        doInbox = True
        if isinstance(item, mimemessage.MIMEMessage):
            for smartRef in self.itemSmarts:
                if smartRef.getItem().extractItem(item) == "filed":
                    doInbox = False
            self.extractItem(item)
        else:
            for smartRef, (type, decode) in self.partSmarts.items():
                if type.startswith('text/plain'):
                    smart = smartRef.getItem()
                    eItem = IExtractable(item, None)
                    if eItem:
                        smart.extractPart(item, item.description, None)
        if doInbox:
            self.inboxRef.getItem().addItem(item)

    def itemRemoved(self, item):
        if isinstance(item, mimemessage.MIMEMessage):
            smarts = self.partSmarts.keys() + self.itemSmarts
            for smartRef in smarts:
                smartRef.getItem().itemRemoved(item)
            self.removeFromInbox(item)

    def removeFromInbox(self, item):
        self.inboxRef.getItem().removeItem(item)
        
    def itemChanged(self, item, changeData):
        allSmarts = self.itemSmarts + self.partSmarts.keys()
        for smartRef in allSmarts:
            smart = smartRef.getItem()
            smart.itemChanged(item, changeData)

    def _extractLoop(self, msg):
        avatar = self.store
        provis = provisioner.IProvisioner(avatar)
        for part in msg.walk():
            pType = part.inferType()
            for smartRef, (type, decode) in self.partSmarts.items():
                if pType.startswith(type):
                    smart = smartRef.getItem()
                    payload = part.get_payload(decode)
                    cost = len(payload) * smart.byteCost
                    if True: #provis.deplete(smart.tankID, cost):
                        smart.extractPart(msg, payload, part)
                        if len(getattr(msg, 'extracts', ())) > 30:
                            return
                        yield None

    def extractItem(self, msg):
        def _(i):
            for e in i:
                pass
        self.store.transact(_, self._extractLoop(msg))

class SmartenupDisplay(Base):
    def render(self, context, data):
        return self.document

    document = taglib.page(title="Smartenup")[
        tags.p["Hello"],
    ]
