# -*- test-case-name: quotient.test.test_people -*-

import rfc822
import time

from twisted.python.components import Interface
from atop.powerup import Powerup
from atop.store import Item, Pool, ItemForwardingComponent, ItemReferenceIndex
from atop.store import IPipe, NameIndex, INDEX_AND_ITEM

from quotient.mimemessage import IMIMEDelivery, ICleanPoint, EmailPowerup

from quotient.web.root import Base, TabularBase, column
from quotient import taglib
from quotient import provisioner
from quotient.smarts.base import ExtractBase, ExtractCollectorBase
from quotient.addressbook import AddressBook, IAddressBook
from quotient.smartenup import Smartenup
from quotient.prefup import IPrefs

from nevow import tags
from nevow import formless
from nevow import freeform


phoneChoices = formless.Choice(["home", "business", "cell"])
emailChoices = formless.Choice(["home", "work"])

class ICard(formless.TypedInterface):
    class CardGroup(formless.TypedInterface):
        title = formless.String()
        organization = formless.String()

    emails = formless.Dictionary()

    phones = formless.Dictionary()
    def setEmailAddress(self, address = formless.String(), label=emailChoices):
        pass
    setEmailAddress = formless.autocallable(setEmailAddress)
    def setPhoneNumber(self, number = formless.String(), label=phoneChoices):
        pass
    setPhoneNumber = formless.autocallable(setPhoneNumber)



class Card:
    """A contact-information holding class; a metaphorical calling card.
    """
    name = None
    __implements__ = ICard,
    def __init__(self, person, title, organization):
        self.person = person
        self.title = title
        self.organization = organization
        self.emails = {}
        self.phones = {}
        self.name = "%s %s" % (title, organization)
        self.dateCreated = time.gmtime(time.time())

    def _updateDictAndIndex(self, dicti, index, label, value):
        previous = dicti.get(label)
        if value is None and previous is not None:
            del dicti[label]
            if not previous in dicti.values():
                self.person.removeFromIndex(index, previous)
        else:
            dicti[label] = value
        self.person.appendToIndex(index, value)

    def setEmailAddress(self, address, label=''):
        self._updateDictAndIndex(self.emails, 'email', label, address)

    def setPhoneNumber(self, number, label=''):
        self._updateDictAndIndex(self.phones, 'phones', label, number)

    def burn(self):
        """Destroy this card and remove all indexes associated with it.  (Call
        within a transaction.)
        """
        for func, data in ((self.setEmailAddress, self.emails),
                           (self.setPhoneNumber, self.phones)):
            for k in data.keys():
                func(None, label=k)
        self.person.cards.remove(self)


class TabularCard(TabularBase):
    def render_name(self, context, data):
        # HAAAAACK:
        ## Figure out which element we are in the list.
        top = context
        while top.parent is not None:
            top = top.parent
        num = getattr(top, 'number', 0)
        top.number = num + 1
        return tags.a(href="%s/%s" % (self.original.person.storeID, num))[
            "%s %s" % (self.original.title, self.original.organization)
        ]
    column(render_name, True, 200)

#    def action_burn(self, pool):
#        print "BNURN", self, pool


class PersonRow(TabularBase):
##    def render_name(self, context, data):
##        tags.a(href=str(data.storeID))[taglib.headerToUTF8(data.name)]
##        #return taglib.headerToUTF8(data.name)
##    column(render_name, True, 300)

    def render_addToAddressBook(self, context, data):
        addBook = IAddressBook(data.store)
        if data in addBook.addressPool():
            return tags.a( href="/me/%s?remove=%s" % (addBook.storeID, data.storeID)) [ "Remove from address book" ]
        else:
            return tags.a( href="/me/%s?add=%s" % (addBook.storeID, data.storeID)) [ "Add to address book" ]
    column(render_addToAddressBook, True, 250, "Add To Address Book")

class IPerson(formless.TypedInterface):
    #personName = formless.String(label="Name")

    def newCard(self, organization=formless.String(), title=formless.String()):
        pass
    newCard = formless.autocallable(newCard)


class Person(Item):
    __implements__ = IPerson
    def recentMessages(self, limit=5):
        """Returns an iterator.
        """
        mp = IPersonCollector(self.store).getMessagesPool()
        if mp is not None:
            rt = self.referenceTo()
            return mp.queryIndex('person', startKey=rt,
                                 endKey=rt, backwards=True, limit=limit)
        return iter([])

    def __init__(self, store, name, creator=None):
        Item.__init__(self, store, name)
        self.cards = []
        if creator:
            self.creatorRef = creator.getRef()
        else:
            self.creatorRef = IPrefs(store).meRef

    def _touchIt(self, newVal):
        self.setIndexKey('name', newVal)
    personName = property(lambda self: taglib.headerToUTF8(data.name or '<No Name>'), _touchIt)

    def newCard(self, organization, title=''):
        """Add a new card for this person with appropriate contact information.
        """
        self.touch()
        r = Card(self, title, organization)
        self.cards.append(r)
        return r

    def burnCards(self, cards):
        self.touch()
        for card in cards:
            card.burn()

    def histogram(self, days=1000):
        msgs = IPersonCollector(self.store).getMessagesPool()
        D = {}
        now = time.gmtime()
        then = time.mktime(now) - (60 * 60 * 24 * days)
        selfref = self.getRef()
        for m in msgs.queryIndex('received',  endKey=then, backwards=True):
            if m.creatorRef == selfref:
                y, m, d = m.dateReceived[:3]
                day = (y,m,d)
                dayMsgs = D.setdefault(day,0)
                D[day] = dayMsgs + 1
        days = D.keys()
        days.sort()
        return [(day, D[day]) for day in days]


class PersonExtract(ExtractBase):
    pass


class IPersonCollector(formless.TypedInterface):
    "The Address Book"
    def newPerson(self, newPersonName=formless.String(label="Name")):
        """Add a Person to the Address Book
        """
        pass
    newPerson = formless.autocallable(newPerson)


class PersonCollector(ExtractCollectorBase):
    __implements__ = IPersonCollector, IPipe, Powerup.__implements__, \
                     provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable
    powerupName = 'People'
    dependencies = [AddressBook, Smartenup]

    def setUpPools(self, store):
        ExtractCollectorBase.setUpPools(self, store, name='People', type=None)
        store.setComponent(IPersonCollector, self)
        self.getMessagesPool().addIndex(ItemReferenceIndex('person', store))
        self.getPeoplePool().addIndex(NameIndex('email'))
        self.getPeoplePool().addIndex(NameIndex())
        # Hack to get it accssible from root for left nav
        store.getRootPool().addItem(self.extractPool())
        # for searching when we scan email
        self.invisible = False
        IPrefs(store).addToPrimaryNav(self)

    def newPerson(self, newPersonName, guessed=False):
        p = Person(self.store, newPersonName)
        self.getPeoplePool().addItem(p)
        return p

    def getPeoplePool(self):
        return self.extractPool()

    def getMessagesPool(self):
        return ICleanPoint(self.store).getItem().sourcePoolRef.getItem()

    def peopleFromMessage(self, item, hdr = 'from'):
        fromAddress = item[hdr]
        if not fromAddress:
            # No from header: let's not bother.
            return
        name, address = item.nameAndAddress()
        if not address:
            # Invalid header: let's not bother.
            return
        folks = list(self.getPeoplePool().queryIndex('email', startKey=address, endKey=address))
        return folks

    def extractItem(self, item):
        if hasattr(item, 'creatorRef') and item.creatorRef:
            # Item already has a creator assigned
            return
        folks = self.peopleFromMessage(item)
        if not folks:
            # doesn't exist.  create!
            name, address = item.nameAndAddress()
            if not name:
                name = address
            person = self.newPerson(name)
            card = person.newCard(address.split('@')[-1])
            card.setEmailAddress(address)
        else:
            person = folks[-1]
        item.setIndexKey('person', person.referenceTo())
        item.creatorRef = person.getRef()

    def itemRemoved(self, item):
        people = self.peopleFromMessage(item)
        for person in people or ():
            try:
                self.getPeoplePool().removeItem(person)
            except KeyError:
                # No such person. Probably a multiple spammer
                pass

    def itemChanged(self, item, notificaion=None):
        pass


## The web adapters really should be in quotient.web, but everything else does it like this so what the heck for now
class PeopleDisplay(Base):
    def renderFilterPreference(self, context, data):
        pref = IPrefs(self.original.store)
        current = pref.getPref('showAllPeople')
        selected = None
        if current is None or current == '0':
            selected = 'selected'
        return tags.form(action="/me/%s/freeform_post:setPref" % pref.storeID)[
            tags.input(type="hidden", name="pref", value="showAllPeople"),
            tags.select(name="value", onchange="submit()")[
                tags.option(value="1")["Show all people"],
                tags.option(value="0", selected=selected)["Show only people in address book"]
            ],
            tags.input(type="submit")
        ]

    def renderPeople(self, context, data):
        from quotient.web import pools
        pref = IPrefs(self.original.store)
        current = pref.getPref('showAllPeople')
        if current is None or current == '0':
            addBook = IAddressBook(self.original.store)
            pl = addBook.addressPool()
        else:
            pl = self.original.getPeoplePool()

        return pools.PoolBrowser(pl, renderAddPool=False)


    document = taglib.page(title="Address Book")[
        renderFilterPreference,
        renderPeople,
        freeform.configure,
        tags.a(href=lambda context, data: data.extractRef.storeID)[ "go see the people pool" ]
    ]


def recent(context,data):
    L = []
    for m in data.recentMessages():
        L.append(tags.div[tags.a(href=str(m.storeID))[
            taglib.headerToUTF8(m['subject'] or '<No Subject>')]])
    if L:
        return L
    return tags.div["no messages"]


def showMessages(context, data):
    from quotient.web import pools
    return pools.PoolBrowser(IPersonCollector(data.store).getMessagesPool(), data.recentMessages())

def renderGallery(context, data):
    from quotient.web import pools
    from quotient.smarts.image import IThumbnail
    ic = IThumbnail(data.store)
    snaps = ic.extractsFrom(data)
    snaps = list(snaps)
    return pools.PoolBrowser(ic.extractPool(), snaps, suggestedView="Gallery")

def histogram(context, data):
    return str(data.histogram())

def showcards(context,data):
    for card in data.cards:
        t = tags.div(_class="person-card")
        t[tags.div(_class="person-title")[card.title],
          tags.div(_class="person-org")[card.organization]]
        if card.phones:
            d1 = tags.div(_class="person-phones")
            t[d1]
            for label, phone in card.phones.items():
                d1[tags.span(_class="person-label")[label],
                   tags.span(_class="person-phone")[phone]]
        if card.emails:
            d2 = tags.div(_class="person-emails")
            t[d2]
            for label, email in card.emails.items():
                d2[tags.span(_class="person-label")[label],
                   tags.span(_class="person-email")[tags.a(href='compose?to='+email)[email]]]
            yield t


def showCards2(context, data):
    from quotient.web import pools
    return pools.PoolBrowser(data, data.cards)


class PersonDisplay(Base):
    document = taglib.page(title=("Person Detail: ", lambda context, data: taglib.headerToUTF8(data.name or '<No Name>')))[

#        tags.div(_class="recent")[tags.h3['Recent Messages'],
#                                  recent],
#        tags.div(_class="cards")[tags.h3['Contact Cards'],
#                                 showcards],
        showMessages,
        freeform.configure,
        showCards2,
        renderGallery,
        #histogram,
    ]

    def getChild(self, name, request):
        from quotient.web import root
        try:
            i = int(name)
        except ValueError:
            return root.FourOhFour(self.original)
        try:
            return CardDisplay(self.original.cards[i], self.original.store)
        except IndexError:
            return root.FourOhFour(self.original)


class CardDisplay(Base):
    adaptsAvatar = True
    ### HAAACK
    def __init__(self, card, avatar):
        Base.__init__(self, card)
        self.original = card
        ## ugggggh
        card.store = avatar
        self.avatar = avatar

    def backToPerson(self, context, data):
        return tags.div[
            tags.a(href="../%s" % self.original.person.storeID)[
                "back to %s" % self.original.person.name
            ]
        ]

    document = taglib.page(title="Card Detail")[
        backToPerson,
        freeform.configure
    ]

