
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""
System Storeup for handling required tasks such as sending bounces,
password reset information, etc.
"""

import gc

from atop import credup
from atop import powerup
from atop.sched import SubScheduler

from twisted.cred.checkers import ICredentialsChecker
from twisted.python import components
from twisted.python import log

from quotient import stats
from quotient import provisioner
from quotient.prefup import Prefup
from quotient.reset import PasswordResetter
from quotient.proto.smtpout import OutgoingMailup

from quotient.stats import formatSize

class SignupDisallowed(Exception):
    pass

class AdministrativeRestriction(SignupDisallowed):
    pass

def makeUser(store, username, password, powerups, free):
    # XXX This part sucks
    if free:
        tag = IControlPanel.USER_CREATION
    else:
        tag = IControlPanel.TICKET_REDEMPTION

    if not IControlPanel(store).actionAllowed(tag):
        raise AdministrativeRestriction("Signup is currently disabled")

    cc = ICredentialsChecker(store)
    a = credup.Avatar(store, username)
    ps = powerup.IPowerStation(a)
    pv = ps.installPowerup(provisioner.Provisioner)
    provisioner.oneYearService.grantTo(pv)

    prefs = ps.installPowerup(Prefup)
    for p, visible in powerups:
        pup = ps.installPowerup(p)
        if visible:
            prefs.makeVisible(pup)
    a.setInsecurePassword(password)
    a.setSecurePassword(password)
    personalize(a, username)
    cc.addAvatar(a)
    return a

def personalize(store, username):
    from quotient.smarts.person import IPersonCollector, Person
    pc = IPersonCollector(store, None)
    if pc is None:
        return
    from quotient.addressbook import IAddressBook
    from quotient.prefup import IPrefs
    person = IPersonCollector(store).newPerson(username)
    IPrefs(store).setMe(person)
    # Add me to the address book
    IAddressBook(store).addPerson(person)

systemUsers = [
    ("postmaster", (SubScheduler, OutgoingMailup)),
    ("janitor", (PasswordResetter,))
    ]

def makeSystemUsers(store):
    c = ICredentialsChecker(store)
    for name, pups in systemUsers:
        a = credup.Avatar(store, name)
        ps = powerup.IPowerStation(a)
        ps.installPowerup(provisioner.Provisioner)
        provisioner.oneYearService.grantTo(provisioner.IProvisioner(a))
        c.addAvatar(a)
        map(powerup.IPowerStation(a).installPowerup, pups)

class ISysup(components.Interface):
    pass

class Sysup(powerup.Storeup):
    __implements__ = (ISysup,)

    def setUpPools(self, store):
        makeSystemUsers(store)
        store.setComponent(ISysup, self)

class IControlPanel(components.Interface):
    USER_CREATION = 'user-creation'
    TICKET_REDEMPTION = 'ticket-redemption'

    MESSAGE_INDEXING = 'disable-indexing'
    EXTERNAL_GRABBING = 'disable-grabbing'

    def allowAction(self, action, yesno):
        pass

    def actionAllowed(self, action):
        pass

class ConvenientHackRemoveBeforeOnePointOh(dict):
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            self[key] = stats.AveragingResourceMeter(key, 60 * 15)
            return self[key]

class ControlPanel(powerup.Storeup):
    __implements__ = (powerup.Storeup.__implements__, IControlPanel)

    RESOURCE_METERS = ('transaction-commits', 'transaction-aborts',
                       'pickled', 'unpickled', 'cache-hits',
                       'page-renders', 'authorized-logins',
                       'unauthorized-logins', 'messages',
                       'bytes')

    SERIALIZED_LENGTH_WARNING = 8192

    def setUpPools(self, store):
        store.service.notifyOnStartup(self)
        self.startup()

        self.meters = ConvenientHackRemoveBeforeOnePointOh()
        for n in self.RESOURCE_METERS:
            self.meters[n] = stats.AveragingResourceMeter(n, 60 * 15)

        self.actions = {IControlPanel.USER_CREATION: True,
                        IControlPanel.TICKET_REDEMPTION: True,
                        IControlPanel.MESSAGE_INDEXING: True,
                        IControlPanel.EXTERNAL_GRABBING: True}
        store.setComponent(IControlPanel, self)

    def startup(self):
        log.addObserver(self.observe)

    def observe(self, events):
        # This might be optimized by adding a "channel" key
        # which could be dispatched upon in O(1) time

        if events['message']:
            return

        if 'db_put' in events:
            L = len(events['pickle'])
            if L > self.SERIALIZED_LENGTH_WARNING:
                log.msg("Warning: %-6d bytes pickled from %s" % (L, events['item']))
            self.meters['pickled'].record(L)
            return

        if 'db_get' in events:
            L = len(events['pickle'])
            if L > self.SERIALIZED_LENGTH_WARNING:
                log.msg("Warning: %-6d bytes unpickled from %s" % (L, events['item']))
            self.meters['unpickled'].record(L)
            return

        if 'db_txn' in events:
            if events['db_txn'] == 'commit':
                self.meters['transaction-commits'].record()
            elif events['db_txn'] == 'abort':
                self.meters['transaction-aborts'].record()
            return

        if 'db_cache' in events:
            if events['db_cache'] == 'hit':
                self.meters['cache-hits'].record()
            return

        if 'login' in events:
            if events['login'] is None:
                self.meters['unauthorized-logins'].record()
            else:
                self.meters['authorized-logins'].record()
            return

        if 'storage' in events:
            if events['storage'] == 'used':
                self.meters['bytes'].record(events['amount'])
                self.meters['messages'].record()

        if 'http_render' in events:
            self.meters['page-renders'].record()
            return

    def allowAction(self, action, yesno):
        self.touch()
        self.actions[action] = yesno

    def actionAllowed(self, action):
        return self.actions.setdefault(action, False)

class ISystemAdministrator(components.Interface):
    def shutdown(self):
        pass

class Adminup(powerup.Powerup):
    # Mostly, but not entirely, just a stub
    # Allows the ControlPanel for the Store to be rendered

    __implements__ = (powerup.Powerup.__implements__, ISystemAdministrator,)

    powerupName = "Administrative Controls"

    storeDependencies = [ControlPanel]

    def setUpPools(self, avatar):
        avatar.setComponent(ISystemAdministrator, self)

    def shutdown(self):
        from twisted.internet import reactor
        reactor.callLater(1, reactor.stop)


from nevow import tags
from nevow import freeform
from nevow import formless

from quotient import taglib
from quotient.web.root import Base

class ISystemSettings(formless.TypedInterface):
    allowUserCreation = formless.Boolean(label="New Users",
                                         description="Allow new users to register without claiming a ticket")
    allowTicketRedemption = formless.Boolean(label="Ticket Redemption",
                                             description="Allow new users to register by claiming a ticket")

    allowMessageIndexing = formless.Boolean(label="Fulltext Indexing",
                                            description="Allow newly received messages to be indexed.")
    allowExternalGrabbing = formless.Boolean(label="External Account Grabbing",
                                             description="Allow messages to be retrieved from external accounts.")

class IAdminDisplay(formless.TypedInterface):
    SystemSettings = ISystemSettings

    def shutdown(self):
        """Shut the server down"""
    shutdown = formless.autocallable(shutdown)

def renderMeter(name, period, format):
    def render(self, context, data):
        m = IControlPanel(self.original.store.store).meters[name]
        return format(m.average(period))
    return render

def makeResourceRow(name, format='%4.2f'.__mod__):
    f = lambda i: tags.td[renderMeter(name, i, format)]
    return tags.tr[tags.td[name], map(f, (60, 60 * 5, 60 * 15))]

class AdminDisplay(Base):
    __implements__ = (IAdminDisplay,)

    def render_garbage(self, context, data):
        return tags.p["Number of Garbage Objects:", len(gc.garbage)]

    document = taglib.page(title="Administrative Controls")[
        render_garbage,
        freeform.configure,
        tags.table()[tags.tr[tags.td["Resource Usage Averages"],
                             tags.td["1 min"], tags.td["5 min"], tags.td["15 min"]
                             ],
                     makeResourceRow("transaction-commits"),
                     makeResourceRow("transaction-aborts"),
                     makeResourceRow("pickled", formatSize),
                     makeResourceRow("unpickled", formatSize),
                     makeResourceRow("cache-hits"),
                     makeResourceRow("page-renders"),
                     makeResourceRow("authorized-logins"),
                     makeResourceRow("unauthorized-logins"),
                     makeResourceRow("messages"),
                     makeResourceRow("bytes", formatSize),
                     ]
        ]

    def action(tag):
        def get(self):
            cp = IControlPanel(self.original.store.store)
            return cp.actionAllowed(tag)
        def set(self, value):
            cp = IControlPanel(self.original.store.store)
            return cp.allowAction(tag, value)
        return get, set

    allowUserCreation = property(*action(IControlPanel.USER_CREATION))
    allowTicketRedemption = property(*action(IControlPanel.TICKET_REDEMPTION))

    allowMessageIndexing = property(*action(IControlPanel.MESSAGE_INDEXING))
    allowExternalGrabbing = property(*action(IControlPanel.EXTERNAL_GRABBING))

    def shutdown(self):
        return ISystemAdministrator(self.original).shutdown()
