# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import datetime

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from atop import store
from atop import sched
from atop import credup
from atop.test import unittest

from twisted.protocols import imap4
from twisted.cred.portal import Portal, IRealm
from twisted.cred.checkers import ICredentialsChecker

from quotient.test.emaildata import test_messages
from quotient.test import helper

from quotient import mimemessage
from quotient.proto import imapup
from quotient import grabbers
from quotient import searchup
from quotient import smartenup

class MsgSet(tuple):
    pass

class Loopback(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = (sched.SystemScheduler, credup.StoreRealm,)

    def setUpClass(self):
        grabbers.IMAP4MessageGrabber.allowInsecureLogin = True

    def tearDownClass(self):
        grabbers.IMAP4MessageGrabber.allowInsecureLogin = False

    def dbSetUp(self):
        self.avatar = self.createAvatar([imapup.IMAPAccountUp, grabbers.GrabberPowerup, sched.SubScheduler], name="USAR")
        self.avatar.setInsecurePassword("passWORD")
        self.avatar.setSecurePassword("PASSword")

        self.msg_strings = test_messages

        for m in self.msg_strings:
            mr = mimemessage.IMIMEDelivery(self.avatar)
            mr.createMIMEReceiver(False).feedStringNow(m)

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        nc = searchup.ISearch(self.avatar).checkpointDefer
        nc and nc.cancel()

    def dbtestWholeThing(self):
        p = Portal(IRealm(self.store))
        cc = ICredentialsChecker(self.store)
        cc.addDomain("localhost")
        cc.addAvatar(self.avatar, "localhost")
        p.registerChecker(cc)

        sFactory = imapup.IMAP4Factory(p)
        sFactory.debug = True
        from twisted.internet import reactor
        p = reactor.listenTCP(0, sFactory, interface="127.0.0.1")
        gp = grabbers.IGrabberAdder(self.avatar)
        cFactory = gp.addIMAPGrabber("USAR", "passWORD", "127.0.0.1", p.getHost()[2], "USAR@localhost", 0, False, None)

        val = unittest.deferredResult(cFactory.run())

        arrival = mimemessage.IMIMEDelivery(self.avatar).getArrivalRef().getItem()
        msgs = list(arrival.queryIndex('uid', result=store.ITEM))
        self.assertEquals(len(msgs), len(self.msg_strings) * 2)
        half = len(msgs) / 2

        begin, end = msgs[:half], msgs[half:]

        origs = []
        for x in begin:
            f = x.openFile()
            f.seek(x.bodyOffset)
            origs.append(f.read())
        news = []
        for x in end:
            f = x.openFile()
            f.seek(x.bodyOffset)
            news.append(f.read())
        for m in zip(origs, news):
            self.assertEquals(*m)


class Account(unittest.TestCase, helper.ProvisionerMixin):
    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        pups = (mimemessage.EmailPowerup, imapup.IMAPAccountUp)
        self.avatar = self.createAvatar(pups, name="Usar")

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        nc = searchup.ISearch(self.avatar).checkpointDefer
        nc and nc.cancel()

    def testLogout(self):
        acc = imap4.IAccount(self.avatar)
        acc.logout()

    def testCreate(self):
        acc = imap4.IAccount(self.avatar)

        self.assertRaises(imap4.MailboxCollision, acc.create, "inbox")

        self.failUnless(acc.create("foo"))
        p = self.avatar.rootPoolRef.getItem().getIndexItem("name", "foo")
        self.failUnless(isinstance(p, store.Pool))

    def testAddMailbox(self):
        acc = imap4.IAccount(self.avatar)
        self.assertRaises(RuntimeError, acc.addMailbox, "foo", "bar")

    def testSelect(self):
        acc = imap4.IAccount(self.avatar)
        arr = smartenup.IInbox(self.avatar).inboxRef.getItem()
        self.assertIdentical(acc.select("inbox").original, arr)
        self.assertEquals(acc.select("powerups").original.name, "powerups")
        self.assertEquals(acc.select("adoijwwodij"), None)

    def testDelete(self):
        acc = imap4.IAccount(self.avatar)
        self.assertRaises(imap4.MailboxException, acc.delete, "inbox")

    def testRename(self):
        acc = imap4.IAccount(self.avatar)
        self.assertRaises(imap4.MailboxException, acc.rename, "inbox", "foo")

    def testSubscriptions(self):
        acc = imap4.IAccount(self.avatar)

        self.failIf(acc.isSubscribed("foo"))

        self.avatar.transact(acc.subscribe, "foo")

        self.failUnless(acc.isSubscribed("Foo"))
        self.failUnless(acc.isSubscribed("foo"))

        self.avatar.transact(acc.unsubscribe, "Foo")

        self.failIf(acc.isSubscribed("Foo"))
        self.failIf(acc.isSubscribed("foo"))

        self.assertRaises(imap4.MailboxException, self.avatar.transact, acc.unsubscribe, "foo")

    def testListing(self):
        acc = imap4.IAccount(self.avatar)

        self.assertEquals([(m[0], m[1].original.name) for m in acc.listMailboxes("", "i%ox")], [("Inbox", "Inbox")])
        self.assertEquals([(m[0], m[1].original.name) for m in acc.listMailboxes("", "in%y")], [])

    def testNamespaces(self):
        acc = imap4.IAccount(self.avatar)

        self.assertEquals(acc.getPersonalNamespaces(), [["", None]])
        self.assertEquals(acc.getSharedNamespaces(), None)
        self.assertEquals(acc.getUserNamespaces(), None)

class Mailbox(unittest.TestCase, helper.ProvisionerMixin):
    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        pups = (mimemessage.EmailPowerup, imapup.IMAPAccountUp)
        self.avatar = self.createAvatar(pups, name="Usar")

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        nc = searchup.ISearch(self.avatar).checkpointDefer
        nc and nc.cancel()

    def dbtestUIDValidity(self):
        p = imapup.Poolbox(self.avatar, self.avatar.rootPoolRef.getItem(), False)
        u1 = p.getUIDValidity()
        p = imapup.Poolbox(self.avatar, self.avatar.rootPoolRef.getItem(), False)
        u2 = p.getUIDValidity()

        self.assertEquals(u1, u2)

        p = imapup.Poolbox(self.avatar, mimemessage.IMIMEDelivery(self.avatar).arrivalRef.getItem(), False)
        u3 = p.getUIDValidity()

        self.failIfEqual(u1, u3)

    def dbtestUIDNext(self):
        p = imapup.Poolbox(self.avatar, mimemessage.IMIMEDelivery(self.avatar).arrivalRef.getItem(), False)

        n = p.getUIDNext()
        self.assertEquals(n, p.getUIDNext())
        self.assertEquals(n, p.getUIDNext())
        self.assertEquals(n, p.getUIDNext())

        store.Item(self.avatar, name="foo")

        self.assertEquals(n + 1, p.getUIDNext())
    dbtestUIDNext.todo = "getUIDNext seems to actually allocate a UID, rather than just peeking"

    def dbtestAddMessage(self):
        p = imapup.Poolbox(self.avatar, mimemessage.IMIMEDelivery(self.avatar).arrivalRef.getItem(), True)

        d = p.addMessage(StringIO(test_messages[0]), ('\\Answered',))
        m = unittest.deferredResult(d)

        a = list(mimemessage.IMIMEDelivery(self.avatar).arrivalRef.getItem().queryIndex())
        self.assertEquals([m], [e.index_uid(p.original) + 1 for e in a])

    def dbtestStore(self):
        md = mimemessage.IMIMEDelivery(self.avatar)
        msg1 = md.createMIMEReceiver(True).feedStringNow(test_messages[0])
        msg2 = md.createMIMEReceiver(True).feedStringNow(test_messages[1])

        p = imapup.Poolbox(self.avatar, md.arrivalRef.getItem(), True)

        msg1uid = msg1.index_uid(p.original) + 1
        msg2uid = msg2.index_uid(p.original) + 1

        # This will set \\Seen on both messages
        self.failUnless(p.store(MsgSet((msg1uid, msg2uid)), ('\\Seen',), 1, True))

        self.assertEquals(msg1.flags, msg1.SEEN)
        self.assertEquals(msg2.flags, msg2.SEEN)

        # This will unset \\Seen on just msg1
        self.failUnless(p.store(MsgSet((msg1uid,)), ('\\Seen',), -1, True))

        self.assertEquals(msg1.flags, 0)
        self.assertEquals(msg2.flags, msg2.SEEN)

        # This will set flags to \\Answered on msg2
        self.failUnless(p.store(MsgSet((msg2uid,)), ('\\Answered',), 0, True))

        self.assertEquals(msg1.flags, 0)
        self.assertEquals(msg2.flags, msg2.ANSWERED)

        # This will set \\Seen on both messages
        self.failUnless(p.store(MsgSet((1, 2)), ('\\Seen',), 1, False))

        self.assertEquals(msg1.flags, msg1.SEEN)
        self.assertEquals(msg2.flags, msg2.SEEN | msg2.ANSWERED)

class Protocol(unittest.TestCase):

    def dbSetUp(self):
        # Fake a response from the server, make sure only the right
        # UIDs are selected.

        self.f = grabbers.IMAP4MessageGrabber().filterOld

        self.results = {
            1: '01-Jan-2000 SomejunkyTimeSpec GMT',
            2: '02-Jan-2000 AnotherJunkySpec EST',
            3: '03-Jan-2000 Whenever DST',
            4: '04-Jan-2000 HuhHuhHuh PST',
            5: '30-Jan-2000 TimeTimeTime UTC',
            6: '31-Jan-2000 MoreTime CST',
            7: '01-Feb-2000 HowAboutThis MST'}

    def testDeletionFiltering(self):
        uids = self.f(self.results, 7, datetime.date(2000, 1, 4))
        self.assertEquals(uids, [])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 7))
        self.assertEquals(uids, [])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 8))
        self.assertEquals(uids, [])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 9))
        self.assertEquals(uids, [1])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 10))
        uids.sort()
        self.assertEquals(uids, [1, 2])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 11))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 12))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 13))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 31))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4])

        uids = self.f(self.results, 7, datetime.date(2000, 3, 1))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4, 5, 6, 7])
