# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time

from os.path import join as opj

from twisted.cred import portal
from twisted.protocols import pop3

from atop import sched
from atop import store
from atop import credup
from atop import powerup
from atop.test import unittest

from quotient.test import helper
from quotient.test import emaildata
from quotient import mimemessage
from quotient.proto import popup
from quotient import grabbers

class TestPopup(unittest.TestCase, helper.ProvisionerMixin):

    def dbSetUp(self):
        self.msgsObjs = []
        self.avatar = self.createAvatar([popup.Popup])
        self.popup = pop3.IMailbox(self.avatar)
        for m in emaildata.test_messages:
            mr = mimemessage.IMIMEDelivery(self.avatar).createMIMEReceiver(False)
            self.msgsObjs.append(mr.feedStringNow(m))
        self.popup.login()

    def testLogin(self):
        self.assertEquals(self.popup.msgs, [o.storeID for o in self.msgsObjs])
        self.assertRaises(popup.AlreadyLocked, self.popup.login)

    def testListMessageSizes(self):
        self.assertEquals(
            list(self.popup.listMessageSizes()),
            [len(m) for m in emaildata.test_messages])

    def testGetMessageSize(self):
        self.assertEquals(
            [self.popup.getMessageSize(i) for i in range(len(self.msgsObjs))],
            [len(m) for m in emaildata.test_messages])

    def testGetMessage(self):
        self.assertEquals(
            [self.popup.getMessage(i).read() for i in range(len(self.msgsObjs))],
            emaildata.test_messages)

    def testGetUidl(self):
        d = dict.fromkeys([self.popup.getUidl(i) for i in range(len(self.msgsObjs))])
        self.assertEquals(len(d), len(self.msgsObjs))

    def testListMessages(self):
        self.assertEquals(
            list(self.popup.listMessages(None)),
            list(self.popup.listMessageSizes()))
        self.assertEquals(
            [self.popup.listMessages(i) for i in range(len(self.msgsObjs))],
            [self.popup.getMessageSize(i) for i in range(len(self.msgsObjs))])

    def dbtestDeleteMessage(self):
        self.popup.deleteMessage(0)
        self.popup.logout()
        self.popup.login()
        self.assertEquals(len(self.popup.msgs), len(self.msgsObjs) - 1)

    def testUndeleteMessages(self):
        self.popup.deleteMessage(0)
        self.popup.undeleteMessages()
        self.popup.logout()
        self.popup.login()
        self.assertEquals(len(self.popup.msgs), len(self.msgsObjs))

class FakeScheduler:
    def callItemLater(self, item):
        pass
    def clock(self):
        return 0

class LoopbackTest(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = (credup.StoreRealm,)

    def dbSetUp(self):
        self.avatar = self.createAvatar([popup.Popup, grabbers.GrabberPowerup], name="user")
        self.avatar.setInsecurePassword("pass")
        self.avatar.setSecurePassword("pass")
        self.popup = pop3.IMailbox(self.avatar)
        for m in emaildata.test_messages:
            mr = mimemessage.IMIMEDelivery(self.avatar).createMIMEReceiver(False)
            mr.feedStringNow(m)
        self.avatar.setComponent(sched.IScheduler, FakeScheduler())

    def dbtestWholeThing(self):
        from twisted.internet import reactor

        p = portal.Portal(credup.IRealm(self.store))
        cc = credup.ICredentialsChecker(self.store)
        cc.addDomain("localhost")
        cc.addAvatar(self.avatar)
        p.registerChecker(cc)
        sFactory = popup.POP3Factory(p)
        p = reactor.listenTCP(0, sFactory, interface="127.0.0.1")
        gp = grabbers.IGrabberAdder(self.avatar)
        cFactory = gp.addPopGrabber('user', 'pass', '127.0.0.1', p.getHost()[2], "user@localhost", 0, False, None)

        val = unittest.deferredResult(cFactory.run())
        reactor.iterate()

        arrival = mimemessage.IMIMEDelivery(self.avatar).getArrivalRef().getItem()
        msgs = list(arrival.queryIndex('uid', result=store.ITEM))
        self.assertEquals(len(msgs), len(emaildata.test_messages) * 2)
        half = len(msgs) / 2

        begin, end = msgs[:half], msgs[half:]

        self.assertEqual([x.openFile().read() for x in begin], [x.openFile().read() for x in end])

class Protocol(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        self.a = self.createAvatar([grabbers.GrabberPowerup])
        self.f = grabbers.IGrabberAdder(self.a).addPopGrabber("user", "pass", "host", 1234, "", 0, 0, 1)
        N = 10
        self.failUnless(N >= len(emaildata.test_messages))
        mr = mimemessage.MIMEMessageReceiver(self.a, lambda m: None, True)
        msgsList = emaildata.test_messages[:N]
        now = time.time()
        dates = [(now - (i * 60 * 60 * 24)) for i in range(N, 0, -1)]
        self.now = time.gmtime(now)
        uids = map(str, range(100, 100 + N))
        for (m, d, u) in zip(emaildata.test_messages[:N], dates, uids):
            msg = mr.feedStringNow(m)
            msg.dateReceived = dates.pop()
            self.f.messageReceived(msg, 100, uids.pop())

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        pass

    def testMessageReceived(self):
        self.f.age = 7
        self.assertEquals(self.f.getDeletionList(self.now),
                          ["100", "101", "102", "103"])

        self.f.age = 6
        self.assertEquals(self.f.getDeletionList(self.now),
                          ["104"])

        self.f.age = 5
        self.assertEquals(self.f.getDeletionList(self.now),
                          ["105"])

        self.f.age = 2
        self.assertEquals(self.f.getDeletionList(self.now),
                          ["106", "107", "108"])

        self.f.age = 0
        self.assertEquals(self.f.getDeletionList(self.now), [])

        self.f.age = 1
        self.assertEquals(self.f.getDeletionList(time.gmtime(time.mktime(self.now) + (60 * 60 * 24))),
                          ["109"])
