
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.protocols import smtp
from twisted.internet import defer
from twisted.protocols.dns import Record_MX
from twisted.python import failure

from atop import credup
from atop import powerup
from atop import sched
from atop.store import ITEM
from atop.test import unittest

from quotient.test import emaildata, helper

from quotient import provisioner
from quotient.proto import smtpin
from quotient.proto import smtpout
from quotient.mimemessage import IMIMEDelivery, MIMEMessageReceiver, EmailPowerup

class Delivery(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [credup.StoreRealm]

    def dbSetUp(self):
        self.checker = credup.ICredentialsChecker(self.store)
        self.avatar = self.createAvatar([EmailPowerup], "USAR")

        self.checker.addDomain("DOM")
        self.checker.addAvatar(self.avatar, "DOM")

        self.delivery = smtpin.QuotientMessageDelivery(self.checker)

    def u(self, s):
        return smtp.User(smtp.Address(s), "HELO", None, smtp.Address("foo@bar"))

    def testValidateFrom(self):
        origin = "this is a bogus string.  we currently accept all origins."
        self.assertEquals(self.delivery.validateFrom("EHLO FROM ME", origin), origin)
        
    def testInvalidValidateTo(self):
        self.assertRaises(smtp.SMTPBadRcpt, self.delivery.validateTo, self.u("bogus!"))
        self.assertRaises(smtp.SMTPBadRcpt, self.delivery.validateTo, self.u("user@domain"))

    def dbtestExceededQuota(self):
        for i in [10000, 100, 1]:
            try:
                while True:
                    provisioner.IProvisioner(self.avatar).deplete(provisioner.ETank.QUOTA, i)
            except provisioner.NotEnoughFuel:
                pass
        self.assertRaises(smtp.SMTPBadRcpt, self.delivery.validateTo, self.u("USAR@DOM"))
    
    def dbtestSuccessfulDelivery(self):
        md = self.delivery.validateTo(self.u("USAR@DOM"))
        msg = md()
        for line in emaildata.test_messages[0].splitlines():
            msg.lineReceived(line)
        d = unittest.deferredResult(msg.eomReceived())
        self.failUnless(d)

        p = IMIMEDelivery(self.avatar).arrivalRef.getItem()
        msgs = self.store.transact(list, p.queryIndex(result=ITEM))
        self.assertEquals(len(msgs), 1)
        self.assertEquals(msgs[0].openFile().read(), emaildata.test_messages[0])

class MXHijacker:
    def __init__(self, hosts):
        self.hosts = hosts
        
    def getMX(self, dname):
        try:
            return self.hosts[dname] # defer.succeed(self.hosts[dname])
        except KeyError:
            return defer.fail(failure.Failure())

class Transmission(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [credup.StoreRealm, sched.SystemScheduler]

    def dbSetUp(self):
        self.checker = credup.ICredentialsChecker(self.store)
        self.avatar = self.createAvatar([smtpout.OutgoingMailup, EmailPowerup], "USAR")

        self.xmit = smtpout.IOutgoingMail(self.avatar)
        # self.xmit.mx.resolver.resolvers[2]._parseCall.cancel()

        mr = MIMEMessageReceiver(self.avatar, lambda m: None, True)
        self.msg = mr.feedStringNow(emaildata.test_messages[0])
        
    def dbTearDown(self):
        dc = sched.IScheduler(self.store).nextCall
        dc and dc.cancel()
        
    def dbtestConvenience(self):
        from twisted.internet import reactor
        d = defer.Deferred()
        self.xmit.mx = MXHijacker({'thathost': d})
        L = []
        self.xmit.sendmail = lambda *a: defer.succeed(L.append(a))

        D = self.xmit.sendString("HELLO WORLD", [], 'thatguy@thathost', 'thisguy@thishost'
            ).addBoth(lambda x: reactor.crash())
        def _():
            d.callback(Record_MX(exchange="mxhost"))
        reactor.callLater(0, self.store.transact, _)
        reactor.run()
        

    def testSendingMailGoesToTransitPool(self):
        # Jack the MX process so we can intercept the email
        d = defer.Deferred()
        self.xmit.mx = MXHijacker({'thathost': d})

        L = []
        self.xmit.sendmail = lambda *a: defer.succeed(L.append(a))

        def _():
            self.xmit.sendMIMEMessage("thisguy", "thatguy", "thathost", self.msg)
        self.store.transact(_)

        def _():
            out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
            trans = list(self.xmit.transRef.getItem().queryIndex(result=ITEM))
            sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

            self.assertEquals(out, [self.xmit.transRef.getItem()])
            self.assertEquals(trans, [self.msg])
            self.assertEquals(sent, [])
        self.store.transact(_)

        def _():
            d.callback(Record_MX(exchange="mxhost"))
        self.store.transact(_)

        def _():
            out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
            trans = list(self.xmit.transRef.getItem().queryIndex(result=ITEM))
            sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

            self.assertEquals(out, [self.xmit.transRef.getItem()])
            self.assertEquals(sent, [self.msg])
            self.assertEquals(trans, [])
        self.store.transact(_)

    def testFailedSendStaysInOutbox(self):
        # Jack the MX process so we can intercept the email
        d = defer.Deferred()
        self.xmit.mx = MXHijacker({'thathost': d})

        L = []
        self.xmit.sendmail = lambda *a: defer.succeed(L.append(a))

        def _():
            return self.xmit.sendMIMEMessage("thisguy", "thatguy", "thathost", self.msg)
        msgSend = self.store.transact(_)

        def _():
            out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
            trans = list(self.xmit.transRef.getItem().queryIndex(result=ITEM))
            sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

            self.assertEquals(out, [self.xmit.transRef.getItem()])
            self.assertEquals(trans, [self.msg])
            self.assertEquals(sent, [])
        self.store.transact(_)

        def _():
            d.errback(failure.Failure(smtp.SMTPDeliveryError(450, "Transient forged failure")))
        self.store.transact(_)

        unittest.deferredResult(msgSend)

        def _():
            out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
            trans = list(self.xmit.transRef.getItem().queryIndex(result=ITEM))
            sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

            self.assertEquals(out, [self.xmit.transRef.getItem(), self.msg])
            self.assertEquals(sent, [])
            self.assertEquals(trans, [])
        self.store.transact(_)
        d.result = None    # I don't owe you for World any more, JML -glyph

    def testPermanentFailedGeneratesBounce(self):
        # Jack the MX process so we can intercept the email
        d = defer.Deferred()
        self.xmit.mx = MXHijacker({'thathost': d})

        L = []
        self.xmit.sendmail = lambda *a: defer.succeed(L.append(a))

        def _():
            return self.xmit.sendMIMEMessage("thisguy", "thatguy", "thathost", self.msg)
        msgSend = self.store.transact(_)

        def _():
            out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
            trans = list(self.xmit.transRef.getItem().queryIndex(result=ITEM))
            sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

            self.assertEquals(out, [self.xmit.transRef.getItem()])
            self.assertEquals(trans, [self.msg])
            self.assertEquals(sent, [])
        self.store.transact(_)

        def _():
            d.errback(failure.Failure(smtp.SMTPDeliveryError(550, "Permanent forged failure")))
        self.store.transact(_)

        unittest.deferredResult(msgSend)

        def _():
            out = list(self.xmit.outRef.getItem().queryIndex(result=ITEM))
            trans = list(self.xmit.transRef.getItem().queryIndex(result=ITEM))
            sent = list(self.xmit.sentRef.getItem().queryIndex(result=ITEM))

            self.assertEquals(out, [self.xmit.transRef.getItem()])
            self.assertEquals(sent, [])
            self.assertEquals(trans, [])

            inbox = list(IMIMEDelivery(self.avatar).arrivalRef.getItem().queryIndex())
            self.assertEquals(len(inbox), 1)
        self.store.transact(_)
        d.result = None    # I don't owe you for World any more, JML -glyph

class Factories(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        self.avatar = self.createAvatar([EmailPowerup, smtpout.OutgoingMailup], "av1")

    def dbTearDown(self):
        dc = sched.IScheduler(self.store).nextCall
        dc and dc.cancel()
        
    def testInFactoryBuildProtocol(self):
        portal = object()
        f = smtpin.ESMTPInFactory(self.store, portal)
        f.doStart()
        p = f.buildProtocol(("remotehost", 12345))
        self.assertIdentical(p.factory, f)
        self.assertIdentical(p.delivery, f.delivery)
        self.assertIdentical(p.portal, portal)
