# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
import sys
import time
import shutil

from os.path import join as opj
from os.path import isdir

from atop.credup import Avatar, StoreRealm
from atop.store import Store, Item, SubStore
from atop.store import Pool, DateIndex, UIDIndex
from atop.store import ITEM, INDEX, STORE_ID
from atop.store import Inconsistency, NameIndex
from atop import sched
from atop.test import unittest

from quotient import mimemessage
from quotient.searchup import Searchup, ISearch
from quotient.proto import smtpin
from quotient.test import helper

from twisted.cred.portal import Portal
from twisted.cred.credentials import UsernamePassword
from twisted.protocols import imap4
from twisted.protocols import smtp
from twisted.python.components import Interface
from twisted.python import log
from twisted.python.util import sibpath

mailDirectory = os.environ.get("BENCH_MESSAGES") or "/home/amir/test/test150"

nCopies = int(os.environ.get('N_COPIES') or 10)
nRuns = int(os.environ.get('N_RUNS') or 5)
nMsgs = int(os.environ.get('N_MESSAGES') or 100)

class BenchmarkStore(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [StoreRealm, sched.SystemScheduler]

    nCopies = nCopies

    mailDirectory = mailDirectory

    def dbSetUp(self):
        self.msgs = []
        avatar = self.createAvatar(())

        for f in os.listdir(self.mailDirectory):
            mp = mimemessage.MIMEMessageReceiver(avatar, self.msgs.append, True)
            mp.feedFileNow(file(opj(self.mailDirectory, f)))

    def benchmarkMessageInsertion(self):
        p = self.mktemp()
        s = Store(0, opj(p, "db"), opj(p, "files"))

        # Insert all the messages nCopies times
        def _():
            t = time.time()
            for i in range(self.nCopies):
                a = s.transact(SubStore, s)
                for m in self.msgs:
                    path = a.storeID, a.nextID()
                    s.transact(s._saveItemToPath, m, s.storeID, a.storeID, a.nextID())
            return time.time() - t
        t = s.transact(_)
        self.recordStat({"throughput": (len(self.msgs) / t, "Messages_per_sec")})


class BloatBenchmark(unittest.TestCase, helper.ProvisionerMixin):
    mailDirectory = mailDirectory

    storePowerups = [sched.SystemScheduler]

    def dbTearDown(self):
        pass
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()

    def benchmarkBloat(self):
        a = self.createAvatar([mimemessage.EmailPowerup, Searchup])
        lup = ISearch(a)
        for f in os.listdir(self.mailDirectory):
            mp = mimemessage.IMIMEDelivery(a).createMIMEReceiver(True)
            mp.feedFileNow(file(opj(self.mailDirectory, f)))
        #lup.indexer.optimize()
        lup.indexer.close()
        lup.checkpointDefer.cancel()

        def sizeOf(dname):
            s = 0
            for fname in os.listdir(dname):
                fname = opj(dname, fname)
                s += os.stat(fname).st_size
                if isdir(fname):
                    s += sizeOf(fname)
            return s
        totSize = sizeOf(os.path.join(self.storePath, self.dbPath, '..'))
        fSize = float(sizeOf(mailDirectory))
        self.recordStat({"bloat": (((totSize-fSize) / fSize), "DB_bloat")})

from twisted.protocols import loopback
from twisted.test.proto_helpers import LineSendingProtocol
from quotient.mimemessage import IMIMEDelivery
from quotient.proto import popup
from quotient.proto import imapup
from twisted.cred.portal import Portal
from twisted.cred.portal import IRealm
from twisted.cred.checkers import ICredentialsChecker

class SneakyPOP3Server(popup.LoggingInPOP3):
    totalFileTransfers = 0

    def __init__(self, portal):
        self.portal = portal

    def finishedFileTransfer(self, lastsent):
        r = popup.LoggingInPOP3.finishedFileTransfer(self, lastsent)
        self.totalFileTransfers += 1
        return r

class SneakyIMAP4Server(imap4.IMAP4Server):
    totalFileTransfers = 0

    def spew_body(self, *a, **kw):
        self.totalFileTransfers += 1
        return imap4.IMAP4Server.spew_body(self, *a, **kw)

class EmailThroughput(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [StoreRealm, sched.SystemScheduler]

    nRuns = nRuns
    nMsgs = nMsgs

    mailDirectory = mailDirectory

    def dbSetUp(self):
        self.avatar = self.createAvatar([popup.Popup, imapup.IMAPAccountUp], "USER")
        self.avatar.setInsecurePassword("PASS")
        self.avatar.setSecurePassword("PASS")

        ICredentialsChecker(self.store).addDomain("DOMAIN")
        ICredentialsChecker(self.store).addAvatarToDomain(self.avatar, "DOMAIN")

        msgs = os.listdir(self.mailDirectory)

        assert len(msgs) > self.nMsgs
        for msg in msgs[:self.nMsgs]:
            msg = opj(self.mailDirectory, msg)
            IMIMEDelivery(self.avatar).createMIMEReceiver(True).feedFileNow(file(msg))

    def dbTearDown(self):
        dc = sched.IScheduler(self.store).nextCall
        dc and dc.cancel()
        dc = ISearch(self.avatar).checkpointDefer
        dc and dc.cancel()

    def benchmarkPOP3Server(self):
        gets = ["RETR %d" % x for x in range(self.nMsgs)]
        portal = Portal(IRealm(self.store))
        checker = ICredentialsChecker(self.store)
        portal.registerChecker(checker)

        totWall = 0
        totCPU = 0

        count = self.nRuns * self.nMsgs

        for i in range(self.nRuns):
            s = SneakyPOP3Server(portal)
            c = LineSendingProtocol([
                "USER USER@DOMAIN", "PASS PASS"] + gets + ["QUIT"])
            startCPU = time.clock()
            startWall = time.time()
            loopback.loopbackTCP(s, c)
            endCPU = time.clock()
            endWall = time.time()

            self.assertEquals(s.totalFileTransfers, self.nMsgs)

            totCPU += endCPU - startCPU
            totWall += endWall - startWall

        cpuResult = 0
        if totCPU != 0:
            cpuResult = count / totCPU
        wallResult = 0
        if totWall != 0:
            wallResult = count / totWall

        if wallResult == 0 or cpuResult == 0:
            log.msg("Something screwed up happened, but I don't know what")

        self.recordStat({"CPU": (cpuResult, "POP3_Served_Msgs_per_cpu_sec"),
                         "Wallclock": (wallResult, "POP3_Served_Msgs_per_wall_sec")})

    def benchmarkIMAP4Server(self):
        portal = Portal(IRealm(self.store))
        checker = ICredentialsChecker(self.store)
        portal.registerChecker(checker)

        totWall = 0
        totCPU = 0

        count = self.nRuns * self.nMsgs

        for i in range(self.nRuns):
            s = SneakyIMAP4Server()
            s.portal = portal
            c = LineSendingProtocol([
                "01 LOGIN USER@DOMAIN PASS",
                "02 EXAMINE inbox",
                "03 FETCH 1:100 (UID RFC822.SIZE BODY[] FLAGS)",
                "04 LOGOUT",
                ])
            c.MAX_LENGTH = sys.maxint
            startCPU = time.clock()
            startWall = time.time()
            loopback.loopbackTCP(s, c)
            endCPU = time.clock()
            endWall = time.time()

            self.assertEquals(s.totalFileTransfers, self.nMsgs)

            totCPU += endCPU - startCPU
            totWall += endWall - startWall

        cpuResult = 0
        if totCPU != 0:
            cpuResult = count / totCPU
        wallResult = 0
        if totWall != 0:
            wallResult = count / totWall

        if wallResult == 0 or cpuResult == 0:
            log.msg("Something screwed up happened, but I don't know what")

        self.recordStat({"CPU": (cpuResult, "IMAP4_Served_Msgs_per_cpu_sec"),
                         "Wallclock": (wallResult, "IMAP4_Served_Msgs_per_wall_sec")})


from twisted.test.proto_helpers import LineSendingProtocol
class DumbSMTPClient(LineSendingProtocol):
    def __init__(self, recip, msg):
        L = ["HELO me", "MAIL FROM: origin@domain",
             "RCPT TO: " + recip, "DATA"]
        # Limit this to 1000 lines of data, because our stupid send
        # behavior screws up with larger data sets.
        L.extend([line.lstrip(".") for line in msg.splitlines()][:1000])
        L.append(".")
        LineSendingProtocol.__init__(self, L)

    def lineReceived(self, line):
        if line.startswith("250"):
            self.transport.loseConnection()

class SMTPReceptionBenchmark(unittest.TestCase, helper.ProvisionerMixin):
    mailDirectory = mailDirectory

    storePowerups = [StoreRealm]

    nRuns = nRuns
    nMsgs = nMsgs

    def dbSetUp(self):
        self.avatar = self.createAvatar([mimemessage.EmailPowerup], "AV2")
        ICredentialsChecker(self.store).addDomain("localhost")
        ICredentialsChecker(self.store).addAvatarToDomain(self.avatar, "localhost")
        self.server = smtp.ESMTP({}, None)
        self.server.delivery = smtpin.QuotientMessageDelivery(self.store)

    def benchmarkMessageReception(self):
        totWall = 0
        totCPU = 0

        count = self.nRuns * self.nMsgs

        arr = mimemessage.IMIMEDelivery(self.avatar).getArrivalRef().getItem()

        for i in range(self.nRuns):
            for m in os.listdir(self.mailDirectory)[:self.nMsgs]:
                m = opj(self.mailDirectory, m)
                c = DumbSMTPClient("AV2@localhost", file(m).read())
                startWall = time.time()
                startCPU = time.clock()
                loopback.loopbackTCP(self.server, c)
                endWall = time.time()
                endCPU = time.clock()


                totCPU += endCPU - startCPU
                totWall += endWall - startWall

            def _():
                self.assertEquals((i + 1) * self.nMsgs, len(list(arr.queryIndex())))
            self.store.transact(_)

        cpuResult = 0
        if totCPU != 0:
            cpuResult = count / totCPU
        wallResult = 0
        if totWall != 0:
            wallResult = count / totWall

        if wallResult == 0 or cpuResult == 0:
            log.msg("Something screwed up happened, but I don't know what")

        self.recordStat({"CPU": (cpuResult, "SMTP_Received_Msgs_per_cpu_sec"),
                         "Wallclock": (wallResult, "SMTP_Received_Msgs_per_wall_sec")})
