# -*- coding: Latin-1 -*-

import socket

from server import IChatService
from auth import Avatar

from twisted.protocols import irc
from twisted.internet import defer
from twisted.internet import protocol
from twisted.python import log
from twisted.python import failure
from twisted.python import components
from twisted.application import service
from twisted.application import strports

from twisted import cred
import twisted.cred.credentials

class IChatUser(components.Interface):
    """
    @ivar nickname: This users network handle
    """

    def password(self, password):
        """Set the password associated with this connection.
        
        Protocols with real authentication will want to make this a no-op.
        """
    
    def user(self, user, mode, unused, realname):
        """Set the username and other associated metadata."""
    
    def nick(self, nickname):
        """Change this user's nickname."""

class IRCServerError(Exception):
    pass

class UnknownCommand(IRCServerError):
    code = irc.ERR_UNKNOWNCOMMAND

class AlreadyRegistered(IRCServerError):
    code = irc.ERR_ALREADYREGISTRED

class ErroneousNickname(IRCServerError):
    code = irc.ERR_ERRONEUSNICKNAME

class NicknameInUse(IRCServerError):
    code = irc.ERR_NICKNAMEINUSE

class NickCollision(IRCServerError):
    code = irc.ERR_NICKCOLLISION

class UnavailableResource(IRCServerError):
    code = irc.ERR_UNAVAILRESOURCE

class Restricted(IRCServerError):
    code = irc.ERR_RESTRICTED

class NoOperHost(IRCServerError):
    code = irc.ERR_NOOPERHOST

class PasswordMismatch(IRCServerError):
    code = irc.ERR_PASSWDMISMATCH

class UnknownModeFlag(IRCServerError):
    code = irc.ERR_UMODEUNKNOWNFLAG

class UsersDontMatch(IRCServerError):
    code = irc.ERR_USERSDONTMATCH

def _unknown(*args, **kw):
    raise UnknownCommand()

class AnonymousChatUser:
    __implements__ = (IChatUser,)
    
    passwd = None
    nickname = ''
       
    def __init__(self, proto):
        self.proto = proto
    
    def password(self, password):
        self.passwd = password
    
    def user(self, user, mode, unused, realname):
        if not self.passwd or self.nickname:
            raise Restricted()
        self.nickname = user
        c = cred.credentials.UsernamePassword(user, self.passwd)
        return self.proto.portal.login(c, None, IChatUser
            ).addCallbacks(self._cbLogin, self._ebLogin
            )
    
    def _cbLogin(self, (i, a, l)):
        a.proto = self.proto
        self.proto.avatar = a
        self.proto._logout = l
        a.loggedIn()
    
    def _ebLogin(self, f):
        f.trap(cred.error.UnauthorizedLogin)
        return failure.Failure(PasswordMismatch())

    for f in ('nick', 'oper', 'userMode', 'channelMode', 'service', 'quit',
        'squit', 'join', 'part', 'topic', 'names', 'channels', 'invide',
        'admin', 'info', 'privmsg', 'notice', 'kick', 'motd', 'lusers',
        'version', 'status', 'time'):
        exec "%s = _unknown" % (f,)

class AuthenticatedChatUser:
    __implements__ = (IChatUser,)

    client = None
    nickname = ''
    password = user = _unknown
    
    def __init__(self, avatar):
        self.avatar = avatar

    def loggedIn(self):
        self.proto.sendMessage(irc.RPL_WELCOME,
            ":Welcome to this server dohicky")
        self.proto.sendMessage(irc.RPL_YOURHOST,
            ":Your host is me, running version 2")
        self.proto.sendMessage(irc.RPL_CREATED,
            ":This server was created by a very sick man.")
        self.proto.sendMessage(irc.RPL_MYINFO,
            "blah", "blah", "blah", "blah")

    def setClient(self, client):
        print 'Setting client to', client
        self.client = client
        self.account = client.account
        self.nickname = self.client.nickname

    def nick(self, nickname):
        raise UnknownCommand() # XXX Sigh
    
    def oper(self, name, passwd):
        raise UnknownCommand() # This is foolish, anyway.
    
    def userMode(self, set, modes):
        raise UnknownCommand() # XXX Sigh
    
    def channelMode(self, set, modes, args):
        raise UnknownCommand() # XXX Sigh
    
    def service(self, nick, res1, distr, type, res2, info):
        raise UnknownCommand() # We won't proxy services, I guess
    
    def quit(self, msg):
        # t.im is somewhat incomplete.
        # self.account.logOff()
        self.proto.transport.loseConnection()
    
    def squit(self, server, comment):
        raise UnknownCommand() # No way!
    
    def join(self, channels, keys):
        for c in channels:
            self.client.joinGroup(c)
    
    def topic(self, channel, topic):
        raise UnknownCommand() # XXX Cache setTopic() calls and regurgitate

    def names(self, channel):
        raise UnknownCommand() # XXX Cache setMembers() calls and regurgitate
    
    def channels(self, channels, server):
        raise UnknownCommand() # XXX Sigh
    
    def admin(self, server):
        raise UnknownCommand() # We won't support this
    
    def info(self, server):
        raise UnknownCommand() # Nor this
    
    def privmsg(self, recipients, message): # XXX Support n/m/o/v
        for r in recipients:
            if r[0] in irc.CHANNEL_PREFIXES:
                self.account.getGroup(r).sendGroupMessage(message)
            else:
                self.client.getPerson(r).sendMessage(message)
    notice = privmsg

    def kick(self, channel, users):
        raise UnknownCommand() # XXX Sigh
    
    def motd(self, server):
        raise UnknownCommand() # XXX Sigh
    
    def lusers(self, mask, target):
        raise UnknownCommand() # XXX Sigh
    
    def version(self, target):
        raise UnknownCommand() # XXX Sigh
    
    def stats(self, query, target):
        raise UnknownCommand() # XXX Sigh
    
    def links(self, server, mask):
        raise UnknownCommand() # XXX Sigh
    
    def time(self, server):
        raise UnknownCommand() # XXX Sigh
components.registerAdapter(AuthenticatedChatUser, Avatar, IChatUser)

class IRCServer(irc.IRC):
    OPERATOR = 'o'
    VOICE = 'v'
    MODERATED = 'm'
    NO_OUTSIDE_MSGS = 'n'
    INVITE = 'i'

    avatar = None
    _logout = None

    def __init__(self, portal, hostname = socket.getfqdn()):
        self.portal = portal
        self.hostname = hostname
        self.avatar = AnonymousChatUser(self)

    def connectionLost(self, reason):
        if self._logout:
            self._logout()
            self.avatar = self._logout = None
    
    def _cbLogOn(self, (i, a, l)):
        self.avatar = a
        self._logout = l
    
    def sendMessage(self, *args, **kw):
        kw['prefix'] = self.hostname
        return irc.IRC.sendMessage(self, *args, **kw)

    def _protectCall(self, command, f, *args, **kw):
        defer.maybeDeferred(f, *args, **kw
            ).addCallback(self._cbProtected,
            ).addErrback(self._ebProtected, command
            ).addErrback(self._ebUnexpected, command
            )
    
    def _cbProtected(self, result):
        for r in result or ():
            self.sendMessage(*r)

    def _ebProtected(self, failure, command):
        failure.trap(IRCServerError)
        self.sendMessage(failure.value.code, command)
    
    def _ebUnexpected(self, failure, command):
        log.err(failure)
        # This is not a very descriptive response.
        # However, it is accurate.
        self.sendMessage(irc.ERR_CANTKILLSERVER, command, ':' + str(failure.value))

    def irc_unknown(self, prefix, command, params):
        print 'Unknown(%r, %r, %r)' % (prefix, command, params)
        self.sendMessage(irc.ERR_UNKNOWNCOMMAND, command, ':Unknown command')
    
    def irc_PASS(self, prefix, params, command='PASS'):
        try:
            p = params[-1]
        except IndexError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.password, p)

    def irc_USER(self, prefix, params, command='USER'):
        try:
            user, mode, unused, realname = params[:4]
        except ValueError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.user, user, mode, unused, realname)
    
    def irc_NICK(self, prefix, params, command='NICK'):
        try:
            n = params[0]
        except IndexError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.nick, n)
    
    def irc_OPER(self, prefix, params, command='OPER'):
        try:
            name, passwd = params[:2]
        except ValueError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.oper, name, passwd)
    
    def irc_MODE(self, prefix, params, command='MODE'):
        target = mode = None
        if len(params) == 0:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
            return
        elif len(params) == 1:
            target = params[0]
        elif len(params) >= 2:
            target, mode = params[:2]
        
        if mode is None:
            if target[0] in irc.CHANNEL_PREFIXES:
                self._protectCall(command, self.avatar.channelMode, None, None)
            else:
                self._protectCall(command, self.avatar.userMode, None, None)
            return

        if mode[0] not in '+-':
            self.sendMessage(irc.ERR_UNKNOWNMODE, command)
        if len(mode) > 1:
            if target[0] not in irc.CHANNEL_PREFIXES:
                self._protectCall(command, self.avatar.userMode, mode[0] == '+', mode[1:])
            else:
                if len(params) < 3:
                    self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
                else:
                    self._protectCall(command, self.avatar.channelMode, target, mode[0] == '+', mode[1:], params[2])
    
    def irc_SERVICE(self, prefix, params, command='SERVICE'):
        try:
            nick, res1, distr, type, res2, info = params[:6]
        except ValueError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.service, nick, res1, distr, type, res2, info)
    
    def irc_QUIT(self, prefix, params, command='QUIT'):
        msg = ":Bye"
        if params:
            msg = params[0]
        self._protectCall(command, self.avatar.quit, msg)
    
    def irc_SQUIT(self, prefix, params, command='SQUIT'):
        try:
            server, comment = params[:2]
        except ValueError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.squit, server, comment)
    
    def irc_JOIN(self, prefix, params, command='JOIN'):
        if params[0] == "0":
            self._protectCall(command, self.part, None)
        elif len(params) == 1:
            channels, keys = params[0], ''
        elif len(params) >= 2:
            channels, keys = params[:2]
        else:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
            return
        
        self._protectCall(command, self.avatar.join, channels.split(','), keys.split(','))
    
    def irc_TOPIC(self, prefix, params, command='TOPIC'):
        if len(params) == 0:
            self.sendCommand(irc.ERR_NEEDMOREPARAMS, command)
        elif len(params) == 1:
            self._protectCall(command, self.avatar.topic, params[0], None)
        elif len(params) >= 2:
            self._protectCall(command, self.avatar.topic, *params[:2])
    
    def irc_NAMES(self, prefix, params, command='NAMES'):
        if len(params) == 0:
            self._protectCall(command, self.avatar.names, None)
        else:
            self._protectCall(command, self.avatar.names, params[0].split(','))
    
    def irc_LIST(self, prefix, params, command='LIST'):
        channels = server = None
        if len(params) == 0:
            self.sendCommand(irc.ERR_NEEDMOREPARAMS, command)
        elif len(params) == 1:
            self._protectCall(command, self.avatar.channels, params[0].split(','), None)
        elif len(params) == 2:
            self._protectCall(command, self.avatar.channels, param[0].split(','), params[1])

    def irc_INVITE(self, prefix, params, command='INVITE'):
        try:
            target, channel = params[:2]
        except ValueError:
            self.sendCommand(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.invite, target, channel)
    
    def irc_ADMIN(self, prefix, params, command='ADMIN'):
        if len(params) == 0:
            self._protectCall(command, self.avatar.admin, None)
        else:
            self._protectCall(command, self.avatar.admin, params[0])
    
    def irc_INFO(self, prefix, params, command='INFO'):
        if len(params) == 0:
            self._protectCall(command, self.avatar.admin, None)
        else:
            self._protectCall(command, self.avatar.admin, params[0])
    
    def irc_PRIVMSG(self, prefix, params, command='PRIVMSG'):
        try:
            recip, msg = params[:2]
        except ValueError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.privmsg, recip.split(','), msg)

    def irc_NOTICE(self, prefix, params, command='NOTICE'):
        try:
            recip, msg = params[:2]
        except ValueError:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            self._protectCall(command, self.avatar.notice, recip.split(','), msg)

    def irc_KICK(self, prefix, params, command='KICK'):
        msg = None
        if len(params) >= 2:
            channels = params[0].split(',')
            users = params[1].split(',')
        if len(params) >= 3:
            msg = params[2]
        
        if len(params) < 2:
            self.sendMessage(irc.ERR_NEEDMOREPARAMS, command)
        else:
            if len(channels) == 1:
                self._protectCall(command, self.kick, channels[0], users)
            elif len(channels) == len(users):
                for (c, u) in zip(channels, users):
                    self._protectCall(command, self.avatar.kick, c, [u])
            else:
                self.sendMessage(irc.ERR_BADCHANMASK, command)

    def irc_MOTD(self, prefix, params, command='MOTD'):
        if len(params) == 0:
            self._protectCall(command, self.avatar.motd, None)
        else:
            self._protectCall(command, self.avatar.motd, params[0])
   
    def irc_LUSERS(self, prefix, params, command='LUSERS'):
        if len(params) >= 2:
            self._protectCall(command, self.avatar.lusers, *params[:2])
        elif len(params) == 1:
            self._protectCall(command, self.avatar.lusers, params[0], None)
        else:
            self._protectCall(command, self.avatar.lusers, None, None)
    
    def irc_VERSION(self, prefix, params, command='VERSION'):
        if len(params) == 0:
            self._protectCall(command, self.avatar.version, None)
        else:
            self._protectCall(command, self.avatar.version, params[0])
    
    def irc_STATS(self, prefix, params, command='STATS'):
        if len(params) == 0:
            self._protectCall(command, self.avatar.stats, None, None)
        elif len(params) == 1:
            self._protectCall(command, self.avatar.stats, params[0], None)
        else:
            self._protectCall(command, self.avatar.status, *params[:2])
    
    def irc_LINKS(self, prefix, params, command='LINKS'):
        if params:
            self._protectCall(command, self.avatar.links, None)
        else:
            self._protectCall(command, self.avatar.links, params[0])
    
    def irc_TIME(self, prefix, params, command='TIME'):
        if params:
            self._protectCall(command, self.avatar.time, None)
        else:
            self._protectCall(command, self.avatar.time, params[0])

class ServerFactory(protocol.ServerFactory):
    protocol = IRCServer

    def __init__(self, *a, **kw):
        self.a, self.kw = a, kw
    
    def buildProtocol(self, addr):
        p = self.protocol(*self.a, **self.kw)
        p.factory = self
        p.service = self.service
        return p

class ChatService(service.Service):
    factory = ServerFactory

    def __init__(self, portal, listenOn):
        self.portal = portal
        self.strport = listenOn
    
    def startService(self):
        service.Service.startService(self)
        self.factory = self.factory(self.portal)
        self.factory.service = self
        self.port = strports.listen(self.strport, self.factory)
    
    def stopService(self):
        self.port.stopListening()
        del self.factory, self.port
        service.Service.stopService(self)
