# -*- coding: Latin-1 -*-

from twisted.python.components import Interface

from twisted.application import service

class IChatService(Interface):
    pass

class IGroup(Interface):
    def addParticipant(self, name):
        pass

    def removeParticipant(self, name):
        pass

    def __nonzero__(self):
        pass

    def receiveMessage(self, origin, message, meta):
        pass

class GroupChat:
    __implements__ = (IGroup,)

    def addParticipant(self, name):
        pass

class ChatMultiplexor(service.MultiService):
    """A collection of communication servers for various protocols.
    
    A ChatService speaks an abstract, lightweight, in-process-only protocol
    to any number of other ChatService objects.  It serves as a central
    point of communication for passing messages between these different chat
    services.
    """
    
    GroupType = GroupChat
    
    __implements__ = (IChatService,)
    
    def _allExcept(self, service, func, *a, **kw):
        return [getattr(s, func)(*a, **kw) for s in self.services if s is not service]

    def userConnected(self, service, name, address):
        """Called by this service's children when a new user has connected.
        
        @type service: Any implementor of C{IChatService}
        @param service: The service to which this user connected.

        @type name: C{str} or C{unicode}
        @param name: The handle by which this user is to be identified.
        
        @type address: C{tuple}
        @param address: A description of the transport via which this user
        connected.
        """
        self._allExcept(service, 'userConnected', self, name, address)
    
    def userDisconnected(self, service, name, address):
        """Called by this service's children when an existing user disconnects.
        
        @type service: Any implementor of C{IChatService}
        @param service: The service to which this user connected.

        @type name: C{str} or C{unicode}
        @param name: The handle by which this user is to be identified.
        
        @type address: C{tuple}
        @param address: A description of the transport via which this user
        connected.
        """
        self._allExcept(service, 'userDisconnected', self, name, address)

    def getGroup(self, service, name):
        """Return an C{IGroup} implementor for the specified group.
        
        This will either create a new group object or return one which was
        previously created.
        
        @rtype: C{IGroup} implementor
        """
        lname = name.lower()
        g = self.groups.get(lname)
        if g is None:
            g = self.groups[lname] = self.GroupType(name)
        return g

    def destroyGroup(self, service, name):
        """Get rid of a group.
        
        It must be empty first.  This should typically be called from an
        {IGroup} object's removeParticipant() method, when the group notices
        itself to be empty.
        
        @raise KeyError: Raised if the specified group name does not exist.
        """
        lname = name.lower()
        assert not self.groups[lname], "Destroying populated groups is an error."
        del self.groups[lname]

    def listGroups(self):
        """Lists all groups which exist.
        
        @rtype: C{list} of implementors of C{IGroup}
        @return: An iterator of all the groups which exist.
        """
        return self.groups.values()
