# -*- coding: Latin-1 -*-

import time

from twisted.im import basechat
from twisted.im import interfaces
from twisted.im.ircsupport import IRCAccount
from twisted.im.tocsupport import TOCAccount
from twisted.im.pbsupport import PBAccount
from twisted.internet import defer
from twisted.python import log

from quotient import formless
from quotient.grabbers import username, password, host, port
from quotient.items.conversation import Conversation

oProto = formless.Choice(
    label = "Network Protocol",
    choices = ["IRC"],
    description = "The protocol the target network runs",
)

oUser = formless.String(
    label = "Network Username",
    description = "The username with which to log onto the target network",
)

oPass = formless.Password(
    label = "Network Password",
    description = "The password with which to log onto the target network",
)

host = formless.String(
    label = "Network address",
    description = "The hostname of the target network's server",
)
port = formless.Integer(
    label = "Network port",
    description = "The port number of the target network's server",
)

class IAccount(formless.TypedInterface):
    protocol = oProto
    username = oUser
    password = oPass
    
    host = host
    port = port

iProto = formless.Choice(
    label = "Inbound Protocol",
    choices = ["IRC"],
    description = "The protocol you will connect to the proxy with",
)
iUser = formless.String(
    label = "Inbound Username",
    description = "The username via which to allow access to the proxy",
)
iPass = formless.Password(
    label = "Inbound Password",
    description = "The password via which to allow access to the proxy",
)

class IInstantMessagingFactory(formless.TypedInterface):
    account = formless.Object(interface=IAccount)

IInstantMessagingObject = formless.Object(
    interfaces.IAccount,
    label="Instant Messaging Account",
    description="An account on an instant messaging network",
)

class PresenceStatus:
    def __init__(self, message='***ERROR***', success=False):
        self.lastChanged = time.time()
        self.message = message
        self.success = success

    def __str__(self):
        return self.message

    def __repr__(self):
        return '<%s: %r>' % (self.__class__, self.message)

class PresenceOffline(PresenceStatus):
    def __init__(self, reason):
        PresenceStatus.__init__(self, "Offline: %s" % (reason,))

class PresenceOnline(PresenceStatus):
    def __init__(self, reason):
        PresenceStatus.__init__(self, "Online: %s" % (reason,))

class PresenceChanging(PresenceStatus):
    def __init__(self, reason):
        PresenceStatus.__init__(self, "In Progress: %s" % (reason,))

def rotateWrap(f):
    def function(self, *args, **kw):
        r = f(self, *args, **kw)
        if self.conv.shouldRotate():
            self.rotateConversation()
    return function

class QuotientConversation(basechat.Conversation):
    __implements__ = (interfaces.IConversation,)

    def __init__(self, person, chatui):
        basechat.Conversation.__init__(self, person, chatui)
        print 'QuotientConversation(%r, %r)' % (person, chatui)
        self.conv = Conversation(chatui.avatar.store, person.name)
        self.name = chatui.avatar.localpart

    def rotateConversation(self):
        self.conv.save()
        self.conv = Conversation(self.chatui.avatar.store, self.person.name)
    
    def show(self):
        pass
    
    def hide(self):
        pass
    
    def sendText(self, text, metadata):
        print 'QuotientConversation.sendText(%r, %r)' % (text, metadata)
        if metadata.get('style', None) == 'emote':
            self.conv.addAction(self.name, text)
        else:
            self.conv.addMessage(self.name, text)
        return basechat.Conversation.sendText(self, text, metadata)
    sendText = rotateWrap(sendText)
    
    def showMessage(self, text, metadata):
        print 'QuotientConversation.showMessage(%r, %r)' % (text, metadata)
        self.chatui.user.proto.sendLine(text)
        if metadata.get('style', None) == 'emote':
            self.conv.addAction(self.person.name, text)
        else:
            self.conv.addMessage(self.person.name, text)
        # return basechat.Conversation.showMessage(self, text, metadata)
    showMessage = rotateWrap(showMessage) 
       
    def changedNick(self, person, newnick):
        print 'QuotientConversation.changedNick(%r, %r)' % (person, newnick)
        self.conv.addMeta('%s is now known as %s' % (person.name, newnick))
        return basechat.Conversation.changedNick(self, person, newnick)
    changedNick = rotateWrap(changedNick)

class QuotientGroupConversation(basechat.GroupConversation):
    __implements__ = (interfaces.IGroupConversation,)
    
    topic = None

    def __init__(self, group, chatui):
        basechat.GroupConversation.__init__(self, group, chatui)
        self.conv = Conversation(chatui.avatar.store, group.name)
        self.name = chatui.avatar.localpart

    def rotateConversation(self):
        self.conv.save()
        self.conv = Conversation(self.chatui.avatar.store, self.group.name)

    def show(self):
        pass
    
    def hide(self):
        pass

    def c(self):
        return self.chatui.user.proto

    def sendText(self, text, metadata):
        self.c().sendLine(text)
        if metadata.get('style', None) == 'emote':
            self.conv.addAction(self.name, text)
        else:
            self.conv.addMessage(self.name, text)
        return basechat.GroupConversation.sendText(self, text, metadata)
    sendText = rotateWrap(sendText)
    
    def showGroupMessage(self, sender, text, metadata):
        if metadata.get('style', None) == 'emote':
            self.c().action('%s!%s@%s' % (sender, sender, 'quotient'), '#' + self.group.name, text)
            self.conv.addAction(sender, text)
        else:
            self.c().privmsg('%s!%s@%s' % (sender, sender, 'quotient'), '#' + self.group.name, text)
            self.conv.addMessage(sender, text)
        # return basechat.GroupConversation.showGroupMessage(self, sender, text, metadata)
    showGroupMessage = rotateWrap(showGroupMessage)

    def setTopic(self, topic, author):
        if self.topic is None:
            self.c().topic(self.name, '#' + self.group.name, topic, None)
        else:
            self.c().topic(None, '#' + self.group.name, topic, author)
        self.topic = topic
        self.conv.addMeta('%s changed the topic to %r' % (author, topic))
    setTopic = rotateWrap(setTopic)

    def setGroupMembers(self, members):
        basechat.GroupConversation.setGroupMembers(self, members)
        self.c().names(self.name, '#' + self.group.name, members)

    def memberJoined(self, member):
        self.c().join(member, '#' + self.group.name)
        self.conv.addMeta('New participant: %s' % (member,))
        return basechat.GroupConversation.memberJoined(self, member)
    memberJoined = rotateWrap(memberJoined)

    def memberChangedNick(self, oldnick, newnick):
        print 'memberChangedNick(%r, %r)' % (oldnick, newnick)
        self.conv.addMeta('%s is now known as %s' % (oldnick, newnick))
        return basechat.GroupConversation.memberChangedNick(self, oldnick, newnick)
    memberChangedNick = rotateWrap(memberChangedNick)

    def memberLeft(self, member):
        self.c().part(member, '#' + self.group.name)
        self.conv.addMeta('Participant departed: %s' % (member,))
        return basechat.GroupConversation.memberLeft(self, member)
    memberLeft = rotateWrap(memberLeft)

class ChatListener(basechat.ChatUI):
    """Aggregator of one or more conversations for a particular account"""
    __implements__ = (interfaces.IChatUI,)

    def __init__(self, avatar, user):
        basechat.ChatUI.__init__(self)
        self.avatar = avatar
        self.user = user

    def getConversation(self, person, Class=QuotientConversation, stayHidden=False):
        return basechat.ChatUI.getConversation(self, person, Class, stayHidden)
    
    def getGroupConversation(self, group, Class=QuotientGroupConversation, stayHidden=False):
        return basechat.ChatUI.getGroupConversation(self, group, Class, stayHidden)

class IRCAccount(IRCAccount):
    __implements__ = (IAccount,)
    protocol = "IRC"
    
    username = "DEFAULT"
    password = "PWD DEFAULT"

class TOCAccount(TOCAccount):
    __implements__ = (IAccount,)
    protocol = "TOC"

class PBAccount(PBAccount):
    __implements__ = (IAccount,)
    protocol = "PB"

class InstantMessagingFactory:
    __implements__ = (IInstantMessagingFactory,)
    status = None
    loggedOn = False

    def __init__(self, avatar, portal, account):
        self.avatar = avatar
        self.portal = portal
        self.account = account
        self.avatar.addTask(self)
        self.setStatus(PresenceOffline("User requested"))

    def setStatus(self, s):
        self.status = s
        print 'InstantMessagingFactory.status =', s

    def format(self):
        return '%s:%s@%s' % (self.account.protocol, self.account.username, self.account.host)

    def run(self):
        pass
    
    def logon(self, user):
        self.loggedOn = True
        self.setStatus(PresenceChanging("Going online"))
        return self.account.logOn(ChatListener(self.avatar, user)
            ).addCallbacks(self._cbLogOn, self._ebLogOn
            )

    def _cbLogOn(self, result):
        self.client = result
        self.setStatus(PresenceOnline("Woop woop"))
        return self.client

    def _ebLogOn(self, failure):
        log.err(failure)
        self.loggedOn = False
        self.setStatus(PresenceOffline(str(failure)))
