# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from types import ClassType
from os.path import join as opj

from quotient import storq
from quotient.storq import Store, Avatar, Item, Pool
from quotient.storq import ITEM, INDEX, STORE_ID

from twisted.trial import unittest

#Why doesn't this work?  I don't know.
#def makeCaller(name):
#    def caller(self):
#        return self.store.transact(getattr(self, name))
#    return caller
#
#class StorageTestMetaclass(type):
#    def __new__(self, name, bases, attrs):
#        newAttrs = {}
#        for f in attrs:
#            if f.startswith('store'):
#                newAttrs['test' + f[len('store'):]] = makeCaller(f)
#        attrs.update(newAttrs)
#        return type(name, bases, attrs)

class LenientPoolIndexIterator:
    didFirst = False

    def __init__(self, store, pool, index, start=None, end=None, result=INDEX):
        self.store = store
        self.pool = pool
        self.index = index
        self.start = start
        self.end = end
        self.result = result
    
    ##
    ## ~25.5s
    def next1(self):
        r = self.pool.queryIndex(
            self.index, startKey=self.start, endKey=self.end,
            result=self.result, skipFirst=self.didFirst)
        self.start = self.store.transact(r.next)
        self.didFirst = True
        return self.start

    ##
    ## ~13s
    def next2(self):
        r = self.pool.queryIndex(
            self.index, startKey=self.start, endKey=self.end,
            result=self.result)
        def _():
            if self.didFirst:
                r.next()
            return r.next()
        self.start = self.store.transact(_)
        self.didFirst = True
        return self.start
    
    ##
    ## doesn't work
    def next3(self):
        if self.didFirst:
            start = self.start + chr(255)
        else:
            self.didFirst = True
            start = self.start
        r = self.pool.queryIndex(
            self.index, startKey=self.start, endKey=self.end,
            result=self.result)
        self.start = self.store.transact(r.next)
        return self.start

    next = next2

    def __iter__(self):
        return self

class StoreMixin:
    def createAvatar(self):
        return Avatar(self.store)

    def setUpClass(self):
        # Speed up the test cases by lowering the hard limits we're worried about bumping into
        self.MAX_LOCKS, storq.MAX_LOCKS = storq.MAX_LOCKS, 50
        self.MAX_LOCKERS, storq.MAX_LOCKERS = storq.MAX_LOCKERS, 50
        self.MAX_OBJECTS, storq.MAX_OBJECTS = storq.MAX_OBJECTS, 50

    def tearDownClass(self):
        storq.MAX_LOCKS = self.MAX_LOCKS
        storq.MAX_LOCKERS = self.MAX_LOCKERS
        storq.MAX_OBJECTS = self.MAX_OBJECTS

    def setUp(self):
        d = self.mktemp()
        self.store = Store(opj(d, "db"), opj(d, "files"))
        if getattr(getattr(self, self.caseMethodName).im_func, 'insertItems', True):
            self.setupDatabase()

    def setupDatabase(self):
        self.avatars = []
        for i in range(self.nAvatars):
            pools = []
            a = self.store.transact(self.createAvatar)
            def _():
                p = Pool(a)
                nItems = max(1, self.nItems / self.nPools)
                for i in range(nItems):
                    p.addItem(Item(a))
                pools.append((p, nItems))
            for p in range(self.nPools):
                self.store.transact(_)
            self.avatars.append((a, pools))
        self.totalItems = self.nAvatars * max(1, self.nItems / self.nPools) * self.nPools

    def tearDown(self):
        self.store.close()

class FewAvatars:
    nAvatars = 2

class ManyAvatars:
    nAvatars = 20

class FewItems:
    nItems = 5

class ManyItems:
    nItems = 50

class FewPools:
    nPools = 3

class ManyPools:
    nPools = 30

class TWSOGMM:
    """The Whole Sort Of General Mish Mash Test Case
    
    Tests a variety of access styles with a variety of object configurations.
    """

    def testSequentialAccess(self):
        # Exhaust all messages in all pools of all avatars one after another
        iterations = 0
        for (avatar, pools) in self.avatars:
            for (p, nItems) in pools:
                L = list(LenientPoolIndexIterator(self.store, p, 'uid', result=INDEX))
                self.assertEquals(len(L), nItems)
                self.assertEquals(L, range(nItems))
                iterations += nItems
        # print iterations
        self.assertEquals(iterations, self.totalItems)

    def testParallelAccess(self):
        # Exhaust all messages in all pools of all avatars
        # alongside/intermingled with (IN PARALLEL TO) each other
        iterators = []
        for (avatar, pools) in self.avatars:
            for (p, nItems) in pools:
                iterators.append((LenientPoolIndexIterator(self.store, p, 'uid', result=INDEX), 0, nItems))
        iterations = 0
        while iterators:
            keep = []
            for (i, expected, nItems) in iterators:
                try:
                    self.assertEquals(i.next(), expected)
                    iterations += 1
                except StopIteration:
                    # print 'Stopping', expected, nItems
                    self.assertEquals(expected, nItems)
                else:
                    # print 'Continuing', expected, nItems
                    keep.append((i, expected + 1, nItems))
            iterators = keep
        # print iterations
        self.assertEquals(iterations, self.totalItems)

    def testUIDAccess(self):
        for (avatar, pools) in self.avatars:
            for (p, nItems) in pools:
                uids = list(LenientPoolIndexIterator(self.store, p, 'uid', result=STORE_ID))
                for u in uids:
                    self.assertEquals(avatar.getItemByID(u).storeID, u)

bases = (StoreMixin, TWSOGMM, unittest.TestCase)
for a in (FewAvatars, ManyAvatars):
    for i in (FewItems, ManyItems):
        for p in (FewPools, ManyPools):
            name = 'TestCase_%s_%s_%s' % (a.__name__, i.__name__, p.__name__)
            globals()[name] = ClassType(name, (a, i, p) + bases, {})
del a, i, p, name, bases
