import urllib, random, time

from twisted.trial import unittest
from twisted.internet import reactor, defer
from twisted.python import log, components
from twisted.protocols import policies
from twisted.spread import pb

import quotient
from quotient.web import newclient, functional



class IJob(components.Interface):
    """a job to be managed by a TaskFactory
    @ivar jobDone: a deferred fired after job completion
    @ivar master: the taskMaster that's controlling this job
    """
    def doWork(self):
        """called to initiate work on this job
        @returns defer.Deferred: called back when job is done
        """
        # TODO: Change this so the user doesn't have to call clockIn, 
        # because that's an implementaiton detail
        # if you want to do job throttling, you need to call 
        # self.master.clockIn()
        pass


class ITaskFactory(components.Interface):
    def getAJob(self):
        """override this function in subclasses to return an object that implements IJob"""
        pass

    def getToWork(self):
        """call this method to start the tasks"""
        pass

    def timeoutFactory(self):
        """override to change behavior on timeout, by default calls allJobsCompleted"""
        pass

    def allJobsCompleted(self):
        """called when the factory has run all jobs"""
        pass


class TaskFactory(object, policies.TimeoutMixin):
    """manages a queue of jobs for a group of workers
    @cvar totalJobsLimit: this is the number of jobs to be run by the workers
    @cvar queuedJobsLimit: this is the maximum number of jobs allowed to be queued
    and waiting for assignment by this master at any one time 
    @cvar activeJobsLimit: how many workers are allowed to be working at any one time
    """
    __implements__ = (ITaskFactory,)

    totalJobsLimit = 100     
    activeJobsLimit = 10
    jobsPerSecond = 8
    clockedInThisSecond = 0
    timeout = 15 
    _totalCompletedJobs = 0
    avgDuration = 0.1
    punchClock = {}
    activeJobs = 0 

    numActiveJobs = property(lambda self: len(self.punchClock))
    lenJobQueue = property(lambda self: len(self.jobQueue))

    def _getState(self):
        return """
_totalCompletedJobs = %s
totalJobsGenerated = %s
punchClock = %s
activeJobs = %s
clockedInThisSecond = %s 
""" % (self._totalCompletedJobs, 
       self.totalJobsGenerated, 
       self.punchClock, 
       self.activeJobs,
       self.clockedInThisSecond)

    state = property(_getState)

    def __init__(self):
        """constructor
        @param reqDelay: delays each call by n seconds, argument to reactor.callLater
        @note: if you want the code to run lockstep (no concurrent requests) use reqDelay = None
        """
        self.delayeds = []
        self.totalJobsGenerated = 0        # total number of jobs that have been queued

    def getAJob(self):
        """override this function in subclasses to return an object that implements IJob"""
        pass

    def getToWork(self):
        log.debug("TaskFactory.getToWork")
        self.setTimeout(self.timeout)
        self.starttime = time.time()
#        d = self._calibrate()
#        d.addCallback(lambda _: self._consume())
        self._consume()
    
#    def _calibrate(self):
#        # figure out how long one iteration of the reactor takes
#        start = time.time()
#        d = defer.Deferred()
#        def _():
#            lag = time.time() - start
#            self.avgDuration += lag
#            log.msg("adjusting for %s seconds of lag" % lag, lag=True)
#            d.callback(None)
#        reactor.callLater(0, _)
#        return d

    def _consume(self):
        self.resetTimeout()

        if self.totalJobsGenerated >= self.totalJobsLimit:
            return

        if self.activeJobs < self.activeJobsLimit:
            #d = self._calibrate()

            def _sked(_):
                self.totalJobsGenerated += 1
                self.activeJobs += 1

                dispatchTime = self.avgDuration

                log.debug(state=self.state)

                j = self.getAJob()

                j._jid = self.totalJobsGenerated

                # wait the calculated amount of time and then run job
                #log.msg('consume: scheduling job %s for %s sec in the future' % (j._jid, dispatchTime))
                #def _(j):
                d = defer.maybeDeferred(j.doWork)
                d.addBoth(self._clockOut, j._jid)

                #self.delayeds.append(reactor.callLater(dispatchTime, _, j))

            _sked(None)
            #d.addCallback(_sked)
        #self.clockedInThisSecond = 0
        self.delayeds.append(reactor.callLater(0, self._consume))
   
    def clockIn(self, id=None):
        self.resetTimeout()
        now = time.time()
        self.punchClock[id] = now
        log.msg("Job %s Clocked In! time: %s" % (id, now - self.starttime), clockIn=None)
        self.clockedInThisSecond += 1

    def _clockOut(self, result, id):
        self.resetTimeout()
        self._totalCompletedJobs += 1
        self.activeJobs -= 1

        log.debug(state="%s" % self.state)
        dur = time.time() - self.punchClock[id]
        log.msg("Job %s Clocked Out! running time %s sec" % (id, dur), clockOut=True)
        assert self._totalCompletedJobs != 0
        self.avgDuration = (dur + self.avgDuration) / 2

        log.msg("average job duration: %s" % self.avgDuration, avg=True)

        del self.punchClock[id]

        if (self._totalCompletedJobs == self.totalJobsLimit and
                self.activeJobs == 0): 
            return self._jobsCompleted()


    def timeoutFactory(self):
        self._jobsCompleted()

    # HACK!
    timeoutConnection = timeoutFactory    

    def _jobsCompleted(self):
        self.setTimeout(None)
        [d.cancel() for d in self.delayeds if d.active()]
        log.msg("zombies: %s" % len(self.punchClock))
        self.allJobsCompleted()

    def allJobsCompleted(self):
        pass


class IPBTaskFactory(components.Interface):
    def remote_getToWork(self):
        """calls getToWork on a remote TaskFactory
        @returns: a deferred that is called back when all tasks are completed
        """
        
    def remote_config(self, **kw):
        """configures class variables on the remote ITaskFactory
        implementation. pass a dict of {'classvarname': 'value'}"""


from twisted.spread import pb

class PBTaskFactoryAdapter(components.Adapter, pb.Root):
    __implements__ = (pb.IPBRoot,)

    def remote_getToWork(self):
        self.original.getToWork()

    def remote_config(self, **kw):
        for k, v in kw.iteritems():
            if hasattr(self.original, k):
                setattr(self.original, k, v)

    def rootObject(self, broker):
        return self


components.registerAdapter(
        PBTaskFactoryAdapter,
        TaskFactory,
        pb.IPBRoot
)



