# -*- test-case-name: quotient.test.test_todo -*-
#
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re
from quotient import itemstore, itempool, util, formless
from quotient.items.phones import Match, LINE_COST
from quotient.provisioner import ETank

class ITodo(formless.TypedInterface):
    text = formless.Text()
    
    attachments = formless.List()


class ITodoManager(formless.TypedInterface):
    def addTodo(self, text=formless.Text()):
        """Add Todo
        
        Add a Todo to your list.
        """
        return formless.Object(interface=ITodo, label="Todo", description="A Todo instance")
    addTodo = formless.autocallable(addTodo)

    currentTodos = formless.List()
    

class PoolRemover(list):
    def __init__(self, pool):
        super(PoolRemover, self).__init__(pool)
        self.pool = pool

    def remove(self, obj):
        self.pool.removeItem(obj)
        super(PoolRemover, self).remove(obj)


class TodoManager(object):
    __implements__ = ITodoManager

    def __init__(self, pool):
        self.pool = pool

    def addTodo(self, text):
        t = Todo(self.pool.store, text)
        self.pool.addItem(t)
        return t

    def _currentTodos(self):
        return PoolRemover(self.pool)
    currentTodos = property(_currentTodos)

    #def save(self):
    # 1) how do we remove an item from the pool after it has been removed from
    # the currentTodos list;
    # 2) what do we do in save
        

class Todo(itemstore.Item):
    __implements__ = ITodo,

    """
    attribute plan:
    id -- human-readable number
    what -- title/subject for task
    when -- deadline
    who  -- current owner of the task (defaults to creator)
    depends, reverseDepends -- tasks related to this one
    subscribers -- who will be notified on change
    parent, children -- nesting/subtask relationships
    priority
    ordinal -- arbitrary ordering
    status -- what's currently happening
    type -- bug? feature? wish?
    estimate -- how long will this take, approximately?
    """

    def __init__(self, store, text=''):
        itemstore.Item.__init__(self, store)
        self.text = text
        self.attachments = []
        # subparts, somehow
        
    def __str__(self):
        """__str__ is used to render a preview of this todo.
        
        Only show the first line, if it has multiple lines.
        """
        return self.text.split('\n')[0]

    def setText(self, text):
        self.text = text
        self.save()

    def indexes_fulltext(self):
        return {'text':self.text}

    def attachItem(self, item):
        self.attachments.append(item.referTo())
        self.save()

qre = re.compile(r'''(?:^|[\.\?\!]\s) # end of the previous sentence
                   (.*\?) # # sentence, then a ques
                   (?:\s|$) # then whitespace''',re.X)

class TodoCollector(util.Forgetter):
    persistenceForgets = ['store']
    contextRemembers = [('store', 'store')]

    def __init__(self, store):
        self.store = store
        
    def findQuestions(self, m):
                                
        avatar = self.store.avatar
        prov = avatar.provisioner
        for part in m.walk():
            if part.inferType().startswith('text/'):
                byteOfft = 0
                for line in part.getBody().xreadlines():
                    if not prov.deplete(ETank.EXTRACTION, LINE_COST, True):
                        break
                    yield None
                    mat = qre.search(line)
                                            
                    if mat is not None:
                        begin, end = mat.span()
                        begin += byteOfft
                        end += byteOfft
                        ##
                        ques = mat.groups()[0]
                        match = Match(m, part.partnum, begin, end)
                        m.addAction(CollectTodo(m, match, ques))
                        ##
                    byteOfft += len(line)
        avatar.save()

    def itemMovedIn(self, pool, message):
        for x in self.findQuestions(message):
            pass
        
    itemRoutedIn = itemMovedIn

    def itemRoutedOut(self, pool, message):
        pass
    itemMovedOut = itemRoutedOut
    
class CollectTodo:
    def __init__(self, message, match, ques):
        self.match = match
        self.ques = ques
        self.question = "Create a todo for '%s'" % (ques)
        self.message = message
        
    def yes(self):
        # if i were a better person, this would get an ItemRef to the
        # todo pool from the TodoCollector, which gets it at
        # instantiation
        t = Todo(self.message.store, text=self.ques)
        t.attachItem(self.message)
        st = self.message.store
        st.getItemByID(st.avatar.pools['todo']).addItem(t)
        t.save()
        self.message.actions.remove(self)
        
    def no(self):       
        self.message.actions.remove(self)
